/*  maskdat.h -- Definitions of mask data structures */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **
**                                                                   **
**                   Copyright (C) 2011 by                           **
**     Ken Clardy, Carnegie Observatories, Pasadena California.      **
**                                                                   **
**      This software is proprietary, and may be used or copied      **
**          only with the written consent of the author.             **
**              All copies must retain this notice.                  **
**                                                                   **
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef  INCLUDE_MASKDAT_H
#define  INCLUDE_MASKDAT_H

/*  ====  Global Macros  ====  */
/* Support for the data types defined here... */

// #define  C_loop(q,a)  for(q=(a); q!=NULL; q=(q->next!=(a))?q->next:NULL)
// C_loop is now defined in ioutils.h which is included next.


/* --NOTE-- problem with "passband" structure.  Defined in and needed
  in the optics definitions header, it is also used here in this
  header.  We probably should include the optics header to get it; or
  put it in a more general header used by both them and us */
/* -- optics.h header information is now here. */

/* May want to put vector definitions from kdcutils here... */

#include	IOUTILH

#include	FEATSH

/* NOTE -- order changed above because ioutils now needs the definition
  of vect2 from kdcutil to occur first.  If this causes problems, put
  an include of kdcutil.h into ioutil.h and remove the kdcutil.h
  include here.   */

/* NOTE -- ioutils.h now indludes kdcutil.h as it is needed there; so
  there is no include of kdcutil.h from this header.        */

/*  ===  Optics Definitions  ===  */

#define  Angstrom  1.0e-7
#define  Micron  1.0e-3
#define  Inch	25.4
#define CAM_ANGLE	( (180.0 + 45.0) * Degree )


/* Other fundamental structures */

typedef	struct	intval {	/* ordered interval */
	double	lo;
	double	hi;
	} intval;

typedef	struct	bbox {		/* bounding box */
	intval	x;
	intval	y;
	} bbox;

/*  === Optics data structures from optics.h === */
/* optics.h is now obsolete, its definitions are here. */

/* passband applies to filters, detector sensitivity, spect. limits */
typedef  struct  pb {
	double	blue;	/* Shortest wavelength */
	double	red;	/* Longest wavelength */
	} passband;

typedef  struct  filt {
	struct  filt  *next, *last;
	passband	pass;
//	int		ident;		/* Internal identification */
	char		name[32];
	} filter;


/* detector must be defined in the detector plane, with
 pixels/mm and borders, areas between live elements, etc. */
typedef  struct  det {
//	passband	sens;
	bbox		bb;	/* Edges */
	vect2		pcount;	/* Pixel count each direction */
	vect2		psize;	/* Pixel sizes each direction */
	double		vrad;	/* Vignetting Radius */
/* This may change to a more detailed map of pixel space in the
  linear coordinates */
	bbox		bbs;	/* Bounds of nod&shuffle detector */
	bbox		bbn;	/* Bounds of normal detector */
	bbox		bbd;	/* Bounds of default detector */
	} detector;

/* Optics definitions for "optutils.h" here */

/*  ====  Optics data structures  ====  */

/* Generalized polynomial */

typedef struct poly {
	int	dummy;	// for alignment
	int     order;
	double  *coef;
	} poly;
/* coef will contain order+1 values */

typedef struct cpoly {
	int	dummy;	// for alignment
	int	order;
	vect2	*coef;
	} cpoly;
/* coef will contain order+1 values,
  each a complex quantity.  */

typedef  struct  datlist {
	struct datlist	*next;
	char	*data;
	struct element *eg;
	} datlist;
// NOTE that datlist is now specific to element
// instrument queues.


// element IS SIMILAR TO xlist STRUCTURE IN optix.h
/* Structure(s) for the gmap ini data stuff... */
typedef  struct  element {
	struct element  *next, *last;
	char	*name;
	int	kw;	// Type of element. (a Keyword value)
	int	flag;		/* The default flag, maybe others */
	void   *data;
	struct  datlist  *head;
// CHANGE THIS TO POINTER TO (VOID) DATA STRUCTURE ?
	} element;

/* Bit values for "flag" word: 
	1 = Default element (not used here)
	2 = Alternate (x) axis modifier
..*/

/* Structures for GISMO operation */
// #ifdef  GISMO
typedef	struct	sbfld {
	bbox	field;
	cpoly	*trans;
	}  sbfld;

/* The "sbfld" structure is a description of the input field boundries
  as a bounding box, and the transform function as a cpoly.  Typically,
  several of these are used to make a set of subfields for a subfield
  instrument such as "gismo".  For that, the data is stored in a
  "sfset" structure containing n "sbfld" structures.  */

typedef	struct	sfset {
	int	n;		// number of sbfld elements
	sbfld	*s;		// array of sbfld elements
	}  sfset;

// #endif  // on GISMO

typedef	struct	stgrism {
	double	lpmm;
	double	angle;
	char*	glass;
	element*  eg;
	} stgrism;
/* The glass string gets matched to a glass entry containing the
  index polynomial when this is used. */
/* Optionally, we could add a pointer to the polynomial, and fill it
  in after reading. */

typedef	struct	egrism {
	double	inangle;
	char	*inglass;
	element	*ig;
	double	gangle;
	double	lpmm;
	char	*eglass;
	element	*eg;
	double	eangle;
	} egrism;
/* The enhanced grism has a lead angle and glass, then a
  grating at some other angle, and an exit glass, ending
  at some exit angle.  Either glass may be null.  Zero angle
  is normal to coordinate system, usual input is negative
  angle and exit is positive angle.  */

typedef	struct	stechel {
	vect3	pris1;
	vect3	pris2;
	vect3	grnor;
	vect3	grdis;
	double	lpmm;
	char*	glass;
	element*  eg;
#ifdef  GTILTMOE
	double	gtilt;	// debug - tilt angle for grating.
	vect3	gtnor;	// debug tilted normal
	vect3	gtdis;	// debug tilted dispers
#endif
	} stechel;
/* Done similarly to grism with vector components
  rather than defining angles.  */

/*  (( Might need the focdat structure ?  ))  */

typedef	struct	gapdat {
	struct	gapdat	*next, *last;
	char	*name;
	char	*instr;
	int	ns;
	double	x1u, x1b;
	double	x2u, x2b;
	} gapdat;
/* Since the gaps need not be orthogonal to coordinates, they can't
  be stored as a bounding box.  The x-u values are at the upper edge
  of the detector, and the x-b at the bottom edge.  */

#define  GAPHEAD  "GapHead"
// GAPHEAD is name of the special gap header element


/* ... End of optutils data structures */

/*  =====  Mask space data  =====  */

typedef struct  avoids  {
        struct avoids  *next, *last;
//      vect2   c;
        double  x, y;
        double  r;
        }  avoids;

// ( This is from cututil.h )
#ifdef  USE_CDAT
/* We expect that this symbol is never defined, but it could
  be defined in mgfeats.h if needed to improve speed.  If defined,
  it should be populated just after cut parameters are read,
  probably used as an array of structures indexed by inst, and
  filled by a subroutine, or filled by the program(s) which read
  the data file itself.  Possibly used as an alternate method of
  data reading and storage to GS_lookup and BNODE methods.  */
typedef struct  {
        vect2   curve;  // diopters, in 2 directions
        double  beam;   // diameter, mm.
        double  slew;   // rate mm/min
        double  cut;
        double  fine;
        double  afoc;   // autofocus range
        double  dx, dy, dz;
	double	flange;
	double	aangle;
	double	aclear;
	double	LongSlit;
	double	SlitGap;
	double	tool;
	double  cte;
	double	dmx;
	double	tcut;
	double	acon;	// acceleration
        avoids  *av;    // avoidance queue
	int	mma;
        }  cdata;
/* cdata contains all the cutting parameters for any given
  instrument.  Typically used in an array by instrument, or
  passed as a (pointer) subroutine argument when used.  */
#endif

/* The object data structure -- data from object list */
typedef  struct  objdat {
//	void	*dummy;		// for alignment
	char	*name;		/* Name, dynamic string */
	char	*comment;	/* In-line comment, dynamic */
	double	ra, dec;	/* Celestial coordinates */
	double	priority;	/* For optimization */
	double	basepri;	/* Basic priority for apri feature */
	namlist	*precom;	/* Pre-comments */
	namlist	*postcom;	/* Post-comments */
	} objdat;

enum	slitshape {
	CIRCLE,
	SQUARE,
	RECTANGLE,
/* Add special, complex shapes here... */
	SPACED_CROSS
	};

typedef  struct  slit {
	int	dummy;		// for alignment
	int	shape;		/* See the shape enum */
	double	width;		/* Width or only dimension */
	double	alen;		/* Length to left */
	double	blen;		/* length to right */
	double	angle;		/* Position angle in radians */
//	double	curve;		/* Curvature in diopters */
	} slit;

// NOTE -- These conflict types are yet unused
enum	conflict_types {
	NO_CONFLICT,
	OD_Y,
	OD_X,
	INT_SPECT,
	INT_IMAGE,
	DUPLICATE,
	BAD_SPOT
	};

/* Conflict node */
typedef  struct  conf_tag {
	struct  conf_tag  *next, *last;
	int	type;
	struct	conf_tag  *bak, *fwd;	/* 2-way per-object list */
	struct  objq  *obj;	/* Primary object back-pointer */
	struct  objq  *other;	/* Other object */
	} cfl;

/* Spectrum edge descriptor */
typedef  struct  edge_tag {
	vect2	p;	/* Position of center */
	vect2	r;	/* Red end */
	vect2	u;	/* blue end */
	double	a, b;	/* Coeficients of curve */
	} sp_edge;

/* Spectrum on Detector descriptor */
/* Used for detail conflict resolution computation */
typedef  struct  sod_tag {
	struct  sod_tag  *next;	/* Allow pop-list for images */
	int	order;		/* The order of this spectrum */
/* The spectrum edges, stored as above */
	sp_edge	e1, e2;
/* The bounding box limits for full compares */
	bbox	bb;
/* Second bounding box for on-detector test */
	bbox	bd;
	vect2	center;    /* DEBUG object at cw location */
	int	on_det;	/* see if any part is on the detector */
	} spect;

/* Object values -- data for the object's spectrum */
typedef  struct  objval {
/* Put here the projected spectra positions on detector as vect2 values */
	/* Pointer to main spectrum traced boundry */
	/* Queue to ghost image spectra */
	int		ncf;	/* Conflict count */
	cfl		*cfq;	/* Conflict queue pointer */
	spect		*sod;	/* Primary spectrum */
	spect		*img;	/* Other order image list */
	bbox		dimg;	/* Direct image limits */
	int		stat[2];	/* Extension status values */
	int		flag;	/* DEBUG flag, normally 0 */
	} objval;


enum  object_types {
	OBJ_UNKNOWN,
	OBJ_OBJECT,
	OBJ_REFERENCE,
	OBJ_GAP,
	OBJ_AREA
	};

/*  Bits in objq.flag */
#define  OBJECT_FLAG    1
#define  OBJECT_ACTIVE  4
#define  OBJECT_SPECIAL 8
#define  OBJECT_VIRTUAL 16
#define  OBJECT_SERVICE (1<<5)
#define  OBJECT_ABSPA  (1<<6)
#define  OBJECT_POSIND (1<<7)

#ifdef  OBJCHECK
/* Bits in cflag check flag */
#define  OC_NEGUSE (1 <<  0)
#define  OC_PRIX   (1 <<  1)
#define  OC_NSHORT (1 <<  2)
#define  OC_PREQ   (1 <<  3)
#define  OC_WIDTH  (1 <<  4)
#define  OC_WIDEX  (1 <<  5)
#define  OC_SHAPE  (1 <<  6)
#define  OC_ALEN   (1 <<  7)
#define  OC_BLEN   (1 <<  8)
#define  OC_LENGTH (1 <<  9)
#define  OC_ANGLE  (1 << 10)
#define  OC_UNCUT  (1 << 11)
#define  OC_MOEX   (1 << 12)
#define  OC_ABXU   (1 << 13)
//sunifdef < #ifdef  OCKWX
//sunifdef < #define  OC_POSV   (1 << 14)
//sunifdef < #define  OC_KEYW   (1 << 15)
//sunifdef < #endif
#endif

/* A queue of objects read in, and having slits assigned */
typedef  struct  objq {
	struct  objq	*next, *last;
	objdat		dat;
	vect2		smpos;	/* Position on the mask (was vect3) */
//#ifdef  GISMO
	vect2		fpos;	/* Focal Plane position */
/* NOTE - fpos is same a smpos except when subfielding is present, then
  fpos is the input focal plane position, and smpos is the output from
  the subfield element.  Use fpos for differential refraction, for example,
  and smpos for cutting slits.
*/
//#endif
	slit		slit;
//#ifdef  GISMO
	slit		islit;	/* Incident image slit */
//#endif
	objval		*val;	/* Conflict data */
	int		type;
	int		flag;
	int		use;	/* Use count */
	int		order;	/* Ordering to restore list */
	vect2		smp1, smp2;	/* For long slit ordering */
#ifdef  NNA
	struct	objq	*nn;	/* Neighbor pointer */
	struct	objq	*nx;	/* Neighbor pointer */
#endif
#ifdef  OBJCHECK
	int		cflag;	/* Check bits flag */
#endif
//#ifdef  GISMO
	int		sfx;	/* Sub-field index */
//#endif
	} objq;

/* Slit edge for extension algorithm: */
typedef	struct	sleg_tag {
	struct	sleg_tag   *next, *last;
	objq	*objec; 	/* The object */
	objq	*confl;		/* The conflict */
	int	side;		/* A or B side, 0 or 1 index */
	int	dummy;		// for alignment
	double	free;		/* Free space (for sorting) */
	} sleg;

/* Status bits for stat[side] */
#define	Stat_Free  1
#define	Stat_Prov  2
#define	Stat_Mark  4

/*  ===  Observation Data  ===  */

//#ifdef  GSDATA
typedef	struct	gpos_tag {
	double	ra;	/* Right Ascension, in hours */
	double	dec;	/* Declination, in degrees */
	double	ep;	/* Epoch, years */
//	int	cn;	/* Catalog Number */
	unsigned long	cn;	/* Catalog Number */
	}  gpos;
//#endif

#define	NSrlim	0.00436	// 1/4 degree

/* Observation flags (bits in Obs.Flag) */
#define  OFxrd   1
#define  OFpb2   2
#define  OFifu   4
#define  OFsxt   8
#define  OFmoe  16
#define  OFexc  (1 <<  5)
#define  OFexe  (1 <<  6)
#ifdef  OBJNAM
#define  OFpon  (1 <<  7)
#endif
//#ifdef  GISMO
//#define  OFgis	(1 <<  8)
//#endif
//#define  OFpim	(1 << 8)
#define  OFinc  (1 << 9)
#define  OFagr  (1 << 10)
#define  OFral  (1 << 11)
//sunifdef < #ifdef  OCKWX
//sunifdef < #define  OFkw   (1 << 12)
//sunifdef < #endif
/* xrd = extra orders;  pb2 = passband 2 active;
  ifu = IFU mode on;  sxt = slit extend feature;
  moe = Multi-Object Echelle in use;
  exc = Spectral extent checked;
  exe = Spectral extent error status;
  pon = Plotting of object names;
  ral = Remove ALignments for detector/gap;
  gis = The gismo subfield splitter is in use; --UNUSED--
  pim = Position Independent mask -- see posind.otl (now reflim==0)
  inc = Inch scale for numbers on debug plot;
  agr = Accurate gismo rotation
  kw  = Keyword input in object list seen
  ...   */

/* Detector status flags (bits in Obs.Dstat) */
#define  DS_ns     1
#define  DS_gaps   2
#define  DS_orient 4

/* Observation structure -- the observing parameters */
typedef  struct  obs {

/* Identification */
	char	oname[16];	/* Observer name (required) */
	char	fname[8];	/* File name (required) */
	char	*title;		/* Observation title */
//	void	*dummy;		// for alignment
	void	*source;	/* Pointer to source program name/version */
// source will usually be a Vers structure

/* Celestial position and instrument orientation */
	double		ra, dec;	/* Field center (radians) */
	double		equinox;	/* Coordinate system */
	double		epoch;	/* Observation date in years and decimal */
	double		angle;	/* Instrument position angle (radians) */
	double		temp;	/* temperature expected */
	double		ha;	/* Hour angle expected */
//	vect3		mpos[3];	/* Position transform matrix */
	matrix3		mxpos;		/* Position transform matrix */

/* Guiders */
//#ifdef  GSDATA
	gpos		gp1;
	gpos		gp2;
	int		autogs;	/* Auto guide star feature flag */
	int		agscan;	/* Scan status, 1 = scanned */
//#endif

/* Wavelength space items */
	passband	pb;	/* Wavelength limits */
	passband	pb2;	/* Limits for on-detector */
	double		cw;	/* Center wavelength */
	double		sw;	/* Spectrum center (mean) wavelength */

/* Objects */
	namlist		*comments;	/* Global comments */
	namlist		*objfiles;	/* Object list files input */
	objq		*ob;	/* Object queue header */
//	void		*dummy2;	// for alignment
	int		consub;		/* Constellation subscript */
	int		reuseob;	/* Re-use of objects */
	int		reuserf;	/* Re-use of reference */
	int		refsel;		/* Reference selector */
	int		reflimit;	/* Reference count limit */
	double		Pmusthave;	/* Must Have priority */
	double		Pdecide;	/* Decision parameter priority */
	double		apri;		/* Alt. extended priority coefficient */

/* Mask space */
	element		*Tel_scope;
	void		*dummy3;	// for alignment
	slit		default_slit;
	slit		reference_slit;

/* Instrument */
	element		*Instr_mt;
	element		*Direct_inst;
// #ifdef  GISMO
	element		*Frontend;	// For GISMO
	void		*dummy3a;	// for alignment
// #endif

/* Disperser */
	element		*Disperser;
	int		order;		/* Disperser order */
	double		D_angle;	/* Disperser (grating) angle */
	passband	ddir;	/* Dispersion direction */
/* definition - multiply ddir.blue by slit.width and add to 
  object position for the blue edge of spectrum; same for red;
  value stored is +0.5 or -0.5 depending on dispersion direction
  relative to slit mask coordinate direction. */
/* actually, ddir is not really used yet... */

/* Detector */
	detector	*Detect;
	int		Dstat;		/* Detector and gap status */
	filter		*filter;
	int		dummy4;		// for alignment

/*  New optical references... */
	element		*elist;		// Element list pointer
	void		*dummy5;	// for alignment

/* Features */
	int		ex_order;	/* Extra order conflicts flag */
	int		dlevel;		/* Debug message level */
	double		minsep;		/* Minimum separation spec. */
	double		dca, dcb;	/* Do not cut l/r in arcsec */
	double		nrbias;		/* N&S rotation bias value, radians */
	int		pb2act;		/* Detector limit flag */
	int		IFUmode;	/* The IFU feature */
	double		IFUra, IFUdec;	/* Offset Center position */
	int		Flag;		/* MOE, later slext, exorder, etc. */
	int		slext;		/* Slit extension flag */
	int		dref;		/* Differential Refraction flag */
	namlist		*pcom;		/* Rotator flip, etc. comments */
	void		*dummy6;	// for alignment
	double		cref;		/* Refraction coeficient */
	vect2		zenith;		/* Zenith (radians) in mask xy */
	int		date;		/* Date as mjd for intgui */
	int		nsexx;		/* Number of objects with long spectra */
	int		actref;		/* Number of active reference objects */
//#ifdef  REJOLDATA
//	int		verif;		/* Verification of input */
//#endif
//	int		dummy7;		// for alignment
	} Obs;


/* We need the grating used, passband limit, overlap used, filter,
and perhaps default slit parameters to be recorded for observing
time and reference */

/* A slit mask data structure for output (and input) containing
 all the data in the SMDF */
typedef  struct  slitmask {
	char	observer[16];	/* Standard identification */
	char	maskfile[8];
	Obs	obs;		/* Observing parameters */
/* Obs structure contains field center, eq, ep, pa values for observ'n */
	namlist	*comments;
/* type or structure containing size limits, curvature, etc. */
	objq	*slitlist;	/* List of slit (object) data blocks */
/* When writing, write only active objects */
/* When reading, the objval stuff is NULL pointer */

	} slitmask;
// above is unused in programs (yet?)


/*  ====  */
#endif     /*  INCLUDE_MASKDAT_H  */
/*  ====  */
