/* msdebug.c == debug dump routines for mask generation programs */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **
**                                                                   **
**                   Copyright (C) 2010 by                           **
**     Ken Clardy, Carnegie Observatories, Pasadena California.      **
**                                                                   **
**      This software is proprietary, and may be used or copied      **
**          only with the written consent of the author.             **
**              All copies must retain this notice.                  **
**                                                                   **
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* All these debug dump routines are collected here, to be included
  when needed by such things as mask generation, mask cutting,
  smdfps utility and future utilities.  These are programs which are
  used in debug and development, but not in production versions.  */


/*  ====  Debug Subroutines  ====  */


static	void	dumpnamlist (namlist *h, char* p)
/* List entries in a name list, with prefix p; */
{
namlist	*r;
    C_loop (r, h) printf ("%s%s\n", p, r->name);
}

static	void	dumpcomments (namlist *h)
/* List all comments with leading ! character in queue */
{
    dumpnamlist (h, "!");
}


#if LocDebug > 0 && defined ThisAintHereAtAll
// The latter thing above effectively moves this to mgutils.c now.
void shownamlist (FILE* f, namlist* n, char* title, char* p)
/* Debug dump of namlist structure.  Optional title field and optional
  line prefix field p are printed.  Output to file f.  Allow null list.  */
{
namlist *r;
char	*x;
    if (title != NULL) fprintf (f, " --- %s ---", title);
    if (n == NULL) {
	fprintf (f, "  == [NULL].\n");
	return;
    } else  fprintf (f, "\n");
    x = (p == NULL) ? "" : p;
    C_loop (r, n) {
	if (r->name == NULL) fprintf (f, " %s[NULL]\n", x);
	else fprintf (f, " %s<%s>\n", x, r->name);
    }
//    fprintf (f, "--- End of list ---\n");
}
#endif  // on LocDebug


static  void	showslit (slit s, char* name)
/* Show the actual slit data... */
{
char	shape[16];
    switch (s.shape) {
	case CIRCLE:
	    strcpy (shape, "Circle");
	    break;
	case SQUARE:
	    strcpy (shape, "Square");
	    break;
	case RECTANGLE:
	    strcpy (shape, "Rectangle");
	    break;
	default:
	    strcpy (shape, "Unknown");
	    break;
    }
    if (s.width == 0.0 && s.alen == 0.0 && s.blen == 0.0 && s.angle == 0.0) {
/* Undefined slit parameters... */
	strcpy (shape, "Undefined");
	printf (" Standard %s slit is %s.\n", name, shape);
    } else {
	printf (" Standard %s slit is %s;\n", name, shape);
	printf ("  width %5.3f  lengths %5.3f %5.3f, angle %5.2f.\n",
	    s.width,  s.alen, s.blen,  s.angle/Degree);
    }
}

int	nobjects (Obs* a)
/* Count objects in Obs structure */
/* REPLACE WITH listcount REFERENCE */
{
    return  listcount (a->ob);
}

int	nrefholes (Obs* a)
/* Count reference holes in Obs structure */
{
objq	*r;
int	n = 0;
    if (a == NULL) return 0;
    C_loop (r, a->ob) if (r->type == OBJ_REFERENCE) n++;
    return  n;
}

static	void	dumpobjq (objq *a)
/* Dump the objq entry in file format */
{
char	t;
objq	*r;
char	abuf[32];
char	dbuf[32];
    if (a == NULL) return;
    C_loop (r, a) {
	dumpcomments (r->dat.precom);
	t = r->type == OBJ_OBJECT ? '@' : '*';

	printf ("%c%s", t, r->dat.name);
/* Add print of sexigisimal ra and dec here */
	strcpy (abuf, sexig (r->dat.ra/Hour, 3, 3));
	strcpy (dbuf, sexig (r->dat.dec/Degree, 3, 2));
/* Above is how to do it; the sexig program returns a local buffer */
	printf (" %s %s", abuf, dbuf);
// works when printed separately; same buffer used in sexig
/* Print the mask coordinates */
	printf (" x %7.2f  y %7.2f", r->smpos.x, r->smpos.y);
	printf (" %5.2f %2d %02X", r->dat.priority, r->type, r->flag);
	printf ("\n");
	dumpcomments (r->dat.postcom);
	printf ("  - - -\n");  /* debug */
// experiment change order of these two...
    }
}


void	show_objqd (objq *a)
/* Show object queue and detector position (checking sort) */
{
objq	*r;
    if (a == NULL) return;
//    printf ("\n Object Queue   Bx      By     X     Y\n");
    printf ("\n Object Queue      X        Y");
    printf ("       Sx       Sy\n");
    C_loop (r, a) {
//	printf (" %-10s  %7.4f %7.4f %8.3f %8.3f\n", 
//		r->dat.name, r->val->B.x, r->val->B.y, r->smpos.x, r->smpos.y);
        printf (" %-10s  %8.3f %8.3f",
                r->dat.name, r->smpos.x, r->smpos.y);
        printf (" %8.3f %8.3f\n",
                r->val->sod->center.x, r->val->sod->center.y);
    }
    putchar ('\n');
}


void	dumpobs (Obs *a)
/* Dump an observation structure */
{
    if (a == NULL) return;
    printf ("\n  --- Observation structure dump --\n");

    printf (" Observer: %s   Filename: %s\n", a->oname, a->fname);
/*    printf (" Title: %s  Object: %s\n", a->title, a->object); */
    if (a->title != NULL) printf (" Title: %s \n", a->title);
    printf (" Field center:  %s", sexig (a->ra/Hour, 3, 2));
	printf (" %s", sexig (a->dec/Degree, 3, 1));
	printf (" (%6.1f)\n", a->equinox);
// fflush (stdout);	// debug
    printf (" Date %8.3f, Position angle %5.1f, temperature %4.1f\n",
	a->epoch, a->angle/Degree, a->temp);
    printf (" Wavelength %6.1f, limits %6.1f - %6.1f\n",
	a->cw, a->pb.blue, a->pb.red);
/* NOTE - need to add optic elements, detector, grating, filter, etc */
//    dumpoptic (a->Telescope, "Telescope");
//    dumpoptic (a->Collimator, "Collimator");
//    dumpoptic (a->Camera, "Camera");
//    dumpgrating (a->grating);
//    dumpdispers (a);
// If needed/desired, use a dump element code on the element members.
// See the optdbug.c for example, the "display_element" program.
    showslit (a->default_slit, "default");
    showslit (a->reference_slit, "reference");
    printf (" Object list files:\n");
    dumpnamlist (a->objfiles, "  -- ");
    printf (" Comments:\n");
    dumpnamlist (a->comments, "  !");
    putchar ('\n');
/* Dump the object queue too (only if dodebug enhanced) */
    if (dodebug > 1) {
	printf ("\n- - Object queue - -\n");
	dumpobjq (a->ob);
    }
    printf ("\n");
}

/* --- Test queues for consistency --- */
// void	check_elemqueue (element* head, char* title)
/* --- Used in maskgen only --- */



/*  ====  End Debug Subroutines  ====  */

/*  =*=*=*=*=*=  =*=*=*=*=*=  =*=*=*=*=*=  =*=*=*=*=*=  =*=*=*=*=*=  */
/*  =*=*=*=*=*=     SEPARATOR -- OLD VERSIONS BELOW     =*=*=*=*=*=  */
/*  =*=*=*=*=*=  =*=*=*=*=*=  =*=*=*=*=*=  =*=*=*=*=*=  =*=*=*=*=*=  */

/*  ===== ===== ===== ===== ===== ===== ===== ===== ===== =====  */
/*  ===== ===== ===== ===== ===== ===== ===== ===== ===== =====  */
/*  ===== ===== ===== ===== ===== ===== ===== ===== ===== =====  */
/*  ===== ===== ===== ===== ===== ===== ===== ===== ===== =====  */
/*  ===== ===== ===== ===== ===== ===== ===== ===== ===== =====  */
