/* SMDF plot utility - Read an SMDF file into an observation
  structure as defined in maskdat.h - and interactively plot the
  map as is done in maskgen.  */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **
**                                                                   **
**                   Copyright (C) 2010 by                           **
**     Ken Clardy, Carnegie Observatories, Pasadena California.      **
**                                                                   **
**      This software is proprietary, and may be used or copied      **
**          only with the written consent of the author.             **
**              All copies must retain this notice.                  **
**                                                                   **
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* The SMDF file name is the argument in the command line.  If not
  present, it is prompted for.  Multiple file names may be present.
  For each file, an interactive plot is generated.  */

/* The symbol  MASKONLY  is used here.  It is defined for the include
  of the smplot.c code to include that code compiled in such a way as
  to ony plot the mask and not the detector plane.   */

/*  ====  STANDARD INCLUDES  ====  */
#include	<stdio.h>
#include	<stdlib.h>
#include        <string.h>
#include	<unistd.h>	/* for access */

/*  ====  GLOBAL DEFINITIONS  ====  */

//const	char	PROGNAME[] = "SMDF Plot utility";
//const	char	VERSION[]  =  "0.80";
// int     dodebug = 0;	/* Global debug flag */
// dodebug is defined in optutils.h which is included with mgutils.h

#include	"mgutils.h"
const	char	VERSION[]  =  Xstr(SMFPLTV);

typedef struct  {
	int	dbug;
	int	pwin;
	int	pnam;
	int	inch;
	Pvfs	*cp;
	Pvfs	*fp;
	}  opts;

// Standard Version structure
const	Vers	Program = { NULL, NULL,  SMFPLTV, SMFPLTR, SMFPLTT };
#define  APPLICATION_NAME  "SMDF Plot utility"
const	char	PROGNAME[] = SMFPLTT;
// mgutils expect PROGNAME and VERSION strings to exist as global labels

/*  ====  Utility programs, library programs, debug code.  ====  */

//#include	"smplot.c"
// smplot.c is window plot of mask -- obsolete

// #include	"smdfutil.c"
// smdfutil stuff is now in mgutils.c
/* The routines being tested; they will also be included
  by all the programs which use the SMDF file format */


/*  ===  Plotting Programs  ===  */

#define  MASKONLY
#include  "dbplot.c"
/* The above two lines will locally compile a version of dbplot which
  shows only the mask data rather than the more general plot which
  has a mask/detector plane switch button.  Only the mask plane is
  active when the  MASKONLY  symbol is defined.  */

/*  ===  Debug Subroutines  ===  */

#include	"msdebug.c"
// Standard debug and dump routines
// Must follow plotting program include.


/*  ===  Subject test routines  ===  */


int	testsmd (char* fn, opts opt)
{
Obs	*obs;	/* The observation structure */
int	n, m, k;

/* Check for and handle any null input cases */
    if (fn == NULL) return 0;
    if (*fn == '\0') return 0;

/* Call to read the file */
//    obs = read_smdf (fn);
    obs = read_smdf (fn, NULL, 0);
    if (obs == NULL) {
	fprintf (stderr, " Found no observations in %s.\n", fn);
	return  0;
    }
    if (opt.inch) obs->Flag |= OFinc;

/* Check/warn if obs structure is wrong version */
    k = comp_vers ((Vers*)obs->source, opt.fp);
    if (k < -2) {
	printf (" ** Warning!  File made by UNKNOWN software!\n");
    } else if (k != 0) {
	printf (" ** Warning!  File made by %s version %s is ",
		((Vers*)obs->source)->tag, fvstr(((Vers*)obs->source)->v) );
   	switch (k) {
	    case -2: printf ("Obsolete");		break;
	    case -1: printf ("Older");			break;
	    case  0: printf ("Current");		break;
	    case  1: printf ("advanced (beta)");	break;
	    default: printf ("unknown");		break;
	}
	printf (" software.\n");
    }

/* Call dump routine to dump file */
    if (dodebug) dumpobs (obs);

#ifdef  OBJNAM
	if (opt.pnam) obs->Flag |= OFpon;
//	printf (" Plot object names is %s.\n", opt.pnam ? "on" : "off"); //debug
#endif

//    if (dodebug != 1) obs->dlevel = dodebug;
// dodebug default 1, -d 0, +d 2 ==> dlevel 0, 1, 2
//    obs->dlevel = (dodebug >= 2) ? dodebug : 1 - dodebug;  // temporary debug
// Done because the -d flag sets dlevel to 0 and no flag sets it to 1.

/* obs->dlevel got set to dodebug during read_smdf  so that check_source
  could be debug enabled.  Can also be set when obs is created.  */

    n = nobjects (obs);
    m = nrefholes (obs);
    k = n - m;
    printf (" Found %d slits and %d reference holes.\n", k, m);
    printf (" Total of %d objects.\n", n);

/* Test the plotting */
    if (n > 0) pwin (obs, MASK);

/* Clean out storage of obs structure and obj queue */
    kill_obs (obs);
    return  1;

}


	/* - - - - - - - - *\
	|    Main Program   |
	\* - - - - - - - - */

/*  ====  MAIN PROGRAM  ====  */

#undef  PROMPTLOOP

int  main (int argc, char** argv) {
char    *pp;    /* Scan variable */
char    px;     /* Scan character */
int     i;
int     dohelp = 0;
char	infile [120];	/* For prompting */
opts	opt;
namlist	*head = NULL;
char	*filename;
int	k, m, n;	/* Process counts */
namlist	*comps=NULL;

/* Get parameter as name of SMDF file to read */
/* Usual flags, debug and help */

/* Standard input logic:  If argument given, take that argument as the
  sole specification to process.  If no argument is given, prompt for
  an argument, and repeat the prompting until a null reply.  */

    opt.dbug = 0;	/* Or - put data in define above */
    opt.pwin = 0;
    opt.pnam = 0;	/* plot names */
    opt.inch = 0;	/* Inch mode on numbers */
    opt.cp = NULL;
    opt.fp = NULL;
    dodebug = 0;	/* NOTE = temporary for debugging */

/* Scan for switches and arguments */
    for (i=1; i<argc; i++) {    /* argv[0] is command name */
        pp = argv[i];
        if (*pp == '-') {       /* Scan normal switch */
            while ((px = *++pp) != '\0') {
                if (px >= 'A' && px <= 'Z') px += 'a' - 'A';
                if      (px == 'd')  dodebug = 1;  /* Debug output */
/* May want -d to cause debugging later? */
                else if (px == 'h')  dohelp = 1;  /* show help */
                else if (px == 'i')  opt.inch = 1;  /* Inch mode plot */
#ifdef  OBJNAM
                else if (px == 'n')  opt.pnam = 1;  /* Plot names */
#endif
	    }
	} else if (*pp == '+') {        /* Scan positive switch */
            while ((px = *++pp) != '\0') {
                if (px >= 'A' && px <= 'Z') px += 'a' - 'A';
                if      (px == 'd')  dodebug = 2;      /* Debug output */
                else if (px == 'h')  dohelp = 1;  /* show help */
            }
        } else if (*pp == '?') {
                dohelp = 1;
        } else {
#ifdef debug
            printf ("Adding %s to list.\n", pp);  /* debug */
#endif
	    head = pushtxt (pp, head);	/* Collect arguments */
	    printf ("Argument %s.\n", pp);  /* debug */
        }
    }    

    if (dohelp) {
	printf ("\n %s -- Version %s", APPLICATION_NAME, VERSION);
	printf ("\n  Parameters are: ");
	printf ("\n    SMDF File name ");
	printf ("\n   Switches");
	printf ("\n     -d = Some debug information");
	printf ("\n     +d = More debug information");
#ifdef  OBJNAM
	printf ("\n     -n = Plot object names");
#endif
	printf ("\n     -i = Display scale numbers in inches");
	printf ("\n     -h = Display this help");
	printf ("\n   Plots mask interactively in a window.");
/*	printf ("\n   Dump(s) written to standard output."); */
/*	printf ("\n   Output file name is prompted for."); */
	printf ("\n \n");
//	return (1);
    }

    if (dodebug && !dohelp)
	printf ("\n %s -- Version %s\n", APPLICATION_NAME, VERSION);

/* See if this version is any good, or if we continue.  */
//    comps = pushtxt (MASKCTT, comps);
    comps = pushtxt (SMFPLTT, comps);
    k = check_ver (comps, &(opt.cp), &(opt.fp));
    comps = kill_name (comps);
// if feature exists, we could exit on k < -2 here.
    if (dohelp) return (1);
    if (k <= -2) {
	printf (" -- Obsolete software will not run.\n");
	return (-1);
    }

/*  -- Loop to process arguments, or prompt if no arguments -- */

    for (k=n=m=0; k >= 0; ) {	// Forever loop
	if (head == NULL) {	// No arguments, or looped out
#ifdef  PROMPTLOOP
	    if (m > 0) break;	// Had real arguments, looped out
#else
	    if (n > 0) break;	// No loop after any counted argument
#endif
	    askmestr (infile, 110, "SMDF file:");
	    if (*infile == '\0') break;	// Exit on any null reply
	    filename = dynamstr (infile);
	} else {		// We have a command line argument given
	    filename = pop_name (&head);
	    if ( filename == NULL) continue;
	    if (*filename == '\0') continue;
	    m++;	// Count number of command line arguments given
	}
	n++;		// Count number of arguments requested
	k += testsmd (filename, opt);	// Count number processed
	free (filename);	// MUST be dynamic
    }

/* All done, write a message about how it came out... */
    printf ("\n Processed %d of %d files.\n", k, n);
    opt.cp = kill_pvfs (opt.cp);
    opt.fp = kill_pvfs (opt.fp);
    return  0;
}

/* Some change notes

    08/03-13 - Version 0.83
	Add test of version for input data, and message

    08/02-27 - Version 0.82
	Add inch mode support for -i flag
	Supporting wfccd instrument in dbplot.c now
	Refuse execution if version is obsolete
	Minimum Version for release 2.03

    07/10-23 - Version 0.81
	Add verson control support.
	Minimum Version for release 2.00 and version control.

    07/09-28 - Version 0.80
	New utility called smdfplt; includes dbplot.c with the MASKONLY
	symbol defined to enable use of dbplot code.  Other uses e.g.
	maskgen, use dbplot.c without MASKONLY to get mask/detector
	selection active; they also have dbplot compiled independently
	to dbplot.o.  Special entry in makefile needed to link this
	properly without dbplot.o but including utils.o etc.

    03/10-01 - Version 0.75
	General update, newer plotting routine

    02/04-30 - Version 0.62
	Change smplot to use matrix method for inverse coordinates

    02/03-21 - Veraion 0.60
	Insert standard main program logic,
	standard program name and version globals.

    02/03-18 - Version 0.57
	Add count of objects and reference holes

.. */
