/* Utility programs for performance timing */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **
**                                                                   **
**                  Copyright (C) 2009 by                            **
**     Ken Clardy,  Carnegie Observatories,  Pasadena California.    **
**                                                                   **
**      This software is proprietary, and may be used or copied      **
**          only with the written consent of the author.             **
**              All copies must retain this notice.                  **
**                                                                   **
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* ---------------------------------------------------------------------- */

#include	<stdlib.h>	// malloc, free
#include	<stdio.h>	// sprintf
#include	"timepak.h"

// Define subroutines called from here...
// extern	char*	dynamstr (char* a);  // use strdup instead

/* ---------------------------------------------------------------------- */

/*  The 2009 extension adds the following features:
  > Elapsed absolute time in addition to CPU time
  > Better report formatting codes on bit-based names
  > The "auto" flag does auto start and auto destroy
  > T_new adds a flag to call sequence, bit mapped
  > T_rep stops timer if running, report flag now bit mapped
  > T_rep takes **T_timer argument, to allow timer to be killed
  > T_xit does both report and print
  */

/* ---------------------------------------------------------------------- */

// T_timer* T_new (title, flags)
// void T_clr (T_timer a)
// beg end rep (begin, end, report)
//   twhole = T_new ("Overall");
//   tscan = T_new ("Scanning");
//  T_clr (twhole);
//  T_clr (tscan);
//  T_beg (tscan);
//  ...
//  T_end (tscan);
//  tta = T_new ("Whole system:", TF_auto);
//  bb = T_rep (tta, TR_titlr|TR_elap|TR_cput|TR_count|TR_long);

static	double	ATime (void)
/* Return absolute time in seconds */
{
struct	timeval	t;
double	r;
    gettimeofday (&t, NULL);
    r = (double)t.tv_sec + (double)t.tv_usec/1.e6;
    return  r;
}

T_timer*  T_new (char* title, int flag) {
/* Create a timer block, clear it and return pointer */
T_timer	*b;
    b = (T_timer*) calloc (1, sizeof (T_timer) );
    b->title = strdup (title);
//    T_clr (b);
    b->flag = flag;
    if (flag & TF_auto) T_beg (b);
    return  b;
}


T_timer*  T_kil (T_timer* b) {
/* Remove title, return storage, return null */
    if (b == NULL) return b;
    if (b->title != NULL) free (b->title);
    b->title = NULL;
    T_clr (b);
    free (b);
    return NULL;
}

void	T_clr (T_timer* b) {
/* Clear count and status fields in block (not title) */
    if (b == NULL) return;
    b->et    = 0.0;
    b->ct    = 0.0;
    b->abst  = 0.0;
    b->start = 0;
    b->n     = 0;
    b->state = 0;
    b->flag  = 0;
    b->er    = 0;
}

void	T_beg (T_timer* b) {
/* Start timing */

// Check for null case
    if (b == NULL) return;

// See if already running (an error)
    if (b->state > 0) {
	(b->er)++;
	if (b->flag & (TF_rest | TF_auto)) {
	    T_end (b);
	} else {
	    b->state = 0;
	}
    }

// Start the timer and set state
    b->abst = ATime();
    b->start = clock();
    (b->state)++;
}

void	T_end (T_timer* b) {
/* Stop timing */
clock_t	e;
double	t;

    if (b == NULL) return;

// See if running (error if not)
    if (b->state <= 0) {
	(b->er)++;
	b->state = 0;
	return;
    }

// Find the elapsed time and total it
    e = clock() - b->start;
    t = ATime() - b->abst;
    b->et += (double)e / (double)CLOCKS_PER_SEC;
    b->ct += t;
    (b->n)++;
    (b->state)--;
}

/*  ====  Report section  ====  */

static int add_time (char* buffer, const int ml, int tl, char* lab, 
		double tx, int nt)
/* Utility to add to buffer the time tx with label lab, and count nt;
  return the updated tl value, respecting max line ml.  */
{
char	line[128];
int	j;
	j = sprintf (line, " %s = %6.3f", lab, tx);
	if (tl+j > ml) {strcat (buffer, "\n");  tl = 0;}
	strcat (buffer, line);
	tl += j;
	if (nt > 1) {
	    j = sprintf (line, " (%d @ %6.3f ms)", nt, 1.e3*tx/nt );
	    if (tl+j > ml) {strcat (buffer, "\n"); tl = 0;}
	    strcat (buffer, line);
	    tl += j;
	}
	return  tl;
}

char*	T_rep (T_timer* *b, int r) {
/* Return dynamic buffer with formatted report */
//char	*buf;
//char	xbuf[32];
char	buffer[256];
char	line[128];
char	*tx;
int	tl=0;	// length of current line
int	ml;	// maximum line requested
int	j;	// scratch

    if (b == NULL) return  NULL;
    if (*b == NULL) return  NULL;

/* If timer is auto, we stop it if needed.  */
    if (((*b)->flag & TF_auto) && ((*b)->state > 0)) T_end (*b);
    if ((*b)->n < 1) {	// Null case, no time registered
	if ((*b)->flag & TF_auto) *b = T_kil (*b);
	return  NULL;
    }

    *buffer = '\0';
    ml = (r & TR_long) ? 80 : 60;

/* Put in title if requested */
    if (r & TR_titlr) {
	tx = ((*b)->title == NULL) ? " " : (*b)->title;
	tl = sprintf (buffer, "%s", tx);
    }

/* Add line feed if requested */
    if ((r & TR_titlf) || (tl > ml)) {
	strcat (buffer, "\n");
	tl = 0;
    }

/* Report elapsed time if requested */
    if (r & TR_elap) {
	tl = add_time (buffer, ml, tl, "ET", (*b)->ct,
		(r & TR_count) ? (*b)->n : 0 );
    }

/* Report CPU time if requested */
    if (r & TR_cput) {
	tl = add_time (buffer, ml, tl, "CPU", (*b)->et,
		(r & TR_count) ? (*b)->n : 0 );
    }

/* Add count if not otherwise output;  DEBUG feature only */
//    if ((r & (TR_elap | TR_cput)) && TR_count == 0 ) {
    if ( (r & (TR_elap | TR_cput)) && (*b)->n < 2 ) {
	j = sprintf (line, " (%d)", (*b)->n);
	strcat (buffer, line);
	tl += j;
    }

/* Error report if requested */
    if (r & TR_errp) {
	j = sprintf (line, " (%d errors)", (*b)->er);
	if (tl+j > ml) {strcat (buffer, "\n");  tl = 0;}
	strcat (buffer, line);
	tl += j;
    }

/* If the timer is auto, we destroy it after writing. */
    if ((*b)->flag & TF_auto) *b = T_kil (*b);

/* Return a copy of the buffer */
    if (tl > 0) strcat (buffer, "\n");
    if (strlen(buffer) < 1) return  NULL;
    return  strdup (buffer);

/* The old default report form had features
  TR_titlr + TR_cput + TR_count + TR_errp + TR_long  
  (put those in the Mark II report call.  */
}

void	T_xit (T_timer* *b, int r)
/*  Report and print on stderr, and clean the resulting output
  character buffer.  If b is auto, it is also destroyed.  */
{
char	*c;
//int	f;
    if (b == NULL) return;
    if (*b == NULL) return;
//    fprintf (stderr, " ** Ending %s\n", (*b)->title);	// DEBUG only
//    fflush (stderr);				// DEBUG only
    if ( (*b)->state > 0) T_end(*b);	// does not need auto flag here
//    f = (*b)->flag;
    c = T_rep (b, r);		// Form the report (kill *b if auto)
//    if (f & TF_auto) *b = NULL;
    if (c == NULL) return;
    if (strlen (c) > 0) {
	fprintf (stderr, " %s", c);	// Print to stderr
	fflush (stderr);
    }
    free (c);			// Clean up the report
}

/* ---------------------------------------------------------------------- */

/*   ----    The Mark II system    ----   */

// Usage:
//	TX_con (20)
//	TX_beg (12);
//	TX_enm (12, "subject detail");
//	TX_rep (-1);

void	TX_con (int n) {
/* Control.  If n > 0, establish a set of n timers.
  If n == 0, remove the current timers.  */
T_timer	*t;
int	i;

/* Remove any existing timers before making any more */
    if (T_TX != NULL) {
	for (i=0; i<N_TX; i++) {
	    t = &(T_TX[i]);
	    if (t->title != NULL) free (t->title);
	    t->title = NULL;
	}
	free (T_TX);
    }
    T_TX = NULL;
    N_TX = 0;

/* If we have n > 0, establish n timers, all cleared. */
    if (n > 0) {		// Establish timers
	T_TX = (T_timer*) calloc (n, sizeof (T_timer) );
	N_TX = n;
	for (i=0; i<N_TX; i++) {
	    t = &(T_TX[i]);
//	    t->title = NULL;
//	    T_clr (t);
	    t->flag = TF_rest;
	}
    }
}

static	int	TX_ndx (int n) {
/* Common null case and index code */
    if (T_TX == NULL) return -1;
    if (N_TX < 1) return -1;
    if (n > N_TX) return -1;
    if (n < 0) return -1;
    return (n-1);
}

void	TX_beg (int n) {
/* Start timing for timer n.  */
T_timer	*t;
int	i;

    if ( (i = TX_ndx(n)) < 0) return;	// Null cases and index
    t = &(T_TX[i]);
    T_beg (t);
}

void	TX_end (int n) {
/* Stop timing for timer n.  */
T_timer	*t;
int	i;

    if ( (i = TX_ndx(n)) < 0) return;	// Null cases and index
    t = &(T_TX[i]);
    T_end (t);
}

void	TX_nam (int n, char* title) {
/* Apply given title to timer n. */
T_timer	*t;
int	i;

    if ( (i = TX_ndx(n)) < 0) return;	// Null cases and index
    if (title == NULL) return;
    t = &(T_TX[i]);
    if (t->title == NULL)   t->title = strdup (title);
}

void	TX_enm (int n, char* title) {
/* Both end and name - useful */
    TX_end (n);
    TX_nam (n, title);
}

void	TX_clr (int n) {
/* Clear timer n */
T_timer	*t;
int	i;

    if ( (i = TX_ndx(n)) < 0) return;	// Null cases and index

/* Clear the specified timer */
    t = &(T_TX[i]);
    T_clr (t);
}

void	TX_rep (int n) {
/* Report.  If n > 0, report timer n.
  If n == 0, report all timers.
  If n < 0, report all timers and remove them.  */
T_timer	*t;
char	*c;
int	i;
int	j, k;

/* Bypass any request for null data report */
    if (T_TX == NULL) return;
    if (N_TX < 1) return;

/* Establish report range, j to k */
    if (n > 0) {
	k = n;
	j = k - 1;
    } else {
	j = 0;
	k = N_TX;
    }

/* Report range of nodes */
    for (i=j; i < k; i++) {
	t = &(T_TX[i]);
	if (t->title != NULL || t->n > 0) {
//	    c = T_rep (t, 1);
	    c = T_rep (&t,
	      TR_titlr + TR_cput + TR_count + TR_errp + TR_long  );
	    printf (" [%2d] %s\n", i+1, c);
	    free (c);
	}
    }

/* If requested, kill the data set */
    if (n < 0) TX_con (0);

}



/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
