/*  versup.c == Version Support utility  */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **
**                                                                   **
**                   Copyright (C) 2010 by                           **
**     Ken Clardy, Carnegie Observatories, Pasadena California.      **
**                                                                   **
**      This software is proprietary, and may be used or copied      **
**          only with the written consent of the author.             **
**              All copies must retain this notice.                  **
**                                                                   **
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* Provides support for version control and verification as a stand-alone
  utility.  Will create and verify the version control file maintained
  in the FTP area.   */

/*  === Common includes === */
#include	<stdio.h>
#include	<string.h>
#include	<stdlib.h>

#include	<sys/types.h>	// for stat
#include	<sys/stat.h>	// for stat
#include	<unistd.h>	// for access, stat
#include	<time.h>

/*  ====  GLOBAL DEFINITIONS  ====  */

const	char	PROGNAME[] = "Version Control utility";
/* Version information in mgfeats.h */
/* mgutils seems to need text version still */

#include	MGUTILH
const	char	VERSION[] = Xstr(VERSUPV);  // temporarily needed in mgutils

// Standard Version structure
const	Vers	Program = { NULL, NULL,  VERSUPV, VERSUPR, VERSUPT };
//const	char	PROGNAME[] = VERSUPT;
// mgutils expect PROGNAME and VERSION strings to exist as global labels

typedef struct  {
	int	dbug;	// print level
	int	tfil;	// Compare to FTP rather than compiled
	int	rfil;	// Source is file rather than compiled
	int	cfil;	// Compare to file rather than ftp/compiled
	int	self;	// Do a self test on source
	int	ncmp;	// Omit the compare to second data
	int	dlcf;	// Delete local ftp data (only)
	int	wvfd;	// Write compiled data to a file
//	int	edit;	// Edit mode, read and write
	}  opts;


/*  ===  Utility Programs  ===  */


/*  ==  Actual application subroutines  ==  */


	/* - - - - - - - - - - - - - - *\
	|				|
	|    Reporting  Subroutines	|
	|				|
	\* - - - - - - - - - - - - - - */



char	*avstr (char* vst)
/*  Add spaces to left of vst to put the '.' into position
  3 in the string.  Return a new dynamic string, and free
  storage provided as vst.  */
{
char	*b;
//char	spx[4];
int	d;
    d = 3 - (strchr(vst, '.') - vst);  // add this many characters
    b = malloc (strlen(vst) + d + 2);
    b[d] = '\0';
    while (d > 0) b[--d] = ' ';
    strcat (b, vst);
    free (vst);
    return  b;
}


int	hasfeat2 (char* name, namlist* list, namlist* head)
/*  Search the given namlist for presence of name, return 1 if
  found, and 0 if not found. 
  Special 2 == loop stops if head is next.  Used to find duplicates.  */
{
int	r = 0;
namlist	*p;
//    C_loop (p, list) {
    if (list == head) return  r;
    for (p=list; p!=NULL; p=(p->next!=head)?p->next:NULL) {
	if (strcmp(name, p->name)) continue;
	r = 1;
	break;
    }
    return  r;
}


int	dispvfs (Pvfs* pp)
/*  Display on stdout a report of the content of the given Pvfs
  structure.  Show date field, all versions, and list features.  */
/*  Return count of errors found */
// was dispver
{
Vers	*v;
namlist	*n;
int	k=0;
float	s;
//int	nc = 0;

    if (pp == NULL) return  k;
    printf (" From: %s ", pp->name);
    printf ("   Date:  %s\n", pp->date);
    printf (" --Component--   Version   Min. Version\n");
    C_loop (v, pp->ver) {
char	*vs, *ms, *cs;
	vs = avstr(fvstr(v->v));
	ms = avstr(fvstr(v->m));
	printf ("   %-8s\t%-8s   %s", v->tag, vs, ms);
	if (strcmp(v->tag, VCONCVT) == 0) {
	    s = vschecks(pp->ver);
	    if (fvcmp(v->v, s)) {
		cs = fvstr(s);
		printf ("\t <-- %s\n", cs);
		free (cs);
	    } else  printf ("\n");
	} else if (fvcmp (v->v, v->m) < 0) {
	    printf ("\t <-- ** Version error\n");
	    k++;
	} else  printf ("\n");
	free (vs);
	free (ms);
    }

    printf ("\t\t  ---  Total of %d features.  ---\n", listcount(pp->feat));
    displist (pp->feat);

/* Check for duplicate features... */
    C_loop (n, pp->feat) {
	if (hasfeat2 (n->name, n->next, pp->feat)) {
	    printf (" ** Warning: duplicate feature:  %s\n", n->name);
	    k++;
	}
    }
    return  k;
}


int	showdif (Vers* x, Vers* r)
/* Report on a line how versions in x compare to refernce versions
  in r.  Handle null cases correctly.  */
/*  Return the error count  */
{
int	k=0;
char	*vx;
char	*vr;
char	*mr;
char	*stat;
const	char	xxx[] = "  ---";
const	char	format[] = "   %-8s\t%-8s  %-8s  %-8s\t%s\n";
// This keeps all 3 kinds of line in sync.  The version should be
// 8 or fewer characters of form xxx.xxxx with first and last two
// possibly or likely blank.

    if (x == NULL && r == NULL) return k;  // nothing to compare
    else if (r == NULL) {
	vx = avstr(fvstr(x->v));
	printf (format, x->tag, vx, xxx, xxx, "Unknown");
	k++;
	free (vx);
    } else if (x == NULL) {
	vr = avstr(fvstr(r->v));
	mr = avstr(fvstr(r->m));
	printf (format, r->tag, xxx, vr, mr, "Missing");
	k++;
	free (vr);
	free (mr);
    } else {
	vx = avstr(fvstr(x->v));
	vr = avstr(fvstr(r->v));
	mr = avstr(fvstr(r->m));
	switch (vvcmp (x, r)) {
	    case  -2:	stat = "Obsolete"; k++;		break;
	    case  -1:	stat = "Older";			break;
	    case  0:	stat = "Current";		break;
	    case  1:	stat = "Beta";			break;
	    default:	stat = "Other";	   k++;		break;
	}
	printf (format, r->tag, vx, vr, mr, stat);
	free (vx);
	free (vr);
	free (mr);
    }
    return  k;
}


int	compvfs (Pvfs* p, Pvfs* c)
/*  Report the comparison of versions p to reference versions c, which
  must be different.  Will find components in c which are not present
  in p in addition to those reported from p not in c.  Find differences
  in the set of features also.   */
/*  Return a count of errors found during compare operation */
// was compver
{
Vers	*pv;
Vers	*cv;
namlist	*dn = NULL;
namlist	*x;
int	n;
int	k = 0;

/*  Check and bypass null cases  */
    if (p == NULL) return k;
    if (c == NULL) return k;
    if (p == c) return k;

/*  Write heading  */
    printf ("\n");
    printf ("  Data from: %20s  Date:  %s\n",        p->name, p->date);
    printf ("  Reference data from: %15s  ( %s )\n", c->name, c->date);
    printf (" --Component--  Version   Ref. Ver.   Min.      Status\n");

/* Loop to check version */
    C_loop (pv, p->ver) {
	if (strcmp(pv->tag, VCONCVT) == 0) continue;	// skip cs
	cv = findver (pv->tag, c);
	k += showdif (pv, cv);
    }

/* Look for missing versions from p which are in c */
    C_loop (cv, c->ver) {
	pv = findver (cv->tag, p);
	if (pv) continue;
	k += showdif (pv, cv);
    }

/* Find differences in features.  */
    C_loop (x, p->feat) {	// Look at all features in p
	if (hasfeat(x->name, c->feat) == 0)
	    dn = pushtxt (x->name, dn);	// Feature in p not in c
    }
    n = listcount (dn);
    k += n;
    if (n) {
	printf (" Found %2d extra features not in reference:\n", n);
	displist (dn);
	dn = kill_name (dn);
    }

    C_loop (x, c->feat) {	// Look at all features in c
	if (hasfeat(x->name, p->feat) == 0)
	    dn = pushtxt (x->name, dn);	// Feature in c not in p
    }
    n = listcount (dn);
    k += n;
    if (n) {
	printf (" Missing %2d features present in reference:\n", n);
	displist (dn);
	dn = kill_name (dn);
    }
    return  k;
}




	/* - - - - - - - - - - - - - - *\
	|				|
	|	Test Subroutines	|
	|				|
	\* - - - - - - - - - - - - - - */


/*  ===  test subroutine here  ===  */

int	selftest (Pvfs* p)
/*  Do standard self test(s) on data in structure p, and report the
  error count in the return.  */
{
Vers	*v;
char	*vs, *ms;
int	k=0;
namlist	*nl;

/* Check for version mis-match in each version set. */
    C_loop (v, p->ver) {
	if (strcmp(v->tag, VCONCVT) == 0) continue;
	if (fvcmp(v->v, v->m) < 0) {
	    vs = fvstr (v->v);
	    ms = fvstr (v->m);
	    printf (" ** Error for %s version %s minimum %s\n", v->tag, vs, ms);
	    k++;
	    free (vs);  free (ms);
	}
    }

/* Check for duplicate features */
    C_loop (nl, p->feat) {
	if (hasfeat2 (nl->name, nl->next, p->feat)) {
	    printf (" ** Warning: duplicate feature:  %s\n", nl->name);
	    k++;
	}
    }

    if (k > 0) printf ("  %d errors found in %s.\n", k, p->name);

    return  k;
}


int	verfile (char* filename, opts opt)
/* Process the options.
  If filename is NULL, use VERSFIL as a default.
  If opt.dlcf, delete the local ftp data only and return 0.
  If opt.wfvd, current data will be written to filename.
  If opt.rfil, read from file; otherwise input is current data.
  If opt.cfil, compare to file unless doing opt.rfil
  If opt.tfil, use FTP in preference to current data.
  Return error count.					.. */
/* Write a report on the input unless the -v or -z flags are
  in use; do compare to other file unless -c is in use.  */
{
char	*source;
char	*buffr;
Pvfs	*pp = NULL;
Pvfs	*cv = NULL;
Pvfs	*fp = NULL;
Pvfs	*ap, *bp;
int	k = 0;
int	sf = 0;  // source file given


/* Take care of some immediate options */
    if (opt.dlcf) {	// Delete any local file (only)
char	*lf;
/*  Form the local file name for the file to be deleted */
	lf = malloc (strlen(VERSFIL) + strlen(LOCALpfx) + 2);
	sprintf (lf, "%s%s", LOCALpfx, VERSFIL);
/*  Unlink the file if it can be accessed.  */
	if (opt.dbug > 0) printf (" Deleting %s\n", lf);
	if (access(lf, F_OK) == 0) unlink (lf);
	free (lf);
/* Can return now, no Pvfs storage has been allocated. */
	return  k;
    }

/* Obtain current compiled data if needed (!rfil || !tfil || wvfd || ncmp) */
// OK, let's just do this anyway; 
// needed if  wvfd || !rfil || !tfil
// not needed only if tfil && (rfil || cfil)
// ==> needed if  wvfd || !tfil || (!rfil && !cfil)
	cv = makevfs ();

/* Obtain the actual file name to use */
    if (filename == NULL) source = VERSFIL;
    else		{ source = filename;  sf = 1; }
    if (opt.dbug > 0) printf (" File name: %s\n", source);

// Write to file moved below for logic reasons 10/12-16 KDC

/* Normalize flags; if both rfil and cfil are on, turn cfil off. */
    if (opt.rfil && opt.cfil) opt.cfil = 0;	// put in main, too.

/* Read file data if needed (rfil OR cfil)  */
    if (opt.rfil || opt.cfil) {
	buffr = getafile (source);
	if (buffr == NULL) printf (" NULL result.\n");
	else {
	    if (opt.dbug) printf (" Reading \"%s\" ",  source );
	    if (opt.dbug > 1)printf (" Found %d characters.\n", (int)strlen(buffr));
	    if (opt.dbug > 1)printf (" Content: %s\n", buffr);
	    pp = getvfs (buffr, source);  // Add file name as argument
	    if (pp == NULL) printf (" Could not parse buffer.\n");
	    free (buffr);
	}
    }

/* Write compiled data (or data read) to file request */
    if (opt.wvfd) {
Pvfs	*p;	// Where we write from...
//	if (opt.self) k = selftest (cv);
/* If reading data from file (rfil on) then we write from version structure pp;
  otherwise, use current version data in version structure cv as a default.  */
	p = opt.rfil ? pp : cv;
	k = selftest (p);
	if (k) goto clear_out;  // was return  k;	// Return all errors.
/* Fix up internal checksum in buffer p (could be made into subroutine)  */
	{	// Change version on VCONCVV to vschecks(p->ver)
Vers	*vp;
	    vp = findver (VCONCVT, p);
	    if (vp != NULL) vp->v = vschecks(p->ver);
	}
	buffr = putvfs (p);
	if (opt.dbug > 1)
	    printf (" Constructed buffer with %d characters.\n",
		(int)strlen(buffr));
	putafile (buffr, source);
        if (opt.dbug > 0) printf (" Wrote %d characters in buffer to %s\n",
	    (int)strlen(buffr), source );
	if (opt.dbug > 1) printf (" Content: %s\n", buffr);
	free (buffr);
	goto clear_out;  // was return  k;
    }

/* Obtain FTP data if needed (tfil || !rfil)  */
    if (opt.tfil || !opt.rfil) fp = getcurver (VERSFIL);

/* Set up the source and reference pointers */
//    ap = opt.rfil ? pp : cv;
//    bp = opt.tfil ? fp : cv;
    if (opt.rfil) ap = pp;
    else if (opt.tfil) ap = fp;
    else ap = cv;

    if (opt.cfil && ap != pp) bp = pp;
    else if (opt.tfil && ap != fp) bp = fp;
    else bp = cv;

    if (ap == bp) bp = fp;	// in case no flags, both were cv.

/* Need a big debug to explain options and ap,bp values */
    if (opt.dbug > 1) {
char	afn[8], bfn[8];
	printf (" File steering options:  r: %d  f: %d  t: %d\n",
		opt.rfil, opt.cfil, opt.tfil);
	printf (" Other options:   c: %d  s: %d\n",
		opt.ncmp, opt.self);
	if      (ap == cv) strcpy (afn, "cur");
	else if (ap == fp) strcpy (afn, "ftp");
	else if (ap == pp) strcpy (afn, "file");
	else               strcpy (afn, "other");
	if      (bp == cv) strcpy (bfn, "cur");
	else if (bp == fp) strcpy (bfn, "ftp");
	else if (bp == pp) strcpy (bfn, "file");
	else               strcpy (bfn, "other");
	printf (" Data compares are:  %s-%s\n", afn, bfn);
    }

/* Process the input part */
//    if (opt.self) k = selftest (ap);
    k = dispvfs (ap);	// Display the given item
/* The k result from both of the above should be the same, so we do not
  increment k by the second result if they were both done.  In fact, the
  self test is so redundant here, that we just drop it.  */
    if (opt.self && k > 0) goto clear_out;  // was return k;
/* The self test flag truncates execution if errors are found */

/* Return if no compare is indicated */
    if (opt.ncmp) goto clear_out;  // was return  k;

/* Process the compare of versions */
    k += compvfs (ap, bp);	// compare to other version...

// Find out how to free Pvfs storage??
clear_out:	// clean up the dynamic Pvfs storage
    pp = kill_pvfs (pp);
    cv = kill_pvfs (cv);
//    fp = clean_pvfs (fp);	// managed in getcurver
/* Return any status derived */
    return  k;
}


	/* - - - - - - - - *\
	|		    |
	|    Main Program   |
	|		    |
	\* - - - - - - - - */

/*  ====  MAIN PROGRAM  ====  */

#undef  PROMPTLOOP

int  main (int argc, char** argv) {
extern char *optarg;
extern int optind,opterr,optopt;
int c;
int     dohelp = 0;
// The following symbol is nominally undefined
#ifdef  LOOPMETHODUSEDASOLDWAY
char	outfile [120];	/* For prompting */
#endif
opts	opt;
namlist	*head = NULL;
char	*filename = NULL;
#ifdef  LOOPMETHODUSEDASOLDWAY
int	k, m, n;	/* Process counts */
#else
int	k = 0;
#endif
char	*vfld;
int	xx = 0;		/* Exit flag */
int	yy = 0;		/* test mode */

/* Usual flags, debug and help */

/* Standard input logic:  If argument given, take that argument as the
  sole specification to process.  If no argument is given, prompt for
  an argument, and repeat the prompting until a null reply.  */

    opt.dbug = 0;	/* Or - put data in define above */
    opt.tfil = 0;	/* Comparison is to FTP (1) or current (0) */
    opt.rfil = 0;	/* Read file, source is file (1), not compiled (0) */
    opt.cfil = 0;	/* Compare to file unless rfil on */
    opt.self = 0;	/* Self test; check source data */
    opt.ncmp = 0;	/* Non compare, source only */
    opt.dlcf = 0;	/* Deletion of local file requested. */
    opt.wvfd = 0;	/* Write version file data */

/* Decode all the options of the command line */
/* :: follows letter with optional argument, : follows one with required
  argument, otherwise there are no arguments to the options.  */
    while((c = getopt(argc, argv, "cd::fqrstvyzh")) != -1) switch(c) {
	case 'c':	opt.self=opt.ncmp=1; 	break;  // just do self test
	case 'd':				// debug level
	    opt.dbug = optarg == NULL ? 1 : atoi (optarg);	break;
	case 'f':	opt.cfil = 1;	break;	// compare to file
	case 'q':	xx = 1;		break;	// display version and exit
	case 'r':	opt.rfil = 1;	break;	// read file
	case 's':	opt.self = 1;	break;	// self test
	case 't':	opt.tfil = 1; 	break; // use ftp;  had opt.rfil=1;
	case 'v':	opt.wvfd = 1;	break;	// write file
	case 'y':	yy = 3;		break;	// debug test for return data
	case 'z':	opt.dlcf = 1;	break;	// just drop local file
	case 'h':	dohelp = 1;	break;	// parameter help
	default:
	    printf (" unused %c option\n", c);	break;
    }  /* end of switch, only statement in the getopt while loop */
    if (opt.dbug > 1) {
	printf (" Options:  c%d d%d r%d f%d t%d s%d v%d z%d\n",
		opt.ncmp, opt.dbug, opt.rfil, opt.cfil, opt.tfil,
		opt.self, opt.wvfd, opt.dlcf);
    }

    for (c = optind; c < argc; c++) {
	filename = dynamstr (argv[c]);
	head = pushtxt (argv[c], head);  /* Collect arguments */
	if (opt.dbug > 0) fprintf (stderr, "-Argument %s.\n", filename);
/* NOTE that dynamstr must be used to put any static string into a namlist,
  usually in  get_nam, and the like.  It will cause a crash at the
  end when the name queue is dropped, if this is not done correctly.
  That is why pushtxt is used above.  */
    }

/* Normalize flags; if opt.rfil and opt.cfil are both on, turn
  opt.cfil off as they cannot both be used.  */
    if (opt.rfil && opt.cfil) opt.cfil = 0;

    vfld = fvstr (VERSUPV);

    if (dohelp) {
	printf ("\n %s -- Version %s", PROGNAME, vfld);
	printf ("\n  Parameters are: ");
	printf ("\n     File name  [%s]", VERSFIL);
	printf ("\n    Default is to compare current vs FTP data");
	printf ("\n   Switches");
	printf ("\n    -d* = Set debug level to (*) [1]");
	printf ("\n     -h = Display this help");
	printf ("\n     -q = Display version and exit");
	printf ("\n     -r = Read file as source");
	printf ("\n     -f = Read file as reference (if not source)");
	printf ("\n     -t = Use FTP data as source");
	printf ("\n     -v = Write current data to file");
	printf ("\n     -s = Self test of data");
	printf ("\n     -c = Self test only, no compare");
	printf ("\n     -z = Delete local FTP data only");
	printf ("\n   Supports version control file.");
/*	printf ("\n   Output file name is prompted for."); */
	printf ("\n \n");
	return (0);
    }

    if (!opt.dlcf)
	printf ("\n %s -- Version %s\n", PROGNAME, vfld);
//    if (opt.ncmp) opt.tfil = 0;  /* Holdover from old meaning of tfil */
    if (xx) return  0;	// Implement -q
    if (yy) return  yy;	// debug only -- see ~/imacs/linux/verfile

/*  -- Loop to process arguments, or prompt if no arguments -- */
#ifdef  LOOPMETHODUSEDASOLDWAY
    for (k=n=m=0; k >= 0; ) {	// Forever loop
	if (head == NULL) {	// No arguments, or looped out
#ifdef  PROMPTLOOP
	    if (m > 0) break;	// Had real arguments, looped out
#else
	    if (n > 0) break;	// No loop after any counted argument
#endif
	    askmestr (outfile, 110, "Output file:");
	    if (*outfile == '\0') break;	// Exit on any null reply
	    filename = dynamstr (outfile);
	} else {		// We have a command line argument given
	    filename = pop_name (&head);
	    if ( filename == NULL) continue;
	    if (*filename == '\0') continue;
	    m++;	// Count number of command line arguments given
	}
	n++;		// Count number of arguments requested

	verfile (filename, opt);  k++;
	free (filename);	// MUST be dynamic
	filename=NULL;
    }
#else
    k = verfile (filename, opt);
    if (k > 0) printf (" ** Total of %d errors detected.\n", k);
#endif

/* Free dynamic storage used for version field and filename. */
    free (filename);
    free (vfld);
    return  k;
}

/* Some change notes

    10/12-16 - Version 0.83
	-- Add -e (edit) option to read & write a named file
	-- Edit mode; if -r and -v are both active, read the file first
	and then write from that data rather than current.

    08/11-05 - Version 0.82
	-- Convert to "getopt" utility support
	-- Improve option flags for greater versatility
	-- Clean up some code in main and verfile
	--> Minimum version for release 2.06+
	-- (2010/12-15) Make compatable with new getcurver (clean_pvfs)

    08/03-19 - Version 0.81
	-- Using global "Program" pointer.
	-- Return 0 status when giving help.

    07/10-24 - Version 0.80
	-- Add clear of Pvfs storage where needed.
	-- Minimum Version for release 2.00 support.

    07/10-23 - Version 0.76
	-- Add checksum support.

    07/10-19 - Version 0.75
	-- On the -s flag, to test on first data and exit early
	with status if in error.
	-- Add -c flag to just do self test and return status
	-- Add -z flag to delete any local file (only).
	-- Totally reorganize the processing subroutine.

    07/10-17 - Version 0.70
	-- General programs all moved to ioutils.c and mgutils.c
	-- Early test code removed, doing real things only

    07/10-16 - Version 0.60
	Testing of local file and FTP file reading

    07/10-15 - Version 0.50
	Removed many general subroutines to ioutils; noted that ioutils is
	not made automatically, check makefile.
	(also, this routine's make does not compile mgutils or depend on
	mgfeats.h -- it really should).
	Remaining routines need to go to mgutils.c
	Now mostly the report formatting programs which remain here.

    07/10-12 - Version 0.40
	Change names of structures and members, and names of
	some programs.

    07/10-12 - Version 0.30
	Add report of data, and compare of data with
	a reference data set.
	Generalized display of a list of features.

    07/10-11 - Version 0.17
	Add struct from basic data, and buffer from struct.
	Will eliminate buffer from basic data (not needed).

    07/10-10 - Version 0.15
	Changing version for linking with mgutils

    07/10-09 - Version 0.10
	Initial test version, testing file formats, read from/to file.

.. */

