/* Balanced Tree algorithms...   See Knuth, V3, P455 */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **
**                                                                   **
**                 Copyright (C) 1994, 2001 by                       **
**     Ken Clardy, Carnegie Observatories, Pasadena California.      **
**                                                                   **
**      This software is proprietary, and may be used or copied      **
**          only with the written consent of the author.             **
**              All copies must retain this notice.                  **
**                                                                   **
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include	<stdlib.h>	/* malloc, free */
#include	"btree.h"

/*  ------  Add a node to the tree  ------  */
void *btree_add (head, data, subr)
/* Return data address found, same as data if new */
BNODE	**head;
void	*data;		/* Object to add to tree */
int	(*subr)();	/* Comparison subroutine */
{
BNODE	*n;
BNODE	*p, *q, *r, *s, *t;
BNODE	*v, *w;
int	a;
int	x;

/* Create a new node to hold the data */
    n = (BNODE*) malloc (sizeof (BNODE));
    n->llink = n->rlink = NULL;
    n->balance = 0;
    n->rank = 1;
    n->data = data;

/* Check for special case, null head */
    if (*head == NULL)  {
	*head = n;
	return data;
    }
/* in above, etc, note that head is pointer to pointer, and
  not pointer itself so that it may be modified... */

/* Look for where the new node goes */
    t = *head;
    p = s = t;
    w = NULL;
    while (p != NULL)  {
/* NOTE -- Here, if compare equal, we terminate search;
    It would also be possible to allow equal values, but then
    we could only find one and not all of them... */
	x = (*subr) (n->data, p->data);
	if (x == 0)  {
	    free ((char*)n);
	    while (w != NULL) {	/* Remove rank increments */
		v = w->llink;
		--(v->rank);
		v = w->rlink;
		free ((char*)w);
		w = v;
	    }
	    return p->data;
	} else if (x < 0)  {
	    ++(p->rank);
	    v = (BNODE*) malloc (sizeof (BNODE));
	    v->rlink = w;
	    v->llink = p;	/* Track rank increments */
	    w = v;
	    q = p->llink;
	    if (q == NULL)  {
		p->llink = q = n;
		break;
	    }
	}  else  {
	    q = p->rlink;
	    if (q == NULL)  {
		p->rlink = q = n;
		break;
	    }
	}
	if (q->balance != 0)  {
	    t = p;
	    s = q;
	}
	p = q;
    }
    while (w != NULL) {	/* Free increment track storage */
/* In this case, do not restore the ranks, insertion being made. */
	v = w->rlink;
	free ((char*)w);
	w = v;
    }

/* Adjust the balance factors... */
    r = p = ((*subr) (n->data, s->data) < 0) ? s->llink : s->rlink;
    while (p != q)  {
	x = (*subr) (n->data, p->data);
	if (x == 0) break;	/* Should occur only if p == q */
	else if (x < 0) {
/* NOTE that balance factors of nodes between s and q are currently zero */
/* ALSO balance of s is non-zero, and balance of q (just added) is zero */
	    (p->balance)--;	/* Book says set to -1 ? */
	    p = p->llink;
	}  else  {
	    (p->balance)++;
	    p = p->rlink;
	}
	if (p == NULL) {
	    break;	/* Should not do this, catch error */
	}
    }

/* Balancing act... */
    a = ((*subr)(n->data, s->data) < 0) ? -1 : 1;
/* In the following, we implicitly do "s->balance += a" */
    if (s->balance == 0) {
/* Tree has grown higher */
	s->balance = a;
	/* Increment total height counter */
	return data;
    }
    if (s->balance == -a)  {
/* Tree is more balanced */
	s->balance = 0;
	return data;
    }
/* Now assume s->balance is == a ... */
/* if (s->balance == a) return data; */ /*DEBUG! THIS SHOULD BYPASS BALANCING */
    if (r->balance == a)  {
/* Single rotation */
	p = r;
	if (a > 0)  {
	    s->rlink = r->llink;
	    r->llink = s;
	    r->rank += s->rank;
	} else {
	    s->llink = r->rlink;
	    r->rlink = s;
	    s->rank -= r->rank;
	}
	s->balance = r->balance = 0;
    }
    if (r->balance == -a)  {
/* Double rotation */
	if (a > 0)  {
	    p = r->llink;
	    r->llink = p->rlink;
	    p->rlink = r;
	    s->rlink = p->llink;
	    p->llink = s;
	    r->rank -= p->rank;
	    p->rank += s->rank;
	} else {
	    p = r->rlink;
	    r->rlink = p->llink;
	    p->llink = r;
	    s->llink = p->rlink;
	    p->rlink = s;
	    p->rank += r->rank;
	    s->rank -= p->rank;
	}
	s->balance = (p->balance == a) ? -a : 0;
	r->balance = (p->balance == -a) ? a : 0;
	p->balance = 0;
    }

/* Finishing touch, update the top pointer */
/* Due to transform, we must put p where we got s from */
    if (s == *head) {
	*head = p;
    } else {
	if (s == t->rlink)
	    t->rlink = p;
	else
	    t->llink = p;
    }
    return data;
}

/*  NOTE -- We do not have a routine here to remove (delete) a node
  from the b-tree.  Doing a delete would involve re-linking the node's
  sub-trees, with either a left or right rotation, the longer sub-tree's
  head would replace this head, with change of pointer in the parent,
  and re-adjustment of rank and balance all the way back up the tree.
  Since this implementation has no back links to parent nodes, a complex
  method similar to the add algorithm would be needed.  Fugedaboudit. */

/*  ------  Count nodes in tree  ------  */
int  btree_length (head)
/* Return total of nodes in tree */
BNODE	**head;
{
BNODE	*p;
int	sum;
/* Sum rank fields of head and all right subtrees... */
    for (sum=0, p = *head; p != NULL; p = p->rlink)
	sum += p->rank;
    return sum;
}

/*  ------  Index tree as a linear list  ------  */
void *btree_index (head, n)
/* Return the data pointer of item n */
BNODE	**head;
int	n;
{
BNODE	*p;

/* Initialize */
    p = *head;

/* Check sum and rank to see where to go... */
    while (p != NULL)  {
	if (n < p->rank)  {
	    p = p->llink;
	} else if (n > p->rank)  {
	    n -= p->rank;
	    p = p->rlink;
	} else {
	    break;
	}
    }
/* if p null, return NULL, otherwise return pointer from p */
    if (p != NULL) return p->data;
    else  return  NULL;
}

/*  ------  Visit all nodes of the tree, in order  ------  */
void  btree_visit (node, subr, args)
/* Note that no comparisons are needed here */
BNODE	*node;
void	(*subr)();
void	*args;
{			/* Implemented as an ideal recursive algorithm */
    if (node == NULL) return;			/* Omit any null cases */
    btree_visit (node->llink, subr, args);	/* Visit the left sub-tree */
    (*subr) (node, args);			/* Visit the node itself */
    btree_visit (node->rlink, subr, args);	/* Visit the right sub-tree */
}


/* Enhancement == Add range subroutine to restrict visits to a range
  of nodes.  The "range (node)" will return -1 if node is outside the
  range to the left, 0 if within range, and +1 if outside the range
  to the right.  Only nodes within range will be "visited" i.e. call the
  action subroutine, and those will be done in sort order.  This range
  subroutine should be related to the normal decision routine, or possibly
  call that decision routine for two (global) range values. */

/*  ------  Visit range of nodes of the tree, in order  ------  */
void  btree_visit_rest (node, range, subr, args)
/* Note that no comparisons are needed here */
BNODE	*node;
int	(*range)();
void	(*subr)();
void	*args;
{			/* Implemented as an ideal recursive algorithm */
int	r;	/* Should be put on stack... */
    if (node == NULL) return;			/* Omit any null cases */
    r = (*range) (node);	/* Check on proper range */
    if (r >= 0) btree_visit (node->llink, subr, args);	/* Visit left  */
    if (r == 0) (*subr) (node, args);			/* Visit node  */
    if (r <= 0) btree_visit (node->rlink, subr, args);	/* Visit right */
}


/*  ------  Find data pointer with given key  ------  */
void *btree_find (root, key, subr)
/* Return pointer to node with key, using comparison subroutine */
BNODE	**root;		/* The tree root from which to search */
void	*key;		/* Key to pass to compare routine */
int	(*subr)();	/* Special comparison subroutine, key to object */
{
BNODE	*p, *q;
int	x;

/* Check for special case, null head */
    if (*root == NULL)  return NULL;

/* Look for the key itself */
    p = *root;
    while (p != NULL)  {
	x = (*subr) (key, p->data);
	if (x == 0)	return p->data;	/* Found it! */
	else if (x < 0)	q = p->llink;
	else		q = p->rlink;
	p = q;
    }
    return NULL;
}

/*  ------  Kill a tree, and recycle the waste products  ------  */
BNODE *btree_kill (node, subr)
/* Return a null pointer for good form. */
BNODE	*node;		/* The node from which to kill */
int	(*subr)();	/* Subroutine to clear data */
{		/* Implemented as an ideal recursive algorithm */
    if (node == NULL) return NULL;	/* Omit any null cases */
    node->llink = btree_kill (node->llink, subr);  /* Kill left */
    node->rlink = btree_kill (node->rlink, subr); /* Kill right */
    (*subr) (node->data);		/* Clear the data */
    free ((char*)node);			/* Clean node itself */
    return  NULL;
}

/*  =*=  */
