/* ioutils.c -- Input/Output Utility subroutines */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **
**                                                                   **
**                   Copyright (C) 2011 by                           **
**     Ken Clardy, Carnegie Observatories, Pasadena California.      **
**                                                                   **
**      This software is proprietary, and may be used or copied      **
**          only with the written consent of the author.             **
**              All copies must retain this notice.                  **
**                                                                   **
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/*  ====  Needed Includes  ====  */
#include	<stdlib.h>
#include	<stdio.h>
#include	<string.h>
#include	<time.h>
#include	<unistd.h>	// for bits in access
#include	<math.h>	// for sexigisimal formatting; exp, log
#include	<ctype.h>	// for toupper/tolower

/* Needed for libcurl support */
#include <curl/curl.h>
#include <curl/types.h>
#include <curl/easy.h>

/* Typedefs otherwise needed ---------- */

// typedef	struct	vector2 { double x, y; } vect2;
/* That is in lieu of including kdcutil.h for reading vectors. */
/* The kdcutil.h is now included.  */

#define   IS_IOU_C
#include	"ioutils.h"

#ifdef  SYS_Darwin
#include	<sys/fcntl.h>  /* for fcntl */
#endif

/*  ====  Global Data  ====  */

static	char*	FNT[32]={[0 ... 31]NULL};
/*  FNT is used only by subroutines in ioutils.c, so it is defined as
  static here.  If needed, comment this definition and restore the
  global external definition in ioutils.h    */

/*  ====  Standard Definitions  ====  */

/*  ---  Local function prototypes  ---  */
static void	dskill (char* *a);

/*  -- Code, in order specified in utils.h file -- */

/* Some locally defined things which are in mgutils but belong here */
#define  C_loop(q,a)  for(q=(a); q!=NULL; q=(q->next!=(a))?q->next:NULL)
/* Above is really defined in maskdat.h which does not yet get included here */

#define  xalloc(b)  ( (b*) calloc (1,sizeof(b)) )
/* Above is defined in optutils.h, which does not get included here */

typedef	struct	g_link {
	struct	g_link	*next, *last;
	} g_link;

// these are in mgutils.c; if we have a problem, just put in
// the  function prototypes  for these programs.

static	void*	G_pop_c (void* x)
/* Remove structure from circular queue, return pointer to next */
{
g_link	*a = (g_link*)x;
g_link	*p, *q;
    if (a == NULL) return a;
    p = a->next;
    if (p == a) p = NULL;
    q = a->last;
    if (q == a) q = NULL;
    if (p == NULL) p = q;
    if (p != NULL) p->last = q;
    if (q != NULL) q->next = p;
    a->last = a->next = a;
    return p;
}

static	void*	G_head (void* a)
/* Find head of queue, circular or linear */
{
g_link	*p;
    if (a == NULL) return a;
    for (p=a; p->last != NULL; p = p->last) {
	if (p->last == a) return a;
// If circular, return a as the head of the list.
    }
    return  p;
}

static	void*	G_tail (void* a)
/* Find tail of queue, circular or linear */
{
g_link	*p;
// if circular, use a->last; if linear, follow path
    if (a == NULL) return a;
    for (p=a; p->next != NULL; p = p->next) {
	if (p->next == a) break;
// break if found circular, return a is not exactly right.
// If circular, return the value before a as the tail.
    }
    return  p;
}

static	void	G_circ (void* a)
/* Check for linear list a, and make circular if so */
{
g_link	*p, *q;
    if (a == NULL) return;
    p = G_head (a);
    q = G_tail (a);
    if (p->last != NULL && q->next != NULL) return;	// redundant?
    if (p->last == NULL) p->last = q;
    if (q->next == NULL) q->next = p;
/* Should (recursively?) loop until both head and tail are found
  to be without NULL pointers, although logically that can happen
  only once?  */
}

static	void*	G_push_c (void* y, void* x)
/* Put element b (y) before element a (x), and other circularly linked
  items get linked together.  Return new head (a) pointer. */
/* b is tail of list inserted before a */
{
g_link	*a = (g_link*)x;
g_link	*b = (g_link*)y;
g_link	*p, *r;

/* Find queue ends such that a...p and r...b are merged, into
  something like  ...p - r...b - a... ; this replaces the links
  ...p - a... and ...b - r... with the p - r and b - a links. */
/* If a has null predicessor, or b a null follower, circularize
  that list and proceed as before.  */
/* Take care of null cases */
    if (a == NULL) return b;
    if (b == NULL) return a;
/* Link it up */
    if (a->last == NULL) G_circ (a);
    p = a->last;		/* ...p - a... */
    if (b->next == NULL) G_circ (b);
    r = b->next;		/* ...b - r... */
    p->next = r;
    r->last = p;
    b->next = a;
    a->last = b;		/* ...p - r...b - a... */
    return a;
}


/*  ====  String Manipulation  ====  */

void  trimend (char* a, char* space)
/* Remove trailing whitespace from a string */
{
char*	p;
    if (a == NULL) return;
    if (*a == '\0') return;
    for (p=a; *p != '\0'; p++) ;
/* Points p at null termination of string a */
    for (p--; p >= a; p--) {
	if (strchr (space, *p) == NULL) break;
	*p = '\0';
    }
}

void  trimbeg (char* a, char* space)
/* Remove leading whitespace from a string */
{
char*	p;
    if (a == NULL) return;
    for (p=a; *p != '\0'; p++) {
	if (strchr (space, *p) == NULL) break;
    }
/* Points p at first non-whitespace character in string a */
    if (p == a) return;	/* null case */
    while (*p != '\0') *a++ = *p++;
    *a = '\0';
}

void  trims (char* a, char* space)
/* Remove leading and trailing whitespace from a string */
{
char	*b, *p;
    if (a == NULL) return;
    for (p=b=a; *p != '\0'; p++) {
	if (strchr (space, *p) == NULL) break;
    }
/* Points p at first non-whitespace character in string a */
    if (p != a) {	/* null case */
	while (*p != '\0') *b++ = *p++;
	*b = '\0';
    } else {
	for (b=a; *b != '\0'; b++) ;
    }
    if (*a == '\0') return;	/* If became empty */
/* From above, b points to current null end. */
    for (b--; b >= a; b--) {
	if (strchr (space, *b) == NULL) break;
	*b = '\0';
    }
}


char*  dyncat (char* pfx, char* nam)
/*  Return a dynamic string concatenating pfx to nam.
  Any trailing blankspace in pfx, and leading blankspace in
  nam does not appear in the result.  */
{
char	*r = NULL;
char	*pf = NULL;
char	*nx = NULL;
int	lp = 0;
int	ln = 0;

/* Obtain cleaned dynamic copies of pfx and nam, when present */
    if (pfx != NULL) {
	pf = strdup (pfx);
	trimend (pf, BLANKSPACE);
	lp = strlen (pf);
    }
    if (nam != NULL) {
	nx = strdup (nam);
	trimbeg (nx, BLANKSPACE);
	ln = strlen (nx);
    }

/* Satisfy any trivial cases, return cleaned copy */
    if (lp < 1) {
	if (pf) free (pf);
	return  nx;
    } else if (ln < 1) {
	if (nx) free (nx);
	return  pf;
    }

/* Non-trivial case, do an actual concatenation */
    r = malloc (lp + ln + 1);
    strcpy (r, pf);	free (pf);
    strcat (r, nx);	free (nx);
    return  r;
}


// #ifdef  NOTDEFINED
char*  fncat (char* pfx, char* nam)
/*  Return a dynamic string concatenating pfx to nam, putting a /
  between them if needed (i.e. if not already there); and possibly
  removing multiple / between them.
  NOTE that trailing blanks in pfx, and leading blanks in nam
  are removed by this routine.  */

/* ACTUALLY, THIS IS NOT NEEDED.  All uses could use dyncat as
  given above, since pfx has the terminal / already.  */

/* This is revived to be used in creating directory/filename
  strings elsewhere.  One use is in put_cache in skymap.  */

{
char	*r = NULL;
char	*pf = NULL;
char	*nx = NULL;
//int	lp = 0;
//int	ln = 0;

// Check for trivial cases
    if (pfx == NULL) return dynamstr (nam);
    pf = strdup (pfx);
    trimend (pf, "/ \t");
    if (strlen(pf) < 1) {
	free (pf);
	return dynamstr (nam);
    }
    r = dyncat (pf, "/");
    free (pf);
    if (nam == NULL) return  r;
    pf = r;
/* At this stage, pf is a dynamic string consisting of the prefix
  followed by exactly one / character, and the prefix is non-null;
  also, nam is not null.  */

    nx = strdup (nam);
    trimbeg (nx, "/ \t");
/* Now, nx is a dynamic string without a leading / character,
  or possibly it is null.  */

    r = dyncat (pf, nx);
    if (pf) free (pf);
    if (nx) free (nx);
    return  r;
}
// #endif  // not defined; above stuff is dropped.


char  *isotime (char *tbuf)
/* Provide the current time in ISO-8601 format.  Use the given
string if tbuf is not null, otherwise obtain a string dynamically.
Returns the string used.  Up to 20 characters are put into string.
Time format is:  yyyy-mm-ddThh:mm:ss
Requires 19 characters, and a 20 character buffer with null end. */
{

/* utime.h or time.h or sys/timeb.h
  see ~/source/lgsub.c or clock.c 
  utime is file access and modification times
  sys/timeb.h gives milliseconds of day; see man ftime
  time.h is usual time of day and date, see man ctime
  sys/time.h defines timeval structure, seconds + microseconds -
    see man gettimeofday
  Actual current time is from "man -s2 time" -- the time function.
  See man strftime - can convert data and time to string with formatting */
/* For the format see /home/draco/clardy/ISO8601.html, and 
  /home/draco/clardy/text/8601v3.pdf and ~/text/8601.pdf 
 also see  http://www.itl.nist.gov/fipspubs/fip4-1.htm
 or do a search on "iso-8601" for further. */

time_t	tloc;
struct tm *tm;
char	*buf;
/* Get the appropriate string pointer into local buf pointer */
    if (tbuf != NULL)	buf = tbuf;			/* Given */
    else		buf = (char*) malloc (20);	/* Gotten */

/* Obtain current time, put into structure, and format it to string */
    time (&tloc);		/* Gets the current time into tloc */
    tm = localtime (&tloc);		/* Find y-m-d-h-m-s+ in *tm */
    strftime (buf, 20, "%Y-%m-%dT%T", tm);	/* Format time to string */
    return  buf;
}


char*	upcase (char* s) {
/* Change string to uppercase; return pointer to argument */
char*	p;
    if (s == NULL) return s;
    for (p=s; *p != '\0'; p++)
	*p = toupper (*p);
    return  s;
}

char*	lowcase (char* s) {
/* Change string to lowercase; return pointer to argument */
char*	p;
    if (s == NULL) return s;
    for (p=s; *p != '\0'; p++)
	*p = tolower (*p);
    return  s;
}

char*	upcasen (char* s, int n) {
/* Change up to n characters of s to upper case, return pointer to arg. */
char*	p;
    if (s == NULL) return s;
    for (p=s; (*p != '\0' && n > 0); p++, n--)
	*p = toupper (*p);
    return  s;
}

char*	lowcasen (char* s, int n) {
/* Change up to n characters of s to lower case, return pointer to arg. */
char*	p;
    if (s == NULL) return s;
    for (p=s; (*p != '\0' && n > 0); p++, n--)
	*p = tolower (*p);
    return  s;
}




/*  ====  Queue Manipulation  ==== */

#define  DYNAMICNAMLIST
//#undef  DYNAMICNAMLIST
/*  Symbol  DYNAMICNAMLIST  is defined locally to change logic of get_nam to
  assure that a dynamic copy of the given string rather than the given pointer
  is inserted in the resulting namlist structure.  This will allow the resulting
  namlist to be able to be destroyed by kill_name correctly.  The old way is
  equivalent to undef, and the new method to define.  The result of this change
  is a possible memory link if get_nam is called with a dynamic pointer.    */

char	*dynamstr (char *a)
/* Return a dynamic string copy of given string a.  If a is NULL
  or empty, returns NULL */
/* This duplicates strdup (in string.h) but may have more robust
  error checking for null cases.  */
{
char	*r;
int	l;
    if (a == NULL) return NULL;
    l = strlen (a);
    if (l < 1) return NULL;
    r = (char *) malloc (l + 1);
    strcpy (r, a);
    return  r;
}

char    *pop_name (namlist **a)
/* Remove **a from the queue, and return its name (value) pointer.
  Adjust *a to point to next node in queue, or NULL if emptied. */
/* Queue element *a is also freed as a by-product here. */
{
char    *r;
namlist *h;
namlist *p, *f;
    if (a == NULL) return NULL;
    h = *a;
    if (h == NULL) return NULL;
    r = h->name;
    p = h->last;
    f = h->next;
    if (p != NULL) p->next = f;
    if (f != NULL) f->last = p;
    *a = (f == h) ? NULL : f;
    free (h);
    return  r;
}

namlist	*get_nam (char *a)
/* Create a namlist entry for character value a.  Returns NULL
  if a is NULL or empty.  Puts pointer a directly into structure,
  be sure to use dynamstr if you have a static buffer input. */
/* The DYNAMICNAMLIST version puts a dynamic copy of a into the
  created structure, and can be called with a static input.  */
{
namlist	*r;
int	l;
    if (a == NULL) return NULL;
    l = strlen (a);
    if (l < 1) return NULL;
    r = (namlist *) malloc (sizeof(namlist));
    r->last = r->next = r;	/* Pointers never null */
#ifdef  DYNAMICNAMLIST
    r->name = strdup (a);   // dynamic copy used
#else
    r->name = a;
#endif
    return  r;
}

namlist *push_name (namlist *b, namlist *a)
/* Put the queue or element b into queue a, so that it 
  preceeds a.  Return the new head (a) pointer.  */
{
namlist *p;
namlist *r;

/* Take care of null cases */
    if (b == NULL) return a;
/* Find queue ends such that a...p and b...r are merged, into
  something like  ...p - b...r - a... ; this replaces the links
  ...p - a... and ...r - b... with the p - b and r - a links. */
/* If a has null predicessor, we link ...0 - b...r - a...; if
  b has null predicessor, find the next end and use that. */
/* Link it up */
    if (a == NULL) return b;
    p = a->last;
    r = b->last;
    if (r == NULL) {	/* Scan to end */
	for (r = b; r->next != NULL; r = r->next) {
		if (r->next == b) break;
	}
    }
    if (p != NULL) p->next = b;
    b->last = p;
    if (r != NULL) r->next = a;
    a->last = r;
    return a;
}

namlist*  pushtxt (char* text, namlist* a)
/* Get a dynamic string and push it onto the a queue */
/* In the DYNAMICNAMLIST version, the dynamic copy is obtained in the
  get_nam program which we call here.    */
{
namlist *b;
//char	*c;
//    c = dynamstr (text);
//    b = get_nam (c);
#ifdef  DYNAMICNAMLIST
    b = get_nam (text);
#else
    b = get_nam (strdup(text));
#endif
    return  push_name (b, a);
}

namlist	*kill_name (namlist *a)
/* Kill the entire queue from a... */
{
namlist	*b;
char	*c;
    b = a;
    while (b != NULL) {
	c = pop_name (&b);
// The queue element at b is freed, and b adjusted.
//#ifdef  DYNAMICNAMLIST
	if (c != NULL) free (c);
//#endif
/* We might want to do the free only on DYNAMICNAMLIST defined, and omit
  the free operation when undef.  But that causes a memory leak at the
  expense of never trying to free static storage.    */
    }
    return  b;
}

/*  ===  Rational Namlist Management  ===  */
/* Here, all character strings are dynamic, as are all namlist
  queue elements.  Empty strings are allowed.  */

namlist	*new_nam (char *a)
/* Create a namlist entry for character value a.  Enters a NULL name
  pointer if a is NULL.  Puts pointer a directly into structure, and
  assumes *a points to a dynamic string.
  To use with static string, use new_nam (strdup(a))
  Yes, this is very similar to get_nam -- except that it:
  1) Does not reject an empty string, or NULL.
  2) Puts a in structure ASSUMING it to be dynamic.    */
{
namlist	*r;
//    if (a == NULL) return NULL;
    r = xalloc (namlist);
    r->last = r->next = r;	/* Pointers never null */
    r->name = a;

    return  r;
}

#ifdef  NOTDEFINEDNAME   // this function is a comment
static	namlist *push_nl (char* a, namlist* p)
/* Create a new namlist entry for a and put it on queue p, return
  the new head pointer.  */
{
    return  push_name (new_nam (a), p);
}
#endif

/* We use here a local version of strndup, which is otherwise a gnu
  standard function.  It would need an explicit include before
  Fedora 13, but later is defined with string library.  For the
  static function, use a unique name to avoid redefine problem.  */
static	char	*kstrndup (const char *s, size_t n)
/* Duplicate the _gnu_source string function; make a dynamic copy
  of (at most) n characters from s and return it.    */
{
char	*r;
int	j;
    j = (s == NULL) ? 0 : strlen (s);
    j = (j < n) ? j : n;
    r = calloc (1, j + 1);
    if (j < 1) return r;
    strncpy (r, s, j);
    return  r;
}


/*  ====  Prompting  ====  */
void  askme (char* buf, int blen, char* prompt)
/* Write prompt, then read buf from stdin */
{
int	l;
    printf ("%s", prompt);
    *buf = '\0';
    fgets (buf, blen, stdin);	/* Stores newline and nul */
    l = strlen (buf);
/* Kill any control characters at end of reply */
    while (l > 0) {
	if (buf[--l] < ' ') buf[l] = '\0';	/* ASCII specific */
    }
}

void  askmestr (char *buf, int blen, char *message)
/* Use message in prompt, obtain buf from user... */
{
int	l;
    printf (" >>%s ", message);
    *buf = '\0';
    fgets (buf, blen, stdin);	/* Stores newline and nul */
    l = strlen (buf);
/* Kill any control characters at end of reply */
    while (l > 0) {
	if (buf[--l] < ' ') buf[l] = '\0';	/* ASCII specific */
    }
}



/*  ====  I/O Routines  ====  */

/*  ====  File Name Support  ====  */

/* -===-  The system-dependent code  -===- */

#if defined SYS_Linux

/*  ====  Linux Method  ====  */

char*	fnx (int fd)
/*  Return a dynamic string holding the file name which
  was most probably used to open file descriptor  fd.    */
{
char	path[24];
char	result[256];
int	l;
char	*n=NULL;

    /* Read out the link to our file descriptor. */
    switch (fd) {
	case 0:  n = strdup ("stdin");	break;
	case 1:  n = strdup ("stdout");	break;
	case 2:  n = strdup ("stderr");	break;
	default:
	    sprintf(path, "/proc/self/fd/%d", fd);
	    l = readlink(path, result, sizeof(result));
	    if (l > 0) {
		if (l >= sizeof (result)) l = sizeof (result) - 1;
		result[l] = '\0';
		n = strdup (result);
	    } else {
		strcpy (result, "Error.");
		n = NULL;
	    }
//	    n = strdup (result);
		break;
    }
    return  n;
}

#elif defined SYS_Darwin

/*  ====  Darwin Method  ====  */

char*	fnx (int fd)
/*  Return a dynamic string holding the file name which
  was most probably used to open file descriptor  fd.    */
{
char result[256];
int	l;
char	*n=NULL;

    /* Read out the link to our file descriptor. */
    l = fcntl (fd, F_GETPATH, result);
    if (l >= 0) ;  // (not) result[l] = '\0';
    else  return  n;	// or: strcpy (result, "Error.");
    switch (fd) {
	case 0:  n = strdup ("stdin");	break;
	case 1:  n = strdup ("stdout");	break;
	case 2:  n = strdup ("stderr");	break;
	default:  n = strdup (result);	break;
    }
    return  n;
}

#else  // Not Linux or Darwin

/*  ====  Unknown Method  ====  */

/* At least provide the entry point when compiling in a
  system for which a known method has not been coded.  */

char*	fnx (int fd)
/*  Return a dynamic string holding the file name which
  was most probably used to open file descriptor  fd.    */
{
char	*n=NULL;

/* For unknown system, we really have no clue how to find
  the file name from the descriptor.  So we return NULL.  */
    return  n;
/* This is not really a total fail; in practice this program is
  only called by fhn, and when a NULL is returned, fhn will call
  the backup method which uses the global FNT table.    */
}

#endif  // On System Definitions


static	void	initfnt (void)
/*  Check initialization status of FNT table, initialize if needed  */
{
int	i;
    if (FNT[0] == NULL && FNT[1] == NULL && FNT[2] == NULL) {
//	if (dbug >= 4) printf (" ..initfnt: Initialize table.\n");
	for (i=3; i<32; i++) FNT[i] = NULL;
	FNT[0] = strdup ("stdin");
	FNT[1] = strdup ("stdout");
	FNT[2] = strdup ("stderr");
	for (i=3; i<32; i++) FNT[i] = NULL;
    }
}

char*	fnt (int fd)
/*  Return a dynamic string holding the file name from the table,
  which was most probably used to open file descriptor  fd.    */
{
    initfnt ();
    if (fd < 0 || fd >= 32) return  NULL;
    if (FNT[fd] == NULL) return  NULL;
    return  strdup (FNT[fd]);
}

#undef  TABLE_PRIORITY
//#define  TABLE_PRIORITY
/*  Symbol  TABLE_PRIORITY  is defined to give priority to the
  global table look-up method of obtaining a file name string
  from the file descriptor for an open file.  There is also a
  system dependent method available for some systems; when this
  symbol is defined, the table method is used first, and the
  system method is a back-up.  When undefined, the system method
  is tried first and the table method is used as a backup.
  Advantages of the table method are that it is independent of
  the O.S. used, is quicker, and gives the string used in the
  open request; disadvantages are that not all open/close calls
  may be covered in support of the table.
  Advantages of the system method where available are that it
  should always give a full absolute path name for the file;
  disadvantages are that it might be slower or use more system
  resources.    Normally undefined.        */

char*	fhn (FILE* f)
/*  Return a dynamic string holding the file name which
  was most probably used to open stream pointer  f.
  Use fnt and fnx to get an answer.    */
{
char	*n=NULL;
int	fd;

    fd = fileno(f);
#ifdef  TABLE_PRIORITY
    n = fnt (fd);
    if (n == NULL) n = fnx (fd);
#else
    n = fnx (fd);
    if (n == NULL) n = fnt (fd);
#endif
    return  n;
}

char*	fhname (FILE* f)
/*  Return the last part of the file name obtained from the
  (stream pointer) *f.  Return is a dynamic string, or NULL
  if no name.  Intended for use in an error message.
  Calling program should free the name.    */
{
char	*name;
char	*c, *n;
    c = name = fhn (f);	// The full name, potentially
// Scan for '/' followed by non-null character, and if found, (re)set
// the c pointer to the character following '/'.
    while (c != NULL) {
	n = strchr (c, '/');
	if (n == NULL) break;
	if (*(++n) == '\0') break;
	c = n;
    }
// At loop end, c points to desired part of name, or is NULL.
// If c is whole name, return it without further processing.
    if (c == name)  return  name;
// Get a copy of c to return, free the name (if any) and it's done.
    n = (c == NULL) ? NULL : strdup (c);
    if (name != NULL) free (name);
    return  n;
}

void	setfnt (int  fd,  char*  file)
/*  Set entry fd in FHN to a dynamic copy of file if fd is in range.
  It is possible for fileno(FILE*) to return -1 for a null or bad
  file stream, and we ignore that value.    */
{
    if (fd < 0 || fd >= 32) return;
    initfnt ();
/*    if (dbug >= 3) printf (" ..setfnt: replace %d (%s) with (%s)\n",
			fd,  FNT[fd],  file);  */
    if (FNT[fd] != NULL) free (FNT[fd]);
    FNT[fd] = (file == NULL) ? NULL : strdup (file);
}

void	clearfnt (void)
/*  Clean any dynamic entries in FNT.  */
{
int	i;
/*    if (dbug >= 4) printf (" ..clearfnt: Clear table...\n");  */
    initfnt ();
    for (i=0; i<32; i++) {
	if (FNT[i] != NULL) {
/*	    if (dbug >= 4) printf (" ..clearfnt: free %d (%s)\n", i, FNT[i]); */
	    free (FNT[i]);
	    FNT[i] = NULL;
	}
    }
}


/*  ====  Open / Close Support  ====  */

FILE*	iopen (char* fs, char* tp)
/* Open file, provide error message if not possible */
/* Try opening without giving message, use fopen (library program) */
/* Value of tp is typically "r" or "w" or "rb" or "wb" */
{
FILE    *handle;
char    line[120];

    handle = fopen (fs, tp);

    if (handle == NULL)  {
/*	strcpy (line, "Opening \"");
	strcat (line, fs);
	strcat (line, "\"");   */
        sprintf (line, "Opening \"%s\"", fs);
        perror (line);
/* The error message now prints on stderr. */
    } else {
	strcpy (LastXopen, fs);
	setfnt (fileno(handle), fs);
    }
    return  handle;
}

FILE*   xopen (char* fs, char* tp)
/* Open file, provide error message if not possible */
/* Try opening without giving message, use fopen (library program) */
/* Value of tp is typically "r" or "w" or "rb" or "wb" */
/* Does NOT track last opened file specification */
{
FILE    *handle;
char    line[120];

    handle = fopen (fs, tp);

    if (handle == NULL)  {
	sprintf (line, "Opening \"%s\"", fs);
        perror (line);
/* The error message now prints on stderr. */
//    } // else strcpy (LastXopen, fs);
    } else  setfnt (fileno(handle), fs);
    return  handle;
}

FILE*	iopenx (char *fs, char *tp, namlist *exts)
/* Open file fs with I/O type tp.  If open fails, try adding extensions
  from all elements in queue exts in turn.  If still failed, generate
  the error message.  Return opened handle, or NULL if failed. */
{
FILE    *handle;
char    line[120];
char	*bigbuf;
namlist	*a;
int	l;

/* Try the bare thing first */
    handle = fopen (fs, tp);
    if (handle != NULL) {
	strcpy (LastXopen, fs);    // Save the open name
	setfnt (fileno(handle), fs);
	return  handle;
    }

/* OK, we try any given extensions... */
    if (exts != NULL) {
	l = strlen (fs);
	bigbuf = (char *) malloc (2*l + 80);  /* Guessing that's good */
/* Now, loop around the exts list, see if adding ext to fs helps. */
	a = exts;
	while (a != NULL) {
	    strcpy (bigbuf, fs);
	    if (a->name != NULL) {	/* Added protection here */
		strcat (bigbuf, a->name);
		handle = fopen (bigbuf, tp);
		if (handle != NULL) break;
	    }
	    a = a->next;
	    if (a == exts) break;
	}	/* Loops over all extensions listed */
/* At this point, if handle is not null, we could print a line
  giving the name of the file found */
	if (handle != NULL) {
	    printf ("  Opened %s mode \"%s\".\n", bigbuf, tp);
/* May want to fprintf that to stderr... */
	    strcpy (LastXopen, bigbuf);    // Save actual open name
	}
/* Drop the dynamic buffer now that we're done */
	free (bigbuf);
    }  /* Checking non-null exts list */

    if (handle == NULL)  {
        strcpy (line, "Opening \"");
        strcat (line, fs);
        strcat (line, "\"");
        perror (line);
/* The error message now prints on stderr. */
    } else  setfnt (fileno(handle), LastXopen);
    return  handle;
}

void	iclose (FILE* f)
/* Close the file if any, maintain the name table */
{
    if (f == NULL) return;
/*    if (dbug >= 4) printf (" ..iclose: close %d.\n",  fileno(f));  */
    setfnt (fileno(f), NULL);
    fclose (f);
}


/*  Finding of file.  Given a filename and a series of possible locations,
  including literals and environment names, find the proper file name and 
  return it as a dynamic string.  */

// /* The "my_access" and "checkfile" routines are local only.  Not to
//   be put into function prototypes section of ioutils.h  */

// int	my_access (char* path)
// /* Return bits 1 for exist, 2 for readable */
// {
// int	r=0;	// return
// int	k;
//     k = access (path, F_OK);
//     if (k == 0) {	// file exists
// 	r |= 1;
// 	k = access (path, R_OK);
// 	if (k == 0) r |= 2;
// 	return  r;
//     } else  return  0;
// }

// int	checkfile (char* spec, char* name, char* dir)
// /* Return 1 if exists, 2 if readable;
//   Put concatenated name into spec;
//   dir may be environment variable, check both ways.   */
// {
// int	k=0;	// return value
// char	*tev;	// translated environment variable
// 
// /* Try first directory/name */
//     sprintf (spec, "%s/%s", dir, name);
//     k = my_access (spec);
//     if (k & 1) return k;
// // printf (" -- Did not find \"%s\".\n", spec);		//debug
// 
// /* If not found, try directoryname without / */
//     sprintf (spec, "%s%s", dir, name);
//     k = my_access (spec);
//     if (k & 1) return k;
// 
// /* Now, try env translation of dir */
//     tev = getenv (dir);
//     if (tev != NULL) {
// /* Try it with / and again without */
// 	sprintf (spec, "%s/%s", tev, name);
// 	k = my_access (spec);
// 	if (k & 1) return k;
// // printf (" -- Did not find \"%s\".\n", spec);		//debug
// 	sprintf (spec, "%s%s", tev, name);
// 	k = my_access (spec);
// 	if (k & 1) return k;
//     }
// 
// /* If nothing found, we need to return a zero, and blank spec */
//     *spec = AN;
//     return  0;
// }


// char*	findrfile (char* fname, namlist* nl)
// /* Return dynamic string containing the readable filespec; nl is list
//   of locations to search.  NULL returned if not found. */
// /* NOTE - only returns files which exist AND are readable.  If writable
//   file is desired, write another similar program for that! */
// {
// char	spec[256];	// Constructed file spec
// int	k;		// Status of search
// char*	result;
// namlist	*p;
// 
// /* Check all those null cases... */
//     if (fname == NULL) return NULL;
// 
// /* Look at all the elements of nl, return any one found. */
//     for (p=nl; p != NULL; p = p->next) {
// 	if (checkfile (spec, fname, p->name) == 3) goto foundfile;
// 	if (p->next == nl) break;
//     }
// 
// /* Look also in the local directory and parent directory */
//     if (checkfile (spec, fname, "." ) == 3) goto foundfile;
//     if (checkfile (spec, fname, "..") == 3) goto foundfile;
// /* Check also bare file name, might have path already */
//     if (checkfile (spec, fname, ""  ) == 3) goto foundfile;
// 
// /* Fall through = not found; return a null value. */
//     return  NULL;
// 
// /*   foundfile  --  Here we return the dynamic string.  */
// foundfile:
//     result = dynamstr (spec);
//     return  result;
// }


/*  Finding of File, revised and generalized, Mark III */


void	findfile (char** spec, char* mode, char* name, char* dir)
/* Returns result in *spec as a dynamic string. */
/* Looks for file of name in dir, possibly translating these as
  environment variables.  The mode requires existance and if it
  contains r=>readable, w=>writable, e=>executable. */
/* Once spec is not NULL, subsequent calls give no action. */
{
char	buf[256];
char	*tev;
static char *tp=NULL;
static char *cf=NULL;
int	m;
int	nc;

/* Look for a current result, return if so. */
    if (*spec != NULL) return;

/* Without a name, we can't do a thing. */
    if (name == NULL) return;

/* We need to look.  Find the proper access argument... */
    m = 0;
    if (strchr (mode, 'r')) m |= R_OK;
    if (strchr (mode, 'w')) m |= W_OK;
    if (strchr (mode, 'e')) m |= X_OK;
    if (m == 0) m = F_OK;
/* When we call access(name, m) we use the "!" negation in the
  test, as access returns 0 for success... */

/* See if the dir shows we have a null case, special... */
    nc = (dir == NULL);
    if (!nc) nc = (strlen(dir) < 1);
    if (nc) {		// Null directory; look locally
	sprintf (buf, "./%s", name);		// ./name
	if (!access(buf,m)) goto foundfile;
	sprintf (buf, "%s", name);		// name
	if (!access(buf,m)) goto foundfile;
	sprintf (buf, "../%s", name);		// ../name
	if (!access(buf,m)) goto foundfile;
	sprintf (buf, "../../%s", name);	// ../../name
	if (!access(buf,m)) goto foundfile;
	sprintf (buf, "/%s", name);		// /name
	if (!access(buf,m)) goto foundfile;
	tev = getenv (name);
	if (tev != NULL) strcpy (buf, tev);	// getenv(name)
	if (!access(buf,m)) goto foundfile;
    } else {		// Good directory string
//	sprintf (buf, "%s/%s", dir, name);	// dir/name
//	if (!access(buf,m)) goto foundfile;
//	sprintf (buf, "%s%s", dir, name);	// dirname
//	if (!access(buf,m)) goto foundfile;
	cf = fncat (dir, name);
	if (!access(cf,m)) goto foundfile2;
	tev = getenv (dir);
	if (tev != NULL) {
/* New Feature!  In a loop, parse tev into tokens with delimiters of
  ":" or " " and try each one as a directory.  This allows the EV to
  be a PATH formatted value rather than just a directory.
  Uses the "dparse" function defined below.   */
char	*tx;	// scratch, local
// dparse (line *tokn delim)
//	    sprintf (buf, "%s/%s", tev, name);	// tev/name
//	    if (!access(buf,m)) goto foundfile;
//	    sprintf (buf, "%s%s", tev, name);	// tevname
//	    if (!access(buf,m)) goto foundfile;
	    for (tx=tev; ; ) {
		tx = dparse (tx, &tp, ": ");
		if (tp == NULL) break;
		if (strlen(tp) < 1) break;
		cf = fncat (tp, name);
		if (!access(cf,m)) goto foundfile2;
	    }
	    dskill (&tp);
	}
	dskill (&cf);
    }
/* Well, if we didn't get to go to foundfile, we haven't found the
  file, so we just return... */
    return;

	/* - - - - - - *\
	|   foundfile   |
	\* - - - - - - */
foundfile2:
/* The cf pointer has a dynamic good file specification, so that
  is returned, and cf itself is set to NULL.  */
    *spec = cf;
    cf = NULL;
    dskill (&tp);
    return;

foundfile:
/* The buffer contains a good file specification; create a dynamic
  string containing it, and return its pointer. */
    *spec = dynamstr (buf);
    dskill (&tp);
    return;
}



void	evtran (char* *nam)
/* Translate name into Environment Variable if possible.
  Assume *nam is a dynamic string.  If an EV translation is found,
  *nam is freed and replaced with a dynamic copy of the translation.   */
{
char	*e=NULL;
    if ( nam == NULL) return;
    if (*nam == NULL) return;
    if ( (e = getenv (*nam)) != NULL) {
	free (*nam);
	*nam = strdup (e);
    }
}

/* -- The "locfile" subroutine will go here when done.  -- */
/*  Will include
  static void shownamlist () debug routine,
  static void evtran () translate environment name if possible - replace
  by a routine taking dynamic string and replacing it with dynamic
  translation if that is possible.
  char*  locfile ()  */


/*  ====  Parsing Routines  ====  */

static  void	dskill (char* *a)
/* This is a storage management utility.  Several programs, starting with
  dparse and dsexig take as an argument a pointer to a string pointer, and
  the pointer *a should be either NULL or point to a dynamic storage.  It
  is also possible that a itself is NULL, indicating no argument at all.
  When used, the program will need to free the storage at *a, and be able
  to allocate new storage.  To this end, this utility will determine if
  the parameter exists, if storage exists, and properly free it when
  needed, and clear the pointer.  NOTE that if *a points to storage which
  is already free, to static storage, or random things, this will cause
  a crash, and none of these problems can be tested for at execution time.
  When used, the dynamic pointer should be initialized to NULL, and then
  managed by this program.  When free, the pointer should be NULL.    */
{
    if (a != NULL) {
	if (*a != NULL) {
	    free (*a);
	    *a = NULL;
	}
    }
}

char*   skipover (char* line, char* delim)
/* Return line pointer advanced over characters in the delim set. */
{
char*   c;
    if (line == NULL) return line;
    for (c=line; strchr (delim, *c) && *c != AN; c++) ;
    return  c;
/* The alternate implementation: replace everything with:
    return  line + (line == NULL) ? 0 : strspn (line, delim);
-- which tests for NULL but needs valid delim...    */
}

/* Non-destructive parsing.  Given a pointer (within) a line, copy the
  delimited token to a (provided) buffer, and return the advanced
  pointer to the "rest" of the line.  Handle nulls and delimited
  strings intelegently. */
char*   parse (char* line, char* token, char* delim)
/* line points to string data; copy the first token delimited by
  characters from the delim string to the token buffer, and return
  a pointer to the remainder of the line.  */
{
char    *c;
int     j;

/* If token is not null, set its first character to nul */
    if (token != NULL) *token = AN;
/* If line is null, return a null */
    if (line == NULL) return line;
/* Advance line pointer while it points to delimiter characters. */
    line = skipover (line, delim);
/* If line points to nul, return a NULL pointer. */
    if (*line == AN) return NULL;
/* Find the next whitespace if any */
    c = strpbrk (line, delim);
/* If no whitespace left, we just copy to token, and return a
pointer to the terminating nul */
    if (c == NULL) {            // no whitespace in line
        j = strlen (line);
        c = line + j;
    } else {            // token and rest present
        j = c - line;   // length of token
        c = skipover (c, delim);
    }
/* Copy the stuff from line to token */
    strncpy (token, line, j);
    token[j] = AN;      // adds a nul at end
    return  c;
}

/* NOTE that "parse" above advances the line pointer over any delimiter
  characters which follow the token, while the "getok" below does not
  advance the line pointer, leaving it pointing to the next character
  following the token found.  Some programs may assume the former
  action, which may be simulated by adding a call to "skipover" after
  the call to "getok" when needed.    */

char*   dparse (char* line, char* *token, char* delim)
/* line points to string data; copy the first token delimited by
  characters from the delim string to a dynamic token buffer, and return
  a pointer to the remainder of the line.  If token pointer is NULL,
  just skip one token.    */
// xx-xx
/* NOTE that the *token pointer must be NULL or point to a valid dynamic
  storage address which will be freed.  If it points to dynamic storage
  which is already free, or to random things, (either of which cannot be
  detected by testing) the program will crash with an invalid pointer.
  Pointer used should be initialized to NULL, and may be previously assigned
  storage by this program.  If freed, the pointer should be set NULL.      */
{
char	*c = line;
int	j = 0;

/* If *line is NULL, we need to return a NULL */
    if ( line == NULL) return c;
/* If token is not NULL, free *token and make it point to NULL */
    dskill (token);
/* If at end of line, we return the end.  */
    if (*line == AN) return c;
/* Also need to return current place if no delim characters */
    if (delim == NULL) return c;
    if (*delim == AN) return c;
/* Advance *line pointer over any leading delimiter characters. */
    c += strspn (c, delim);	// new "skipover" implementation
/* If line points to nul, return a NULL pointer. */
    if (*c == AN) return c;
/* Find length of line to next delim (count non-delim chars) */
    j = strcspn (c, delim);
/* Get token as a malloc string of the proper length */
    if (j > 0) {
	if (token != NULL) {
//	    *token = malloc (j + 1);
	    *token = calloc (1, j + 1);  // cleared to zero on allocation
	    strncpy (*token, c, j);
//	    (*token)[j] = AN;	// redundant?  NO, not redundant, required.
/* There are (at least) two ways to do this.  One is to explicitly place
  the null terminator character at the end of the n characters copied, 
  while the other is to allocate cleared memory.    */
	}
//	if (token != NULL) *token = strndup (c, j);
// strndup is a GNU feature.  WATCH OUT IN ioutils.c
	c += j;
    } else if (j == 0) {
/* Special feature -- instead of *token being NULL, we allocate
  a single character buffer with a null character in it.  This is
  needed by some uses which formerly assumed a static buffer, and
  do not test for NULL, but test content only.    */
	if (token != NULL) *token = calloc (1, 1);
    }
    c += strspn (c, delim);  // skip trailing delimiters too
    return  c;
}


/*  Advanced non-destructive parsing.  Given a *pointer (within) a line,
  return a dynamic string copy of the delimited token, and advance the
  *pointer to the "rest" of the line.  Handle null cases intelegently. */
char*	getok (char** line, char* delim)
/*  Non-destructive parsing.  Return a dynamic string containing the first
  token delimited by characters in the delim string; and advance the *line
  pointer to the character following the token.  Content of line undisturbed. */
{
char	*r = NULL;
int	j = 0;

/* If *line is NULL, we need to return a NULL */
    if ( line == NULL) return r;
    if (*line == NULL) return r;
/* Also need to return NULL if no delim characters */
    if (delim == NULL) return r;
    if (*delim == AN) return r;
/* Advance *line pointer over any leading delimiter characters. */
    *line += strspn (*line, delim);	// new "skipover" implementation
/* If line points to nul, return a NULL pointer. */
    if (**line == AN) return r;
/* Find length of *line to next delim (count non-delim chars) */
    j = strcspn (*line, delim);
/* Get result as a malloc string of the proper length */
    if (j > 0) {
//	r = malloc (j + 1);
//	strncpy (r, *line, j);
//	r[j] = AN;  // redundant
	r = kstrndup (*line, j);
/* Could use:  r = strndup (*line, j); to replace malloc, strncpy */
	*line += j;
    }
    return  r;
}


int	itok (char* *line, char* delim)
/* Non-destructive parsing.  Return integer value of next
  token in *line, advancing the *line pointer.  */
{
char	*b=NULL;
int	r=0;

//    b = getok (line, delim);
// Solve "skip at end" problem here
    *line = dparse (*line, &b, delim);
    if (b == NULL) return  r;
    r = atoi (b);
//    free (b);
    dskill (&b);
    return  r;
}

double	ftok (char* *line, char* delim)
/* Non-destructive parsing.  Return double precision value of next
  token in *line, advancing the *line pointer.  */
{
char	*b=NULL;
double	r=0.0;

//    b = getok (line, delim);
// Solve "skip at end" problem here
    *line = dparse (*line, &b, delim);
    if (b == NULL) return  r;
    r = atof (b);
//    free (b);
    dskill (&b);
    return  r;
}

double	postok (char* *line, char* delim)
/* Non-destructive parsing.  Return parsed value of next position token
  in *line, advancing *line pointer.  */
{
char	*b=NULL;
double	r=0.0;

//    b = getok (line, delim);
// Solve "skip at end" problem here
    *line = dparse (*line, &b, delim);
    if (b == NULL) return  r;
    r = posvalue (b);
//    free (b);
    dskill (&b);
    return  r;
}

long	ltok (char* *line, char* delim)
/* Non-destructive parsing.  Return long integer value of next
  token in *line, advancing the *line pointer.  */
{
char	*b=NULL;
int	k;
long	r=0;

//    b = getok (line, delim);
// Solve "skip at end" problem here
    *line = dparse (*line, &b, delim);
    if (b == NULL) return  r;
    k = sscanf (b, "%ld", &r);
    if (k != 1) r = 0;
//    free (b);
    dskill (&b);
    return  r;
}


/*  New feature coming soon!   (currently in mgutils.c)
  The new   char*  extok (char* *line, char* fredel, char* specdel)
  function will be coming here.  Characters in fredel may be duplicated,
  but any character in specdel will terminate a token, and can indicate
  a null token.  Typically fredel is whitespace and specdel is a comma.
  The returned dynamic token can be null, and line is advanced past
  the tokan and any following specdel character.  When specdel contains
  characters also in fredel, the skipping of fredel occurs first.
  This simulates "parse" operation when specdel is null.    */

void	xcat (char** a, char* b)
/* Concatenate *a to b, free *a and put dynamic result in *a.
  Intended use is like this:
    xcat (&a, "first part of string");
    xcat (&a, ", another part");
    xcat (&a, ", final part.");      */
{
char	*c;
    c = dyncat (*a, b);
/* If we have a problem with null strings, use dynamstr in dyncat
  in place of strdup.  Should not be a problem, however...    */
    if (*a) free (*a);
    *a = c;
}

char*	extok (char** line, char* fredel, char* specdel)
/* Extended parsing.  Operates similarly to getok, but with two
  delimiter strings.  The free delimiters may be duplicated, but
  the specific delimiters will always terminate a token, and can
  indicate a null token.  Usually fredel is whitespace and specdel
  is something like a comma.  Returns a dynamic string with the
  token found (can be null) and advances *line appropriately.    */
/* This represents a generalization of getok to use two differently
  processed sets of delimiters, and to allow for null tokens.
  Either delimeter set, but not both, may be null.      */
{
char    *r = NULL;
char	*any = NULL;
int     j = 0;
//int	k = 0;
int	fd = (fredel != NULL);
int	sd = (specdel != NULL);

/* If *line is NULL, we need to return a NULL */
    if ( line == NULL) return r;
    if (*line == NULL) return r;
/* Also need to return NULL if no delim characters */
/* ? Should be using = not += here ?   YES - use = here. */
    if (fd) fd = (*fredel == AN) ? 0 : 1;
    if (sd) sd = (*specdel == AN) ? 0 : 1;
    if ( (fd+sd) < 1) return r;
/* Advance *line pointer over any leading free delimiter characters. */
//    if (fd > 1) *line += strspn (*line, fredel);
    if (fd) *line += strspn (*line, fredel);
/* If line points to nul, return a NULL pointer. */
    if (**line == AN) return r;
/* Generate an any delimiter string */
    any = dyncat (specdel, fredel);
    if (any == NULL) return r;
    if (*any == AN) {
	free (any);
	return r;
    }
/* Find length of *line to next delim (count non-delim chars) */
    j = strcspn (*line, any);	// can be zero
/* Get result as a malloc string of the proper length */
    if (j > 0) {
//	r = malloc (j + 1);
//	strncpy (r, *line, j);
//	r[j] = AN;  // redundant
	r = kstrndup (*line, j);
/* Could use:  r = strndup (*line, j); to replace malloc, strncpy */
	*line += j;
    }

/* Fix up the *line pointer with trailing delimiters */
/* Advance over any free delimiters if present. */
    if (fd) *line += strspn (*line, fredel);
/* Advance over exactly one specific delimiter if present */
    if (sd) *line += (strspn (*line, specdel) > 0) ? 1 : 0;
/* Could also use test on specdel containing **line and if true,
  do a (*line)++ operation; i.e. do this:
    if (strchr (specdel, **line) ) ++*line;
  Another way is to use a blend of the last two methods:
    if (sd) *line += strchr (specdel, *line) ? 1 : 0;
  or the perhaps more understandable equivalent:
    if (sd) *line += (strchr (specdel, *line) == NULL) ? 0 : 1;  */
/* Note that, like getok, we don't pass over trailing free delimiters,
  except in the case of specdel being null.   */

/* Return that result we found */
    return  r;	// can be NULL
}


/*  Another new feature!   (currently in mgutils.c)
  A new   namlist*  cdlparse (char* line, char* fredel, char* specdel)
  function will completely parse line into a dynamic namlist of tokens
  as delimited by fredel and specdel (as used in extok).  The list can
  include null tokens.  This namlist can then be processed by further
  parsing logic and ultimately freed.    */

namlist*	cdlparse (char* line, char* fredel, char* specdel)
/*  Full parsing to namelist.  Using extok, parse all tokens in line,
  including possible null tokens, and return the token list as a
  namlist using dynamic storage.  If no tokens are found, this list
  may be null.  It will also be null if line is null, or if no
  delimiters are provided.
  fredel is a string of free delimiters (i.e. whitespace) and specdel
  is a string of special delimiters (i.e. comma).  See extok for how
  the delimiters are used.    */
/*  Execution changes value of pointer *line, but since it is a call
  by value that should not be a problem.  */
{
namlist	*r = NULL;
//namlist *p;
char	*c;
char	*ll = line;

/* Check for null cases */
    if (line == NULL) return r;
    if (fredel == NULL) {
	if (specdel == NULL) return r;
	if (*specdel == AN) return r;
    } else if (*fredel == AN) {
	if (specdel == NULL) return r;
	if (*specdel == AN) return r;
    }
/* The alternte is to compute it using fd and sd as done in
  extok above.  Probably same amount of code either way.  */

/* Parse items while line has data present */
    while (*ll != AN) {
	c = extok (&ll, fredel, specdel);

/* Would use p = get_nam (c) if it were not for DYNAMICNAMLIST which would
  cause an extra dynamic copy to be created, requiring free (c) also.
  Thus it would be:
	p = get_nam (c);
//#ifdef  DYNAMICNAMLIST
	free (c);
//#endif
  The problem is that a null c would not be entered, and the spec 
  requires a null token to be in the name list output.      */
//	r = push_name (p, r);
/* The alternate is
	r = pushtxt (c, r);
	free (c);
  as pushtxt always gets a new namlist and dynamic string
  to push onto queue r.  Then p is not needed, but we need
  to use c to free it and prevent a memory leak.
  The problem with this is that no entry gets made if the
  string c was null or empty, meaning that we cannot put
  in the null token which the specification requires.     */

/* Replace the above stuff with this, and drop p as not needed. */
	r = push_name (new_nam (c), r);
    }

/* return what we found */
    return  r;
}

/*  NEED a better and more consistent namlist subroutine set.  One
  to create a namelist for a given dynamic string, consistently using
  strdup() for any static text.  Another to put that into a queue.
  It would help if we could recognize dynamic storage pointers, but
  failing that we could use
  xx = strdup (a);  free (a);
  which will only free the dynamic strings, and hopefully just ignore
  the free attempt on a static string. 
  Our main problem is that there is no consistent treatment of the
  cases of dynamic/static strings being put in namlist, and no
  consistent treatment of null or empty namlist elements.      */


//static	int	kval (char* tok, char** v)
int	kval (char* tok, double* v)
/*  Parse tok for keyword=value syntax.  Return number of characters in
  keyword, zero if none.  Set *v to the value part, or tokn if no keyword. */
/* Candidate for ioutils.c when it is perfected. */
{
char*	e=NULL;
int	k=0;

    if (tok) {
	e = strrchr (tok, '=');	// maybe use strrchr ?
	k = e ? e - tok : 0;
	if (e) e++; else e = tok;
    }
    if (v) *v = atof (e);
    return  k;
}


/* Find file name */
char*	cfilename (char* fspec)
/* Return a local pointer to a copy of a filename, extracted from
  the fspec text. */
/* Mostly used to extract the program name from execution command */
/* NEED to change return to a dynamic storage buffer, and manage
  that correctly.  */
{
static	char	buf[32];
char	*c, *e;
int	n=0;

/* Take care of a few null cases */
//    if (fspec == NULL) return NULL;
    *buf = AN;
    if (fspec == NULL) return buf;

/* File name starts at start or following final / character */
    c = strrchr (fspec, '/');
    if (c == NULL) c = fspec;
    else c++;

/* File name ends at end or at first dot following start */
    e = strchr (c, '.');
    n = (e == NULL) ? strlen(c) : (int)(e-c);

/* Copy up to 32 characters to buf... */
    if (n > 31) n = 31;
    strncpy (buf, c, n);
    buf[n] = AN;	/* Supply the terminating null */

/* When changing buf to dynamic, do this:
  1) Change buf[32] to char *buf=NULL;
  2) do this:
    buf = calloc (1, n+1);
    strncpy (buf, c, n);
  3) At start, either return NULL buf, or allocate a single null byte for
    the null return case.
..*/
    return  buf;
}



size_t	szx (char* x, int k)
/* Return the number of characters to the last non-zero character in
  buffer x of length k.  Look from end for last non-null character.  */
{
    while ( k > 0 && x[k-1] == '\0' ) k--;
    return  (size_t) k;
}

char*	cpyan (char* dest, char* src)
/*  Copy alpha-numeric characters from src to dest, skip others.
  Return the dest pointer just as strcpy does.  */
{
char  *p = dest;
    if (dest == NULL) return p;
    if (src  == NULL) { *dest = '\0'; return p; }
    for (; *src; src++) if (isalnum(*src)) *dest++ = *src;
    *dest = '\0';
    return  p;
}


/*  ====  Floating Point Version Support  ====  */



char*	fvstr (float a)
/* Return a dynamic string with a formatted as a
  version value.  Leading spaces are dropped.
  Trailing zeros after the first two digits after
  the decimal point are dropped.  Up to 4 decimal
  digits are formatted.  Be sure to free result
  after use.    */
{
char	*b;
char	*p;
char	*c;
int	sl;

    b = calloc (1, 16);
    sprintf (b, "%.4f", a);

/* Locate decimal point and drop trailing zeros */
    p = strchr (b, '.');
    for (sl = strlen(p) - 1; sl > 2; sl--) {  // Length of .xx00
	if (p[sl] == '0') p[sl] = '\0';
	else  break;
    }

/* Move string toward *b if there are leading blanks */
    for (p=c=b; *p == ' '; p++) ;
    while (*p != '\0') {
	*(c++) = *(p++);
    }

/* The pointer b is the dynamic buffer to return */
    return  b;
}

char*	vstr (float a)
/* Return a (static) character string, with argument a formatted
  as a version value.  Leading spaces are dropped, trailing
  zeros after first 2 digits are dropped.  Up to 4 decimal
  digits are formatted.  */
{
static char b[16];
char    *p;
char    *c;
int     sl;

    sprintf (b, "%.4f", a);

/* Locate decimal point and drop trailing zeros */
    p = strchr (b, '.');
    for (sl = strlen(p) - 1; sl > 2; sl--) {  // Length of .xx00
        if (p[sl] == '0') p[sl] = '\0';
        else  break;
    }

/* Move string toward *b if there are leading blanks */
    for (p=c=b; *p == ' '; p++) ;
    while (*p != '\0') {
        *(c++) = *(p++);
    }

/* Return static buffer b, for immediate use only */
    return  b;
}


char*	combinev (float a, float b)
/* Return a dynamic string containing properly formatted
  version from a, with the decimal part of version b
  appended to it.  Compatability with old things here.  */
{
char	*c;	// Main and return buffer
char	*v;	// Temporary home of formatted b

    c = fvstr (a);	// get A.aa in a 16 char buffer.
    v = fvstr (b);	// get B.bb
    strcat (c, strchr (v, '.') );	// get A.aa.bb
    free (v);		// clean up, not needed
    return  c;
}



/*  ====  Parameter File Reading  ====  */

// typedef  void*  user_prog (char* , void* );
// above may be necessary to prototype the user programs.

// Alternate below is to use... char* filespec, user_prog* userprog, ...
#define  Gbufl  760
// Use multiple of 2*n less 8 byte end pad.  3*256 = 768
// Former Gbufl was 512
void*  get_data (char* filespec, void* (*userprog)(char*, void*),
	char* delim,	char* leadcom,
	char* anycom,	char* dxcom,
	char* xdcom,	char* dxdcom )
/* Filespec is the data file to read;
  userprog is a user program which will store the data,
  there are 6 comment character sets based on use:
  1) The delimiter list.  If null, strings 4,5,6 are ignored.  When
the data is read, leading and trailing delimiters are removed.
  2) The characters which, if present in the first position of a line
declare the entire line to be a comment and be ignored.
  3) The characters which introduce a comment if they appear anywhere
in the line, with or without delimiters.
  4) The characters which will start a comment only if they are preceeded
by any character from the delimiter set.
  5) The characters which will start a comment only if they are followed
by any character from the delimiter set.
  6) The characters which will start a comment only if both preceeded
and followed by any character from the delimiter set.
.. */
{
char	line[Gbufl+8];
FILE	*infile;
char	*p;
// int	i;
int	j;
int	mbl=0;	// maximum buffer (data) length
int	mdl=0;	// maximum uncommented data length
void	*V = NULL;
int	kl = 0;		// debug

/* Open, read the file.  For each valid line, call the user
  program to decode it into its own structure.  Return the
  pointer last given by the user program. */

/* Initial stuff.  See if we have a valid file, and open it.
  Return NULL if anything fails. */
    infile = iopen (filespec, "r");
    if (infile == NULL) return NULL;

//    infile = fopen (filespec, "r");
//// NOTE this is redundant with "iopen" in ioutils
//    if (infile == NULL) {
//	strcpy (line, "Opening \"");
//	strcat (line, filespec);
//	strcat (line, "\"");
//	perror (line);
//	return NULL;
//    }
//    strcpy (LastXopen, filespec);

/* Make sure end of line buffer is cleared */
    for (j=7; j > -2; j--) line[Gbufl+j] = '\0';

/* Clearing the last 8 bytes and more of the buffer, and limiting reads
  to Gbufl should prevent any buffer over-run conditions.  The line
  buffer should be safe to send to the user program.  We note that
  fgets "guarantees" to read at most one less than size characters, and
  to supply a null after what is read, however long that may be.  The
  longest buffer text length expected is thus Gbufl-1.   */

/* Loop on reading lines from the file. */
    while (fgets (line, Gbufl, infile) ) {
// Increased buffer size 2007/3-8 for GISMO transform data
// Increased buffer and added padding 2010/8-20 for safety

/* Track maximum buffer data length for debug */
	j = strlen (line);
	if (j > mbl) mbl = j;

/* Remove commentary -- this is the main thing this program does. */
	if (*line == AN) continue;	// Kill blank lines
	if (leadcom != NULL)	// Look for whole line comments
	    if (strchr(leadcom, (int)*line) ) continue;

	if ((p = strpbrk (line, anycom))) {	// unconditional comment
	    *p = AN;
	}


/* Comments preceeded by delimiter */
	for (p=line; p != NULL && *p != AN; p++ ) {
	    if ((p = strpbrk (p, dxcom))) {
		j = p - line;
		if (j < 1) continue;
		if (strchr(delim, line[j-1])) { *p = AN; break; }
	    } else break;
	}

/* Comments followed by delimiter */
	for (p=line; p != NULL && *p != AN; p++ ) {
	    if ((p = strpbrk (p, xdcom))) {
		j = p - line;
		if (line[j+1] == AN) continue;
		if (strchr(delim, line[j+1])) { *p = AN; break; }
	    } else break;
	}

/* Comments between delimiters */
	for (p=line; p != NULL && *p != AN; p++ ) {
	    if ((p = strpbrk (p, dxdcom))) {
		j = p - line;
		if (j < 1) continue;
		if (line[j+1] == AN) continue;
		if (strchr(delim, line[j-1]) &&
		    strchr(delim, line[j+1])) { *p = AN; break; }
	    } else break;
	}

/* Remove trailing cr, lf, delimiter characters if any. */
	j = strlen (line);
	while (j > 0) {
	    if (strchr ("\r\n", line[j-1]) ||
		strchr (delim,  line[j-1]) ) {
		line[j-1] = AN;
		j--;
	    }  else  break;
	}

/* Remove any leading delimiters */
	p = line;
	while ( *p != AN && strchr (delim, *p) ) p++;

	if (*p == AN) continue;	// Skip whole thing if nothing left

/* Track maximum data length provided */
	j = strlen (p);
	if (j > mdl) mdl = j;

/* Call the user program with the remaining non-comment line */
// pointer p points to de-commented line in buffer...
	V = (*userprog) (p, V);
	kl++;	// debug
    }	// End of main loop
    iclose (infile);	// Close up input

//	printf (" ( Read %d lines.)\n", kl);	//debug

/* Special call to user program with line reporting maximum buffer sizes. */
    sprintf (line, "MAX_BUFFER %d %d", mdl, mbl);
// mdl is maximum data line provided to userprog
// mbl is maximum buffer read size found in file including comments
    V = (*userprog) (line, V);

/* Final call to user program with NULL line for its final pointer. */
    V = (*userprog) (NULL, V);

/* Return the final pointer from user program */
    return  V;
}


/*  ====  Sexigisimal decode/encode  ====  */

double	posvalue (char*  field)
/* Recursively decode a position value field into degrees (hours)
  and optionally : minutes ... */
{
double	a;
int	k;
// int	j;
char	*fc;

/* Take any null cases out quickly at start */
    if (field == NULL) return 0.0;
    if (*field == '\0') return 0.0;

/* Note that this algorithm can be fooled by such errors as an embeded
 minus sign, or extra decimal points.  Also, non-numeric values
 get treated as zero.  All these should be obvious errors. */

/* Recursively evaluate any minus sign by returning the negative of
  what we return for all characters following that minus sign. */
    if (*field == '-')	/* Recursive minus sign feature */
	return -posvalue (field+1);

/* Evaluate floating point number before any colon, and add 1/60 of
  any recursively obtained value after a colon. */

    fc = strchr (field, ':');	/* Look for colon */
    if (fc == NULL) {		/* No colon, decode directly */
	k = sscanf (field, "%lf", &a);
	if (k != 1)  a = 0.0;
    } else {			/* Colon, split into two values */
	k = sscanf (field, "%lf:", &a);
	if (k != 1) a = 0.0;
	a += posvalue (fc+1) / 60.0;
    }
    return  a;
}

int	intvalue (char*  field)
/* Decode an integer value field */
{
int	k;
int	d;
    if (field == NULL) return 0;
    k = sscanf (field, "%d", &d);
    if (k != 1) d = 0;
    return  d;
}

long	longval (char*  field)
/* Decode a long integer value field */
{
int	k;
long	d;
    if (field == NULL) return 0;
    k = sscanf (field, "%ld", &d);
    if (k != 1) d = 0;
    return  d;
}

double	floatvalue (char*  field)
/* Decode a floating value field */
{
int	k;
double	d;
    if (field == NULL) return 0.0;
    k = sscanf (field, "%lf", &d);
    if (k != 1) d = 0.0;
    return  d;
}

vect2	vect2value (char* field)
/* Decode a vector value field; 1 or 2 floating values,
  if only 1 present, the second is same as first. */
/* An alternate is to have second be 0.0 if not present,
  for that, see complxvalue function.  */
{
int	k;
vect2	v={0.0,0.0};
    if (field == NULL) return v;
    k = sscanf (field, "%lf %lf", &v.x, &v.y);
    if (k == 1) v.y = v.x;
    else if (k != 2) v = (vect2){0.0, 0.0};
    return  v;
}

vect2	complxvalue (char* field)
/* Decode a complex number field; 1 or 2 floating values,
  if only 1 present, the second is 0.0  */
{
int	k;
vect2	v={0.0,0.0};
    if (field == NULL) return v;
    k = sscanf (field, "%lf %lf", &v.x, &v.y);
    if (k == 1) v.y = 0.0;
    else if (k != 2) v = (vect2){0.0, 0.0};
    return  v;
}

vect3	vect3value (char* field)
/* Decode a vector value field; 1 to 3 floating values,
  any not present are set to zero.    */
{
int	k;
vect3	v={0.0, 0.0, 0.0};
    if (field == NULL) return v;
    k = sscanf (field, "%lf %lf %lf", &v.x, &v.y, &v.z);
    if (k < 3) v.z = 0.0;
    if (k < 2) v.y = 0.0;
    if (k < 1) v.x = 0.0;
    return  v;
}


/*  ===  Encoding data  ===  */


char*	decomp (int i, char s, int n, int r)
/* Decompose i using radix r for n places, using s as a separator */
/* Return is a dynamic string */
{
char	*result;
char	*top;
int	m;

    if (n > 1) {	/* Multi-period, recursion needed */
	m = i % r;
	top = decomp (i/r, s, n-1, r);
	result = (char*)malloc(strlen(top)+12);
	sprintf (result, "%s%c%02d", top, s, m);
	free (top);
    } else {		/* Single period, return value */
	result = (char*)malloc(12);
	sprintf (result, "%02d%c", i, AN);
    }
    return  result;
}


/*  IMPLEMENTATION NOTE:  Make "dsexig" (below) the primary program,
  and have "sexig" and "sexigw" call it for their buffer implementations.
  Use a NULL value of buf when doing this, and free the returned
  dynamic buffer after copying it to local or provided storage.    */

char*	dsexig (double val, int np, int nd, char* *buf)
/* Similar to sexig, but returns the address of the character
  string in a dynamic buffer.  NOTE that the *buf pointer must
  be NULL or point to a valid dynamic storage address which will 
  be freed.  If it points to dynamic storage which is already free, 
  or to random things, (either of which cannot be detected by 
  testing) the program will crash with an invalid pointer.  
  Pointer used should be initialized to NULL, and may be 
  previously assigned storage by this program.  When freed, 
  the pointer should be set NULL.
  With a provided *buf, this program may be used in a printf
  argument string multiple times.  Start *buf as NULL and be
  sure to free it at the program end.      */
/* If buf itself is NULL, just return the dynamic buffer used, and
  let the calling program manage it.  */
// xx-xx

/* Translate val into sexigisimal quantity with np periods and nd
  decimal points in fraction.  Usually np = 3 and nd is small */
/* New feature!  If np is negative, write with a + sign in front
  when the value is positive.  */
{
char	*ip;
char	*fp;
char	*result;
double	e;
int	i;
int	neg;
int	ps;

/* If *buf is not NULL, free *buf and make it point to NULL */
    dskill (buf);

/* Find the sign, store it.  Find the fraction as needed and
store that too.  Get the integer parts in periods using subroutine.
Finally, combine sign, integer and fraction.  */

/* Get the sign and absolute value of val */
    if (val < 0.0) { neg = 1;  val = -val; }
    else	     neg = 0;
    if (np < 0) { ps = 1;   np = -np; }
    else	  ps = 0;

/* Change radix by np periods of 60 */
    for (i=1; i<np; i++) val *= 60.0;

/* Get the fraction string, and integer part. */
    if (nd > 0) {
	if (nd > 12) nd = 12;
	e = 0.5 * exp (-nd * log(10.0));
	val += e;
	i = (int) val;
	val -= (double) i;
	if (val > e) val -= e;
	fp = (char*)malloc(nd+4);
	sprintf (fp, "%*.*f%c", nd+1, nd, val, AN);
    } else {
	fp = (char*)malloc(4);
	*fp = '\0';
	i = (int) (val + 0.5);
    }

/* Get the integer part from recursive subroutine */
    ip = decomp (i, ':', np, 60);

/* Find the decimal point, if any (often, fp will have the format
of 0.xx rather than .xx as desired.) */
    for (i=0; i<4; i++) if (fp[i] == '.' || fp[i] == '\0') break;

/* Obtain dynamic result buffer of needed length,
  and set *buf if possible.  */
    result = malloc (strlen(ip)+strlen(fp)+3-i);
    if (buf != NULL) *buf = result;

/* Write the result into the user buffer */
    if     (neg) sprintf (result, "-%s%s", ip, fp+i);
    else if (ps) sprintf (result, "+%s%s", ip, fp+i);
    else         sprintf (result,  "%s%s", ip, fp+i);
    free (ip);
    free (fp);
    return  result;
}

char*	sexig (double val, int np, int nd)
/* Translate val into sexigisimal quantity with np periods and nd
  decimal points in fraction.  Usually np = 2 and nd is small */
/* Result is a static string here */
/* Note:  Since result is a static string, it needs to be fully used
BEFORE this program is called again.  When printing, if this is called
to format a string, only one such call may be done within a single
printf argument string. */
{
char	*buf;
static	char	result[32];

    buf = dsexig (val, np, nd, NULL);
    strncpy (result, buf, 31);
    free (buf);
    return  result;
}

void	sexigw (char* buf, double val, int np, int nd)
/* Similar to sexig, but avoids the use of a local buffer by
  writing the result into a user supplied character buffer;
  which must be of sufficient length... */
{
char	*dd;

    dd = dsexig (val, np, nd, NULL);
    strcpy (buf, dd);  // Hope it's long enough...
    free (dd);
}

/* Below are the original sexig and sexigw programs, replaced by the
  dsexig dynamic system and the stub implementationa above.  These are
  kept for documentation only.    */


#ifdef  NOT_Defined
//static char*	sexig (double val, int np, int nd)
/* Translate val into sexigisimal quantity with np periods and nd
  decimal points in fraction.  Usually np = 2 and nd is small */
/* Result is a static string here */
/* Note:  Since result is a static string, it needs to be fully used
BEFORE this program is called again.  When printing, if this is called
to format a string, only one such call may be done within a single
printf argument string. */
{
static	char	result[32];
char	*ip;
char	*fp;
double	e;
int	i;
int	neg;

/* Find the sign, store it.  Find the fraction as needed and
store that too.  Get the integer parts in periods using subroutine.
Finally, combine sign, integer and fraction.  */

/* Get the sign and absolute value of val */
    if (val < 0.0) {
	neg = 1;
	val = -val;
    } else  neg = 0;

/* Change radix by np periods of 60 */
    for (i=1; i<np; i++) val *= 60.0;

/* Get the fraction string, and integer part. */
    if (nd > 0) {
	if (nd > 12) nd = 12;
	e = 0.5 * exp (-nd * log(10.0));
	val += e;
	i = (int) val;
	val -= (double) i;
	if (val > e) val -= e;
	fp = (char*)malloc(nd+4);
	sprintf (fp, "%*.*f%c", nd+1, nd, val, AN);
    } else {
	fp = (char*)malloc(4);
	*fp = '\0';
	i = (int) (val + 0.5);
    }

/* Get the integer part from recursive subroutine */
    ip = decomp (i, ':', np, 60);

/* Find the decimal point, if any (often, fp will have the format
of 0.xx rather than .xx as desired.) */
    for (i=0; i<nd; i++) if (fp[i] == '.' || fp[i] == '\0') break;
/* Should be i < 3 above??? */

/* Combine integer parts and fraction parts */
    sprintf (result, "-%s%s%c", ip, fp+i, AN);
    free (ip);
    free (fp);
    if (neg) return result;
    return result+1;
}

//void	sexigw (char* buf, double val, int np, int nd)
/* Similar to sexig, but avoids the use of a local buffer by
  writing the result into a user supplied character buffer;
  which must be of sufficient length... */
/* Translate val into sexigisimal quantity with np periods and nd
  decimal points in fraction.  Usually np = 3 and nd is small */
/* New feature!  If np is negative, write with a + sign in front
  when the value is positive.  */
/* Note:  Since result is a static string, it needs to be fully used
BEFORE this program is called again.  When printing, if this is called
to format a string, only one such call may be done within a single
printf argument string. */
{
//  static	char	result[32];
char	*ip;
char	*fp;
double	e;
int	i;
int	neg;
int	ps;

/* Find the sign, store it.  Find the fraction as needed and
store that too.  Get the integer parts in periods using subroutine.
Finally, combine sign, integer and fraction.  */

/* Get the sign and absolute value of val */
    if (val < 0.0) {
	neg = 1;
	val = -val;
    } else  neg = 0;
    if (np < 0) {
	ps = 1;
	np = -np;
    } else ps = 0;

/* Change radix by np periods of 60 */
    for (i=1; i<np; i++) val *= 60.0;

/* Get the fraction string, and integer part. */
    if (nd > 0) {
	if (nd > 12) nd = 12;
	e = 0.5 * exp (-nd * log(10.0));
	val += e;
	i = (int) val;
	val -= (double) i;
	if (val > e) val -= e;
	fp = (char*)malloc(nd+4);
	sprintf (fp, "%*.*f%c", nd+1, nd, val, AN);
    } else {
	fp = (char*)malloc(4);
	*fp = '\0';
	i = (int) (val + 0.5);
    }

/* Get the integer part from recursive subroutine */
    ip = decomp (i, ':', np, 60);

/* Find the decimal point, if any (often, fp will have the format
of 0.xx rather than .xx as desired.) */
    for (i=0; i<nd; i++) if (fp[i] == '.' || fp[i] == '\0') break;
/* Should be i < 3 above??? */

/* Combine integer parts and fraction parts */
//    sprintf (result, "-%s%s%c", ip, fp+i, AN);
//    free (ip);
//    free (fp);
//    if (neg) return result;
//    return result+1;

/* Write the result into the user buffer */
    if     (neg) sprintf (buf, "-%s%s", ip, fp+i);
    else if (ps) sprintf (buf, "+%s%s", ip, fp+i);
    else         sprintf (buf,  "%s%s", ip, fp+i);
    free (ip);
    free (fp);

}
#endif  // section removed by NOT_Defined

/*  ====  Date decode and encode  ====  */

/* These programs encode and decode an integer date for contemporary
  Gregorian calendar dates.  These may be related to an actual J.D.
  with additional time zone and time of day information, but are
  used here only for calendar date integer arithmetic.  */

int	nmonth (char* name)
/* Return the month number for the first 3 characters of the
  given text.  Return 0 for error inputs.  */
{
const	char	monam[12][4] = {
	"Jan", "Feb", "Mar", "Apr", "May", "Jun",
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
int	m;

    if (name == NULL) return 0;
    for (m=0; m<12; m++) {
	if (strncasecmp(monam[m], name, 3) == 0) return (m+1);
    }
    return  0;
}

/* The "qrem" function is defined in kdcutil.c and kdcutil.h, and the
  kdcutil.h is included by ioutils.h which is included here.  */

long int     mjdate (int y, int m, int d)
{
long int    j, c, jy, jm;
long int    p, q, ky;

//    i = m + 9;
//    jy = y + (i/12) - 1;
//    jm = i % 12;
    qrem (&ky, &jm, m+9, 12);
    ky += y - 1;
//    c = jy / 100;
//    jy -= 100 * c;
    qrem (&c, &jy, ky, 100);	/* Century, year */
    qrem (&p, &q, c, 4);	/* Gregorian period, sub-century */
// formerly: qrem (&p, &q, c, y); [wrong]
//    j = (146097*c)/4 + (1461*jy)/4 + (153*jm+2)/5 + d - 678882;
    j = 146097*p + 36524*q + (1461*jy)/4 + (153*jm+2)/5 + d - 678882;
    return  j;
}

long int mjd4cdate (char* date)
/* Convert the given date string, which is of format "mmm dd yyyy" as
  generated by the GCC compiler's __DATE__ symbol to a mjd.  */
{
int	j;
int	yr, mo, dy;
char	mfld[4];

    if (date == NULL) return 0;
    j = sscanf (date, "%3s %d %d", mfld, &dy, &yr);
    if (j != 3) return 0;
    mo = nmonth (mfld);
    return  mjdate (yr, mo, dy);
}

void    date4mjd (long int mjd, int* year, int* month, int* day)
/* Convert mjd to Gregorian year, month, day */
{
long int        j, y, d, m;
long int	p, q;
 
//    j = (long)(jd-0.5) - 1721118;     /* De-bias JD */
 
    j = mjd + 678882;   // Convert mjd
/* j is days since 29 Feb. year"0" (1 BC) */

// Updated for new code to handle negative years, etc.

//    y = (4*j-1)/146097;         /* Century */
//    j = 4*j - 1 - 146097*y;     /* Quarter days in century */
//    d = j/4;                    /* Days in century */
//    j = (4*d+3)/1461;           /* Year */
//    d = 4*d + 3 - 1461*j;       /* Quarter days in year */
//    d = (d+4)/4;                /* Days */
//    m = (5*d-3)/153;            /* Month */
//    d = 5*d - 3 - 153*m;        /* 1/5 days in month */
//    d = (d+5)/5;                /* Day in month */
//    y = 100*y + j;              /* Combine year, century */

    qrem (&p, &q, j, 146097);	/* Gregorian period and 4-centuries */
    qrem (&y, &j, 4*q-1, 146097);	/* Century and quarter-days */
    y += 4*p;			/* Add Gregorian centuries of p */
    d = j/4;			/* Days in century */
    qrem (&j, &d, 4*d+3, 1461);   /* Year and quarter days */
    d = (d+4)/4;		/* Days */
    qrem (&m, &d, 5*d-3, 153);   /* Month and 1/5 days in month */
    d = (d+5)/5;                /* Day in month */
    y = 100*y + j;              /* Combine year, century */

    qrem (&p, &q, m+2, 12);
    *year = y + p;		/* Correct year */
    *month = 1 + q;		/* Correct month */
    *day = d;                   /* Day of month */
    return;
}

long int mjd4tm (time_t* now)
/*  Return the local mjd for a system time value.  Uses the system call
  for localtime to get the date in current time zone.  Then converts
  to mjd using the mjdate function.  */
{
struct  tm  *tm;
int	year;
long int m;

    tm = localtime (now);	/* Local time in tm structure */
    year = tm->tm_year;
    if (year < 1700) year += 1900;
    if (year < 1970) year += 100;	// does not usually happen
    m = mjdate (year, tm->tm_mon + 1, tm->tm_mday);
    return  m;
}

int	weekofyear (long int mj)
{
int	yr, mo, dy;
long int nyd, mjs;
int	dow, wn;

/* Find week of year for given mjd */
/* mjd is 0 mod 7 on Wednesday */
/* Week 1 contains the first Thursday of the year */

// Find the mjd of jan 1 of same year
    date4mjd (mj, &yr, &mo, &dy);	// Get year
    nyd = mjdate (yr, 1, 1);		// mjd of Jan 1
//    dow = (nyd+3) % 7;	// Sunday = 0
    dow = mod (nyd+3, 7);   // Sunday = 0
/* dow is day of week, sunday = 0, for Jan 1. */

// find mjd of start of week 0 of year
    mjs = nyd - dow;
    if (dow < 5) mjs -= 7;	// 5 is Friday
/* mjs is mjd of Sunday starting week 0 */

// find week number
    wn = (mj - mjs) / 7;
// return that result.
    return wn;
}

//char*	dayandtime (void)
/*  Return a dynamic string containing the current time, in the
  format  yyyy-mm-dd,hh:mm  using the current time zone.  This is
  used to produce the date field for the release.    */
/*  See isotime above for a better way to do this.  */
/* NOTE THAT this result is not strictly dynamic (but it is not used) */
/* THIS PROGRAM IS OBSOLETE */
//{
////static long int  time_value;
//time_t  now;
//static  char  daybuf[24];
//struct  tm  *tm;
////int	year;
//
//    time (&now);		/* Current time */
//    tm = localtime (&now);	/* Local time in tm structure */
//
///*  We don't need to do all this...
//    year = tm->tm_year;
//    if (year < 1700) year += 1900;
//    if (year < 1970) year += 100;
//    sprintf (daybuf, "%4d-%02d-%02d,%02d:%02d",
//	year,
//	tm->tm_mon + 1,
//	tm->tm_mday,
//	tm->tm_hour,
//	tm->tm_min);
//... stuff, just use the strftime function  (see man strftime)  */
//
//    strftime (daybuf, 22, "%Y-%m-%d,%R", tm);	/* Format time to string */
//    return  daybuf;
//}

char*	dayntim (char* date, char* tfld)
/* Return a dynamic string with reformatted date and time in the
  format yyyy-mm-dd,hh:mm from a date in the form Mmm dd yyyy and hh:mm:ss
  as is returned by the GCC DATE and TIME defined macros.  */
{
int	j;
int	yr = 2007;
int	mo = 0;
int	dy = 22;
char	mf[4];
char	*result;

    result = malloc (20);
    j = sscanf (date, "%3s %d %d", mf, &dy, &yr);
    mo = nmonth (mf);
    sprintf (result, "%4d-%02d-%02d,", yr, mo, dy);
    strncat (result, tfld, 5);
    return  result;
}


/*  ====  Binary Data Debug Utilities  ====  */

void	hdstr (char* title, unsigned char* dat, int n)
/*  Dump to stdout the dat array of n bytes, in hex, identifying it by title */
{
int	i, j;
char	ttl[10];

    if (title == NULL)  strcpy (ttl, "-hdstr-");
    else { memset(ttl, ' ', 10); strncpy(ttl, title, 9); }
    ttl[8] = '\0';	// strncpy does not properly truncate

    if (dat == NULL) {
	fprintf (stdout, " %s: has no data.\n", ttl);
	return;
    } else {
	for (i=0; i < n; i += 32) {
	    if (i == 0) fprintf (stdout, " %8s: ", ttl);
	    else  fprintf (stdout, "           ");
	    for (j=i; j < n && j < i+32; j++)
		fprintf (stdout, "%02x", dat[j]);
	    fprintf (stdout, "\n");
	}
    }
}

void	cdstr (char* title, char* dat)
/*  Make a formatted output of dat, so as not to exceed line length,
  and identifying the output by title.  Output to stdout. */
{
int	i, j;
int	n;
char	ttl[12];

    if (title == NULL)  strcpy (ttl, "-hdstr-");
    else { memset(ttl, ' ', 12); strncpy(ttl, title, 11); }
    ttl[10] = '\0';

    if (dat == NULL) {
	fprintf (stdout, " %s: has no data.\n", ttl);
	return;
    } else {
	n = strlen (dat);
	for (i=0; i < n; i += 60) {
	    if (i == 0) fprintf (stdout, " %10s: ", ttl);
	    else  fprintf (stdout, "             ");
	    for (j=i; j < n && j < i+60; j++)
		fprintf (stdout, "%c", dat[j]);
	    fprintf (stdout, "\n");
	}
    }
}



/*  ====  Binary Encoding Utilities  ====  */


int	hex2bin (char* hx, unsigned char* bn, int n)
/* Convert n unsigned binary characters from hx to binary bn
  Return number of binary characters converted.  If fewer
  than n binary characters are present, only the ones
  present will be converted.  Count is characters converted
  correctly; incorrect characters are present as zero.  */
{
int	nr=0;
int	i, j, k;
char	td[4];

    if (bn == NULL) return 0;
    *bn = '\0';
    if (hx == NULL) return 0;
    if (n < 0) return 0;
    j = strlen (hx);
    if (j < 1) return 0;
    k = (j + 1) / 2;		// characters present
    if (n > k) n = k;		// characters to convert

    memset (td, 0, 4);
    for (i=0; i < n; i++) {
	memcpy (td, &(hx[2*i]), 2);
	j = sscanf (td, "%2x", &k);
	bn[i] = (j == 1) ? (unsigned char)k : 0;
	if (j == 1) nr++;	// count correct conversions
//	bn[i] = (j == 1) ? (k & 0xff) : 0;
    }
/* Alternate method:
	j = hx[2*i] - '0';
	if (j > 9) j -= 'A' - '9' - 1;
	if (j > 15) j -= 'a' - 'A';
	if (j < 0 || j > 15) j = 0;
	bn[i] = 16*j;
	j = hx[2*i+1] - '0';
	if (j > 9) j -= 'A' - '9' - 1;
	if (j > 15) j -= 'a' - 'A';
	if (j < 0 || j > 15) j = 0;
	bn[i] += j;
..*/
    return  nr;
}


int	hex4bin (char* hx, unsigned char* bn, int n)
/*  Convert n unsigned binary characters in bn to hex in hx;
  Return number of hex digits placed in hx.  */
{
int	i;

    if (hx == NULL) return 0;
    *hx = '\0';
    if (bn == NULL) return 0;
    if (n < 1) return 0;

    for (i=0; i < n; i++) {
	sprintf ( &(hx[2*i]), "%02x", bn[i] );
    }
    return  2*n;
}


//ar*	hextex (unsigned char* xx, int n)
char*	hextex (char* xx, int n)
/* Enter n bytes from xx into a local static buffer, which is then
  returned as the result.  The n should be no longer than about 40
  characters.  Result is null terminated text, suitable for output.
  This is intended to be a utility for debug output.    */
{
static  char  buf[256];
int	i;

    buf[0] = '\0';
    if (xx == NULL) return  buf;
    for (i=0; i<n && i<255; i++) {
//	sprintf ( &(buf[2*i]), "%02x", *xx++);
// Either one works, the following is better style...
	sprintf ( &(buf[2*i]), "%02x", xx[i] );
    }
    buf[2*i] = '\0';	// redundant
    return  buf;
}


char	sixbc (int c)
/* Return a special extended character encoding the low order 6 bits
  of integer c.  There are 10 digits 0-9, and 52 letters A-Z, a-z.
  Two more characters are '@' and '_' which are between the digits
  and letters.  An inverse appears below.  */
{
    c &= 63;
    if (c < 10) return '0'+c;
    c -= 10;
    if (c < 26) return 'A'+c;
    c -= 26;
    if (c == 0) return '_';
    c--;
    if (c < 26) return 'a'+c;
    return '@';
}


int	bcsix (char c)
/* Inverse of sixbc above; returns integer value producing the six-bit
  character c.  Returns -1 if c is not a valid character.  */
{
    if      (c <  '0') return -1;
    else if (c <= '9') return c - '0';
    else if (c == '@') return 63;
    else if (c <  'A') return -1;
    else if (c <= 'Z') return 10 + c - 'A';
    else if (c == '_') return 36;
    else if (c <  'a') return -1;
    else if (c <= 'z') return 37 + c - 'a';
    else  return -1;
}


/* The basic bit setting and finding operations involve the mechanism
  of finding two adjacent bytes in bn and two masks for the left
  and right such bytes.  The bits can fit in two adjacent bytes in
  any of 4 ways:
	11111100 00000000
	00111111 00000000
	00001111 11000000
	00000011 11110000
  And the next step is the same as the first, shifted one byte right.
  These actual cases come about as follows:
	11111100 00000000 00000000 00000000
	00000011 11110000 00000000 00000000
	00000000 00001111 11000000 00000000
	00000000 00000000 00111111 00000000
  And the next line is the same as the first, 3 bytes over.  So, for
  every 3 bytes and 4 6-bit fields, we have 4 ways of extracting the
  six bits from adjacent bytes.  These can be easily characterized by
  the starting byte and bit offset of the 6-bit field, and if the 
  offset is > 2 bits the field encroaches on the next byte.  The
  offset values go 0, 6, 4, 2, 0; each is 6 mod 8 greater than the
  previous.  Byte index and offset are easily computed by the simple
  expressions  (6*n)/8 and (6*n)%8.  Then the mask is found by
  shifting the basic 63 mask value left by (2-offset).  If this
  shift value is positive, apply it to the indexed byte only,
  and then right shift the result by shift.  If
  negative, right shift by -shift and apply to the indexed byte
  and then left shift by -shift for first part; then left shift
  by 8+shift, apply to next byte and right shift by the same for
  the second part, and or the two parts together.  When the offset
  is zero, just use the basic mask itself.  If the "next" byte is
  past the limit of n, use a zero byte and don't retrieve or store.
  To make an easier operation, one can get the bytes from the
  array into a long int by clear, two or operations with a
  shift between.  Then take the mask and shift, "and" then un-shift
  to retrieve the data.  For storing, clear, get existing data, shift
  and get next byte if necessary, then shift and "or" the data in,
  and store the result byte(s).    */


int	six2bin (char* sx, unsigned char* bn, int n)
/*  Convert n unsigned binary characters into bn from six-bit special
  character string at sx.  Total of 8*n bits are present, stored as
  6 bits in each character of sx by encoding.
  Return is actual number of binary characters found;  if more exist
  in sx, only 8*n bits are transfered, if fewer exist, the number
  found are returned.  If last binary character is exactly zero and
  byte count is odd, the count could be reduced by one.  Return count
  is a count of correctly converted binary characters.  */
{
int	nr=0;	// return value
int	i, k;	// index, count of 6-bit chars.
int	j;	// Other index
int	e=0;	// error flag
int	b, s;	// data byte, shift
// we do need to check that 6*sx = 8*n and adjust accordingly

    if (n < 1) return 0;
    if (sx == NULL) return 0;
    if (bn == NULL) return 0;
    k = (n*8 + 5) / 6;		// number of characters in sx needed
// loop over the characters of sx, obtaining a binary value
// for each character
// then put that binary value into the right part of bn
// and watch out for exceeding length n on last character

    j = strlen (sx);		// number of characters we actually have
    nr = (j * 6 + 7) / 8;	// number of binary characters we could get
    memset (bn, 0, n);		// clear the array
    if (j < k) k = j;		// limit to available characters
    if (nr > n) nr = n;		// limit to number requested also
    for (i=0; i < k; i++) {
	b = bcsix (sx[i]);	// Byte to insert
//	if (b < 0) continue;	// Error bytes don't get stored
	j = (6*i)/8;		// Starting index
	s = 2 - ((6*i)%8);	// shift. 2 - [0,2,4,6] = 2, 0, -2, -4
	if (b < 0) {		// Error byte, reduce nr
	    if (e == 0) nr--;	// Lost at least 1 byte
	    if (s < 0) { nr--; e++; }  // lost overlap byte
	    continue;
	}
	e = 0;
	switch (s) {
	    case  0:
		bn[j] |= b;
		break;
	    case  2:
		bn[j] = b << 2;
		break;
	    case -2:
		bn[j]  |= b >> 2;
 if ((j+1) < n) bn[j+1] = (b << 6) & 255;
		break;
	    case -4:
		bn[j]  |= b >> 4;	// shift -s
 if ((j+1) < n) bn[j+1] = (b << 4) & 255;  // shift 8+s
// THAT SHOULD BE n-1 OR j+1 IN THE TEST ABOVE !!
	}
    }	// loop over all input bytes
//    if (bn[n-1] == 0) nr &= ~1;  // even return if last byte is zero
    return  nr;
}


int	six4bin (char* sx, unsigned char* bn, int n)
/*  Convert n unsigned binary characters in bn to a string of special
  characters at sx, with each character in sx encoding 6 bits from the
  bn array interpreted as a bit field.  Return the number of 6-bit
  characters produced.  */
{
int	i, k;	// index, count of 6-bit chars.
int	j;	// Other index
int	b, s;	// data byte, shift
// we do need to check that 6*sx = 8*n and adjust accordingly

    if (n < 1) return 0;	// or set n = 0 and slide through...
    if (bn == NULL) return 0;
    if (sx == NULL) return 0;
    k = (n*8 + 5) / 6;		// number of characters in sx needed
//    printf (" --six4bin translate %d into %d characters.\n", n, k);  // DEBUG ONLY
// loop over the 6 bit fields present in bn,
// the last one may be a partial field, pad to right with 0
// Extract the 6 bits for each, convert to a character and
// place it in sx
// put a null byte into sx following the last one.

    for (i=0; i < k; i++) {
//	b = bcsix (sx[i]);	// Byte to insert
//	if (b < 0) continue;	// Error bytes don't get stored
	j = (6*i)/8;		// Starting index
	s = 2 - ((6*i)%8);	// shift. 2 - [0,2,4,6] = 2, 0, -2, -4
	b = bn[j];
//	if (s < 0 && j+1 < n) {	// need an extra byte's data
//	    b = (b << 8) | bn[j+1];
	if (s < 0) {		// need extra byte's data
	    b = b << 8;
	    if (j+1 < n) b |= bn[j+1];  // only if present at end.
	    s += 8;
	}
	b &= (63 << s);
	sx[i] = sixbc (b >> s);
    }
    sx[k] = '\0';	// Terminate the string
    return  k;
}


/*  ====  Dynamic Buffer support  ====  */

#define  Hunk  64
//-  dynamic buffer support belongs in ioutils.c
void	Dsadd (Dbuf* d, char* str)
/* Add string str to end of dynamic buffer d, appending storage to d
  as needed.  When needed, add a Hunk of storage and copy the existing
  data to the new size buffer and drop the old one.  */
{
int	l, k, n;
int	s, m;
//char	*b;
    if (d == NULL) return;	// we don't get to make one here
    if (str == NULL) return;

// See how much we need to add.  If zero, leave early.
    l = strlen (str);
    if (l < 1) return;

// See how much is being used and how much is available; see if
// the added stuff will fit.
    k = (d->buf == NULL) ? 0 : strlen (d->buf);
    n = k + l + 1 - d->size;	// This is how much more we need

    if (n > 0) {
// New data does not fit.  Find the next multiple of Hunk which
// will hold the additional amount needed.  Increase the allocation
// by that amount and copy existing data to the new buffer, and
// free the old one, put pointer to new one in d structure.
	m = (n + Hunk -1) / Hunk;
	s = d->size + m * Hunk;   // the new allocation
	d->buf = realloc (d->buf, s);
	if (k == 0) d->buf[0] = '\0';  // realloc does not clear memory
	d->size = s;
    }

    strcat (d->buf, str);	// Append string to buffer
}


/*  ===  Utilities to support Pvfs and Vers structures  ===  */

Vers*	kill_vers (Vers* a)
/*  Remove all elements from queue a, return their storage
  and contents, and return a NULL pointer.  */
{
Vers	*b;
Vers	*c;
    b = a;
    while (b != NULL) {
	c = G_pop_c (b);
	free (b->tag);
	free (b);
	b = c;
    }
    return  b;
}

Pvfs*	clean_pvfs (Pvfs* a)
/*  Clear all data in given Pvfs structure, return argument */
{
    if (a == NULL) return a;
    a->ver  = kill_vers (a->ver);
    a->feat = kill_name (a->feat);
    free (a->date);
    free (a->name);
//    free (a);
    return  NULL;
}

Pvfs*	kill_pvfs (Pvfs* a)
/*  Clear all data in given Pvfs structure, return NULL */
{
    if (a == NULL) return a;
    clean_pvfs (a);
    free (a);
    return  NULL;
}

/*  ====  Support for version control  ====  */

int	fvcmp (float a, float b)
/* Compare a to b; return -1 if a < b, 0 if a == b,
  and +1 if a > b.  Comparison of near equality is
  done correctly here.  */
{
int	r = 0;
float	d;

    d = a - b;
    if (fabsf (d)  < 0.5e-4) return r;	// equality
    if (fabsf(d/a) < 0.5e-5) return r;
    r = ( d > 0.0 ) ? 1 : -1;
    return  r;
}

int	vvcmp (Vers* a, Vers* b)
/* Compare Vers a to Vers b with b as the reference;
  return -2 if a->v is less than b->m, -1 if a->v is at least
  b->m but less than b->v, 0 if a is equal to b->v and +1
  if a->v is greater than b->v.  Uses fvcmp  */
/* Returns mean as follows:
  +1  a is a beta version relative to b
   0  a is same version as reference b; a current
  -1  a is older than b but acceptable; warn user
  -2  a is not acceptable, refuse to use version a.  */
/* We do not compare a->m to anything, only the minimum
  version of the reference b should matter.  */
{
int	r;	// result
    if (a == NULL || b == NULL) return -3;	// unknown compare
    r = fvcmp (a->v, b->v);	// Compare versions
    if (r >= 0) return r;	// equal or better return
// At this point, r is < 0; a->v is less than b->v
    r = fvcmp (a->v, b->m);	// Compare to minimum
    if (r >= 0) return -1;	// Made minimum
    return -2;			// Fails to make minimum
}

int	hasfeat (char* name, namlist* list)
/*  Search the given namlist for presence of name, return 1 if
  found, and 0 if not found.  */
{
int	r = 0;
namlist	*p;
    C_loop (p, list) {
	if (strcmp(name, p->name)) continue;
	r = 1;
	break;
    }
    return  r;
}

Vers*	findver (char* name, Pvfs* head)
/* Find and return pointer to the Vers data for the named
  component from the head structure.  If not found, NULL is returned.  */
{
Vers	*p = NULL;
/* Take care of some null cases first */
    if (head == NULL) return p;
    if (name == NULL) return p;
    C_loop (p, head->ver) {
	if (strcmp(name, p->tag) == 0) return p;
    }
    return  NULL;	// not found
}

int	comp_vers (Vers* v, Pvfs* p)
/*  Compare given version v to data set p as reference.  Return the usual
  values, -3 = unknown, -2 = obsolete, -1 = older, 0 = current, 1 = beta.
  The p data is obtained separately, usually from FTP data.  The v is
  from another Pvfs or is from an Obs.source structure.  */
{
int	r = -3;  // start with unknown case
Vers	*pv;

// Take care of null cases
    if (v == NULL || p == NULL) return r;

// Find the version in p
    pv = findver (v->tag, p);
    if (pv == NULL) return r;

// Call for compare
    r = vvcmp (v, pv);

// Return result.
    return  r;
}


Vers*	addvers (char* tag, float v, float m, Vers* head)
/*  Create a vers structure with data, add it at queue head,
  return the new head. */
{
Vers	*r;
    r = calloc (1, sizeof(Vers));
    r->v = v;
    r->m = m;
    r->tag = dynamstr (tag);
    return  G_push_c (r, head);  // or maybe G_push_cl ?
}

namlist*  addfeat (char* name, namlist* head)
/*  Create a namlist element from name and add to the queue
  before head and return head.  */
{
char	buf[24];
    if (name == NULL) return head;
    sprintf (buf, "[%s]", name);
    return  pushtxt (buf, head);
}

namlist*  addvfeat (char* name, int v, namlist* head)
/*  Create a namlist element from name, v and add to the queue
  before head and return head.  */
{
char	buf[24];
    if (name == NULL) return head;
    sprintf (buf, "[%s %d]", name, v);
    return  pushtxt (buf, head);
}

namlist*  addxfeat (char* name, char* v, namlist* head)
/*  Create a namlist element from a symbol name and value.
  Intended to be called like this:
	p->feat = addzfeat ("Symbol", Xstr(Symbol), p->feat);
  When Symbol is not defined, Xstr(Symbol) evaluates to "Symbol",
  but when it is defined, Xstr(Symbol) evaluates to the definition
  of Symbol enclosed in quotes as a string.  If that definition is 
  blank or not given, the string is Null.
  Thus if the string v is the same as name, nothing is entered into
  the queue as the symbol is not defined.  However, if there is a
  definition, the name is entered.  If the definition is Null,
  nothing else is written, but if the definition is not null, it
  is written after a space.  The resulting namlist element is
  then added to the queue before head, and head is returned.    */
/*  As a special feature, if the definition is numeric zero ("0")
  exactly, the symbol is treated as being undefined.  To have a
  zero value entered, one can use "-0" or "00" or the equivalent;
  even "zero" should compare to a numeric zero valaue.    */
{
char	buf[24];
    if (name == NULL || strlen(name) < 1) return head;
    if (strcmp (name, v) == 0) return head;
    if (strcmp (v, "0") == 0) return head;  // Special case
    if (v == NULL || strlen(v) < 1) sprintf (buf, "[%s]", name);
    else			sprintf (buf, "[%s %s]", name, v);
    return  pushtxt (buf, head);
}

char*	putvfs (Pvfs* head)
/* Put information from head into a dynamic buffer, and return
  the address of that buffer.  */
// was makepbuf
{
char	*result;
Dbuf	bb = {NULL, 0};
Vers	*v;
namlist	*t;

    Dsadd (&bb, head->date);
    C_loop (v, head->ver) {	// Format the versions
char	c[32];
char	*vs, *rs;
	vs = fvstr (v->v);
	rs = fvstr (v->m);
	sprintf (c, " %s:%s(%s)", v->tag, vs, rs);
	Dsadd (&bb, c);
	free (vs);  free(rs);
    }
    C_loop (t, head->feat) {	// Run the namelist
	Dsadd (&bb, " ");
	Dsadd (&bb, t->name);
    }

    result = bb.buf;
/*  printf (" Buffer uses %d of %d characters. (%d Hunks)\n",
	strlen(result), bb.size, bb.size/Hunk);  // debug only  */
    bb.buf = NULL;	// clean up.
    return  result;
}

Pvfs*	getvfs (char* bufr, char* source)
/* Given a buffer from a file or FTP operation, parse the buffer data and
  produce a Pvfs structure containing the data.  If the buffer appears to
  be seriously in error, return NULL.  */
/* The source argument is put in the structure as the name field;  it is
  usually the file name used to obtain the buffer, or the FTP url.  This
  string is used in report titles.  */
//  was findvers
{
Pvfs	*p = NULL;
Vers	*v;
char	*c;		// parsing pointer
char	*s;		// sub-parsing pointer
char	tokn[80];	// token buffer
char	part[20];	// sub-token buffer
//const	char	delim[] = " ";	// space delimited.
char	delim[] = " ";	// space delimited.
/* NOTE - parse should treat delim as constant; change parse (ioutils.c:571)
to use const char* for delim.  Then we can use it here.  */
int	k = 0;	// priorities
int	f = 0;	// features
int	e = 0;	// environment fields

    if (bufr == NULL) return p;
    if (strlen(bufr) < 20) return p;	// Needs some size
    c = parse (bufr, tokn, delim);	// date
    if (strlen(tokn) < 16) return p;	// not big enough
    if (strlen(tokn) > 19) return p;	// too big

/* Looks like we may have a valid buffer.  Make a header, and
  put the date into it.  */
    p = calloc (1, sizeof(Pvfs));
    p->date = dynamstr (tokn);
    p->name = dynamstr (source);
    while (*c != '\0') {	// parse all the other tokens
//	c = parse (c, tokn, delim);	// Get next token
	if (*c == '[') {	// Feature, a special token
//	if (tokn[0] == '[') {	// It's a feature
	    c = parse (c, tokn, "]");	// special parse to end
	    strcat (tokn, "]");  // restore the end
	    c = skipover (c, delim);  // and skip blank after it
	    p->feat = pushtxt (tokn, p->feat);
	    f++;
	} else {		// Must be a version field (or environment)
float	vv, mv;
char	*tag;
	    c = parse (c, tokn, delim);  // do a regular parse
/* If tokn contains "=" this is an environment string, and we make a copy
  of it and do a putenv on that copy.  If not, just continue on.    */
	  if (strchr (tokn, '=')) {	// Environment string
/* First, find the name and check if it is already defined in the
  environment.  If it is, we allow the existing value to remain and
  do not put the token into the environment.  */
	    s = parse (tokn, part, "=");	// part is the name
	    if (getenv(part) == NULL)  // Do the next statement if not in env.
//	    putenv (tokn);
	    putenv (strdup(tokn));
/* NOTE that this is a memory leak.  The putenv man page indicates that in
  some cases the given string becomes part of the environment and should
  not be changed.  If this is the case, we leave the copy.  If it can be
  shown not to be the case, just do the putenv on tokn directly.
  In Fedora 5 the strdup is observed to be required.    */
	    p->feat = pushtxt (tokn, p->feat);	// also save in feat queue
	    e++;
	  } else {	// version field, decode
	    s = parse (tokn, part, ":");
	    tag = dynamstr (part);
	    s = parse (s, part, "(");
//	    v->v = floatvalue (part);	// or atof (part) ?
//	    v->v = strtof (part, NULL);  // needs #define _XOPEN_SOURCE=600
	    vv = atof (part);
	    if (strlen (part) < 2) continue;	// Needs both parts
	    s = parse (s, part, ")");
	    mv = atof (part);
	    v = calloc (1, sizeof(Vers));
	    v->tag = tag;
	    v->v = vv;
	    v->m = mv;
	    p->ver = G_push_c (v, p->ver);
	    k++;
	  }	// End of test on = in token
	}	// End of feature/non-feature test
    }
//    printf (" Found %d priorities, %d features.\n", k, f);  // debug only
    return  p;
}


char*	getafile (char* filename)
/* Returns a dynamic buffer containing file contents of named file. */
{
char	*buf = NULL;
int	bs = 0;
FILE	*f;
int	i;

//printf (" Getafile reading \"%s\".\n", filename);	// debug
    f = xopen (filename, "r");
    if (f == NULL) return  buf;
    for (i=0; ; i++) {
	if (bs <= i) {		// need more storage
	    bs += Hunk;
	    buf = realloc (buf, bs);
	}
	buf[i] = fgetc (f);

	if (buf[i] == EOF) {
	    buf[i] = '\0';
	    break;
	}
    }		// end of reading loop
//printf (" Getafile read %d characters.  Buffer length = %d.\n", i, bs);
    iclose (f);
    return  buf;
}

void	putafile (char* buf, char* filename)
/*  Open a file named filename and write the buffer into
  that file, then close it.  */
{
FILE	*f;

    if (filename == NULL) return;
    if (buf == NULL) return;
    f = xopen (filename, "w");
    if (f) {
	fprintf (f, "%s", buf);
	iclose (f);
    } else {
	printf (" *Could not write file to \"%s\".\n", filename);
    }
}


void	displist (namlist* list)
/*  For the list, format a report in columns for the list members.  */
/* NOTE:  May want to add a FILE* parameter and use fprintf here.  */
{
#define  linesize  78
#define  leftmarg  2
#define  colspace  4
namlist	*n;
char	**buf;
char	*b;	// scratch pointer
int	nc = 0;	// name count, and position
int	mx = 0;	// largest name
int	sz;	// name size
int	npl;	// names per line
int	nl;	// number of lines (buffers)
int	i, j;	// scratch index

/*  Scan the name queue and find number of entries and length of
  the longest one.  Compute the number of lines needed with the
  derived spacing, and allocate buffers for all the lines.  Then
  process the list filling in names to the proper buffers.  Write
  all the buffers in a loop, and free them. 
  (This is from ~/source/backup.c:1456)  */
    C_loop (n, list) {
	sz = strlen (n->name);
	if (sz > mx) mx = sz;
	nc++;
    }
    npl = (linesize - leftmarg) / (mx + colspace);
    if (npl < 1) npl = 1;
    nl = (nc + npl - 1) / npl;

/* Allocate and fill a set of buffers. */
    buf = calloc (nl, sizeof(char*));	// get nl pointers
    for (i=0; i<nl; i++) {
	b = malloc (linesize+1);
	for (j=0; j<leftmarg; j++) b[j] = ' ';
	b[leftmarg] = '\0';
	buf[i] = b;
    }

/* Loop through list filling buffers */
    nc = 0;
    C_loop (n, list) {
	b = buf[nc];		// select the buffer
	if ( (sz=strlen(b)) > leftmarg) {	// need colspace
	    i = sz + colspace;
	    for (j=sz; j<i; j++) b[j] = ' ';
	    b[i] = '\0';
	    sz = i;
	}
	sprintf (b+sz, "%-*s", mx, n->name);
	nc = (nc + 1) % nl;
    }

/* Buffers filled, now print them */
    for (i=0; i<nl; i++) {
	printf ("%s\n", buf[i]);
	free (buf[i]);
    }
    free (buf);
#undef  linesize
#undef  leftmarg
#undef  colspace
}


/*  ====  Support for libcurl actions  ====  */

/* WriteMemoryCallback is local here and specific to libcurl */
// change "struct MemoryStruct" to "Dbuf"
// and members "memory" to "buf"; and "size" remains "size"
static size_t
WriteMemoryCallback(void *ptr, size_t size, size_t nmemb, void *data)
{
size_t	realsize = size * nmemb;
Dbuf	*mem = (Dbuf *)data;

    mem->buf = (char *)realloc(mem->buf, mem->size + realsize + 1);
    if (mem->buf) {
	memcpy(&(mem->buf[mem->size]), ptr, realsize);
	mem->size += realsize;
	mem->buf[mem->size] = 0;
    }
    return realsize;
}

char*	getFTP (char*  url)
/* Given a url, attempt to read the data there and put it in a dynamic
  character buffer which is returned.  NULL return on error.  */
{
CURL	*curl_handle;
Dbuf	chunk;

    chunk.buf=NULL; /* we expect realloc(NULL, size) to work */
    chunk.size = 0;    /* no data at this point */

    curl_global_init(CURL_GLOBAL_ALL);

/* init the curl session */
    curl_handle = curl_easy_init();

/* specify URL to get */
    curl_easy_setopt(curl_handle, CURLOPT_URL, url);

/* send all data to this function  */
    curl_easy_setopt(curl_handle, CURLOPT_WRITEFUNCTION, WriteMemoryCallback);

/* we pass our 'chunk' struct to the callback function */
    curl_easy_setopt(curl_handle, CURLOPT_WRITEDATA, (void *)&chunk);

/* some servers don't like requests that are made without a user-agent
     field, so we provide one */
//?    curl_easy_setopt(curl_handle, CURLOPT_USERAGENT, "libcurl-agent/1.0");

/* Try setting a user name and password for FTP ? */
    curl_easy_setopt(curl_handle, CURLOPT_USERPWD,
		"anonymous:kdcio@obs.carnegiescience.edu");
/* Above observed to work, by temporarily setting verbose on, below.  */
/* If the above doesn't work, use the CURLOPT_QUOTE method; see web page. */
/* Send command "user anonymous kdcio@ociw.edu" via quote if needed. */

// set debug output on (mostly, leave this commented.)
//    curl_easy_setopt (curl_handle, CURLOPT_VERBOSE, 1);

/* get it! */
    curl_easy_perform(curl_handle);

/* cleanup curl stuff */
    curl_easy_cleanup(curl_handle);

/* The buffer is in chunk.buf now. */
    return  chunk.buf;
}



/*  ====  */

