/* Utility programs for Spherical Astronomy */
/* Mask Generation suplemential edition */

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * **
**                                                                   **
**                  Copyright (C) 2011 by                            **
**     Ken Clardy,  Carnegie Observatories,  Pasadena California.    **
**                                                                   **
**      This software is proprietary, and may be used or copied      **
**          only with the written consent of the author.             **
**              All copies must retain this notice.                  **
**                                                                   **
** * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* We need to be able to do these things:
 1) Precess catalog position to given epoch (coords or vector)
 2) Given a center, and catalog position, find projected vector
 3) Inverse, from projected point, find ra/dec
.. */

/* Other uses of these programs:
 1) Do l.s. solution for change x, y, rotation; then convert x, y
  changes into ra/dec changes.
 2) rotation of 2-d x,y values about a given center
  Uses 2-d dot product, vector add, subtract.
.. */

/* Radical idea -- rather than subscripting, we define a 3-vector
as a *structure* with x,y,z components.  Then we can transfer it as
a value, use it as an argument, and define a matrix as an array of
3 vectors.  */


/* ---------------------------------------------------------------------- */
/*  Defines  */
#ifdef    NEED_SINCOS
#undef    HAVE_SC
#else
#define   HAVE_SC
#endif
/* HAVE_SC is defined when the sincos function is available, and can
  save up to 40% execution time when computing both sin and cos of
  the same angle.  If this function is not available, undef here.  */
/* NEED_SINCOS is defined in those systems which need the sincos
  function, having no such thing in their math library.  A lame version
  using sin and cos is implemented here, to keep things compatable.  It
  in no way improves the computational efficiency as a real implementation
  of sincos would naturally do.  */

/* ---------------------------------------------------------------------- */

#include	"kdcutil.h"
#include	<string.h>
/* The string library needed for function 'English' below.  */

/* GLOBALs -------------------------------------------------------------- */

ConsNams  Cnam[Ncon] = {
 {"---", "??", " ", "Unknown Constellation"},
 {"And", "Andromeda", "Andromedae", "Chained Maiden"},
 {"Ant", "Antlia", "Antliae", "Air Pump"},
 {"Aps", "Apus", "Apodis", "Bird of Paradise"},
 {"Aqr", "Aquarius", "Aquarii", "Water Carrier"},
 {"Aql", "Aquila", "Aquilae", "Eagle"},
 {"Ara", "Ara", "Arae", "Altar"},
 {"Ari", "Aries", "Arietis", "Ram"},
 {"Aur", "Auriga", "Aurigae", "Charioteer"},
 {"Boo", "Bootes", "Bootis", "Herdsman"},
 {"Cae", "Caelum", "Caeli", "Engraving Tool"},
 {"Cam", "Camelopardalis", "Camelopardalis", "Giraffe"},
 {"Cnc", "Cancer", "Cancri", "Crab"},
 {"CVn", "Canes Venatici", "Canun Venaticorum", "Hunting Dogs"},
 {"CMa", "Canis Major", "Canis Majoris", "Great Dog"},
 {"CMi", "Canis Minor", "Canis Minoris", "Lesser Dog"},
 {"Cap", "Capricornus", "Capricorni", "Sea Goat"},
 {"Car", "Carina", "Carinae", "Keel"},
 {"Cas", "Cassiopeia", "Cassiopeiae", "Seated Queen"},
 {"Cen", "Centaurus", "Centauri", "Centaur"},
 {"Cep", "Cepheus", "Cephei", "King"},
 {"Cet", "Cetus", "Ceti", "Sea Monster"},
 {"Cha", "Chamaeleon", "Chamaeleontis", "Chameleon"},
 {"Cir", "Circinus", "Circini", "Dividers"},
 {"Col", "Columba", "Columbae", "Dove"},
 {"Com", "Coma Berenices", "Comae Berenices", "*Berenice's Hair"},
 {"CrA", "Corona Australis", "Coronae Australis", "Southern Crown"},
 {"CrB", "Corona Borealis", "Coronae Borealis", "Northern Crown"},
 {"Crv", "Corvus", "Corvi", "Crow"},
 {"Crt", "Crater", "Crateris", "Cup"},
 {"Cru", "Crux", "Crucis", "Southern Cross"},
 {"Cyg", "Cygnus", "Cygni", "Swan"},
 {"Del", "Delphinus", "Delphini", "Dolphin"},
 {"Dor", "Dorado", "Doradus", "Swordfish (Goldfish)"},
 {"Dra", "Draco", "Draconis", "Dragon"},
 {"Equ", "Equuleus", "Equulei", "Little Horse"},
 {"Eri", "Eridanus", "Eridani", "River"},
 {"For", "Fornax", "Fornacis", "Furnace"},
 {"Gem", "Gemini", "Geminorum", "Twins"},
 {"Gru", "Grus", "Gruis", "Crane"},
 {"Her", "Hercules", "Herculis", "*"},
 {"Hor", "Horologium", "Horologii", "Clock"},
 {"Hya", "Hydra", "Hydrae", "Sea Serpent"},
 {"Hyi", "Hydrus", "Hydri", "Water Snake"},
 {"Ind", "Indus", "Indi", "Indian"},
 {"Lac", "Lacerta", "Lacertae", "Lizard"},
 {"Leo", "Leo", "Leonis", "Lion"},
 {"LMi", "Leo Minor", "Leonis Minoris", "Lesser Lion"},
 {"Lep", "Lepus", "Leporis", "Hare"},
 {"Lib", "Libra", "Librae", "Balance"},
 {"Lup", "Lupus", "Lupi", "Wolf"},
 {"Lyn", "Lynx", "Lyncis", "Lynx"},
 {"Lyr", "Lyra", "Lyrae", "Lyre"},
 {"Men", "Mensa", "Mensae", "*Table Mountain"},
 {"Mic", "Microscopium", "Microscopii", "Microscope"},
 {"Mon", "Monoceros", "Monocerotis", "Unicorn"},
 {"Mus", "Musca", "Muscae", "Fly"},
 {"Nor", "Norma", "Normae", "Carpenter's Square"},
 {"Oct", "Octans", "Octantis", "Octant"},
 {"Oph", "Ophiuchus", "Ophiuchi", "Serpent Bearer"},
 {"Ori", "Orion", "Orionis", "Hunter"},
 {"Pav", "Pavo", "Pavonis", "Peacock"},
 {"Peg", "Pegasus", "Pegasi", "Winged Horse"},
 {"Per", "Perseus", "Persei", "Hero"},
 {"Phe", "Phoenix", "Phoenicis", " "},
 {"Pic", "Pictor", "Pictoris", "Painter's Easel"},
 {"Psc", "Pisces", "Piscium", "Fishes"},
 {"PsA", "Piscis Austrinus", "Pices Austrini", "Southern Fish"},
 {"Pup", "Puppis", "Puppis", "Stern"},
 {"Pyx", "Pyxis", "Pyxidis", "Ship's Compass"},
 {"Ret", "Reticulum", "Reticuli", "Reticle"},
 {"Sge", "Sagitta", "Sagittae", "Arrow"},
 {"Sgr", "Sagittarius", "Sagittarii", "Archer"},
 {"Sco", "Scorpius", "Scorpii", "Scorpion"},
 {"Scl", "Sculptor", "Sculptoris", "Sculptor"},
 {"Sct", "Scutum", "Scuti", "Shield"},
 {"Ser1", "Serpens Caput", "Serpentis", "Snake (head)"},
 {"Ser2", "Serpens Cauda", "Serpentis", "Snake (tail)"},
 {"Sex", "Sextans", "Sextantis", "Sextant"},
 {"Tau", "Taurus", "Tauri", "Bull"},
 {"Tel", "Telescopium", "Telescopii", "Telescope"},
 {"Tri", "Triangulum", "Trianguli", "Triangle"},
 {"TrA", "Triangulum Australe", "Tranguli Australis", "Southern Triangle"},
 {"Tuc", "Tucana", "Tucanae", "Toucan"},
 {"UMa", "Ursa Major", "Ursae Majoris", "Great Bear"},
 {"UMi", "Ursa Minor", "Ursae Minoris", "Little Bear"},
 {"Vel", "Vela", "Velorum", "Sails"},
 {"Vir", "Virgo", "Virginis", "Maiden"},
 {"Vol", "Volans", "Volantis", "Flying Fish"},
 {"Vul", "Vulpecula", "Vulpeculae", "Fox"} };


/*  Constellation South boundary table  */

boundlist  Tbound[Fbound] = { {     0, 86400,  21601,  0},
 {     0, 86400,  21120, 85}, { 28800, 52200,  20760, 85},
 { 75600, 82800,  20680, 85}, { 64800, 75600,  20640, 85},
 {     0, 28800,  20400, 20}, { 33000, 38400,  19680, 11},
 {     0, 18000,  19200, 20}, { 38400, 52200,  19200, 11},
 { 63000, 64800,  19200, 85}, { 72600, 75600,  19200, 34},
 {     0, 12630,  18480, 20}, { 41400, 48900,  18480, 11},
 { 59520, 63000,  18000, 85}, { 72600, 74400,  18000, 20},
 { 28680, 33000,  17640, 11}, { 33000, 40800,  17640, 34},
 { 46800, 59520,  16800, 85}, { 11160, 12300,  16320, 18},
 { 73500, 74400,  16080, 34}, { 40800, 43200,  15960, 34},
 {     0,  1200,  15840, 20}, { 50400, 56400,  15840, 85},
 { 84900, 86400,  15840, 20}, { 43200, 48600,  15360, 34},
 { 48600, 51900,  15120, 34}, { 83400, 84900,  15120, 20},
 { 21960, 25200,  14880, 11}, { 72000, 73500,  14760, 34},
 { 73932, 74160,  14620, 20}, { 25200, 28680,  14400, 11},
 { 28680, 30300,  14400, 84}, { 71160, 72000,  14280, 34},
 { 72000, 73932,  14280, 20}, { 82320, 83400,  14180, 20},
 {     0,  8760,  14040, 18}, { 69900, 71160,  13920, 34},
 {  6120,  6870,  13800, 18}, {  8760, 11160,  13680, 18},
 { 11160, 11400,  13680, 11}, { 80340, 82320,  13500, 20},
 { 18000, 21960,  13440, 11}, { 50520, 51900,  13320, 84},
 { 51900, 69900,  13320, 34}, { 11400, 12000,  13200, 11},
 { 79680, 80340,  13200, 20}, { 74160, 79080,  13160, 20},
 {     0,  6120,  12960, 18}, { 21960, 23400,  12960, 51},
 { 43500, 48600,  12720, 84}, { 54900, 56700,  12720, 34},
 { 79080, 79680,  12660, 20}, { 12000, 18000,  12600, 11},
 { 82320, 84000,  12600, 18}, { 56700, 61200,  12360, 34},
 {  7350,  9060,  12120, 63}, { 61200, 65640,  12120, 34},
 {     0,  4920,  12000, 18}, {  4920,  6000,  12000, 63},
 { 23400, 24480,  12000, 51}, { 84000, 86400,  12000, 18},
 { 48600, 50520,  11640, 84}, {     0,  4020,  11520, 18},
 { 84900, 86400,  11520, 18}, { 65430, 65640,  11400, 40},
 { 65640, 68700,  11400, 34}, { 68700, 69000,  11400, 31},
 {  6000,  7350,  11280, 63}, { 30300, 33000,  11280, 84},
 {   600,  3120,  11040, 18}, { 43200, 43500,  10800, 84},
 { 24480, 26520,  10680, 51}, { 78870, 79080,  10560, 31},
 { 78750, 78870,  10500, 31}, { 69000, 69840,  10440, 31},
 { 33000, 36600,  10080, 84}, { 36600, 38820,   9600, 84},
 { 55560, 56700,   9600,  9}, { 56700, 58800,   9600, 40},
 { 33300, 34500,   9540, 51}, {     0,  9060,   8820,  1},
 {  9060,  9240,   8820, 63}, { 69690, 69840,   8760, 52},
 { 16200, 16890,   8640, 63}, { 78240, 78750,   8640, 31},
 { 78750, 79200,   8640, 45}, { 23520, 26520,   8520,  8},
 { 26520, 27900,   8520, 51}, {     0,  7200,   8400,  1},
 { 79200, 82140,   8400, 45}, { 82140, 82320,   8280, 45},
 { 82320, 84600,   8280,  1}, {  9240,  9780,   8160, 63},
 { 38820, 39600,   8160, 84}, { 43200, 44400,   8160, 13},
 { 27900, 33300,   8040, 51}, { 33300, 35580,   8040, 47},
 {  2580,  5070,   7920,  1}, { 54660, 55560,   7920,  9},
 { 84600, 85500,   7700,  1}, { 44400, 47700,   7680, 13},
 { 85500, 86400,   7520,  1}, { 50250, 50520,   7380, 13},
 {  8700,  9780,   7360, 81}, {  9780, 16200,   7360, 63},
 { 16200, 17100,   7200,  8}, { 65430, 69690,   7200, 52},
 { 39600, 43200,   6960, 84}, { 70800, 75300,   6960, 31},
 { 17100, 21180,   6840,  8}, { 35580, 37800,   6840, 47},
 { 47700, 50250,   6840, 13}, {     0,   240,   6720,  1},
 {  5070,  6000,   6720, 81}, { 21180, 23520,   6720,  8},
 { 28380, 28800,   6720, 38}, { 75300, 78240,   6720, 31},
 { 69330, 70800,   6600, 31}, {  6900,  8700,   6540, 81},
 { 58200, 58800,   6480, 27}, { 54300, 54660,   6240,  9},
 { 54660, 58200,   6240, 27}, { 66120, 67920,   6240, 52},
 { 38700, 39600,   6120, 47}, { 67920, 69330,   6120, 52},
 {  6000,  6900,   6000, 81}, {  2580,  3060,   5700, 66},
 { 37800, 38700,   5640, 47}, { 76500, 77100,   5640, 89},
 { 20520, 21180,   5480, 79}, {   240,   510,   5280,  1},
 { 57300, 57720,   5280, 76}, { 21180, 22380,   5160, 38},
 { 71400, 72900,   5100, 89}, { 67920, 69300,   5060, 89},
 {   510,  3060,   5040,  1}, { 72900, 74040,   4920, 89},
 { 28110, 28380,   4800, 38}, { 74040, 76500,   4680, 89},
 { 69300, 71400,   4600, 89}, { 11820, 12120,   4560,  7},
 { 67920, 68400,   4440, 71}, { 20520, 20760,   4320, 60},
 { 22380, 22710,   4200, 38}, { 68400, 71400,   3880, 71},
 { 17880, 19200,   3840, 79}, { 57300, 57900,   3840, 40},
 { 71400, 72900,   3780, 71}, { 16620, 17880,   3720, 79},
 { 19200, 20160,   3720, 79}, { 46200, 48600,   3600, 25},
 { 62100, 65700,   3440, 40}, { 42720, 46200,   3360, 25},
 { 27000, 28110,   3240, 38}, { 60300, 62100,   3080, 40},
 {     0,   510,   3000, 62}, { 20160, 20760,   3000, 79},
 { 25200, 27000,   3000, 38}, { 76020, 76800,   3000, 62},
 { 22710, 24960,   2880, 38}, { 65700, 67920,   2880, 40},
 { 75150, 75780,   2840, 32}, { 75780, 76020,   2840, 62},
 { 41460, 42720,   2640, 46}, { 22470, 22710,   2400, 60},
 { 24960, 25200,   2400, 38}, { 28110, 28530,   2400, 12},
 { 85800, 86400,   2400, 62}, {  6000, 11820,   2380,  7},
 { 72510, 73080,   2040, 32}, { 48600, 54300,   1920,  9},
 { 81900, 85800,   1800, 62}, { 28530, 33300,   1680, 12},
 { 33300, 38700,   1680, 46}, { 65700, 67184,   1500, 59},
 { 67184, 67920,   1500,  5}, { 75000, 75150,   1440, 32},
 { 25200, 25260,   1320, 15}, { 65700, 66330,   1080, 77},
 { 57900, 60300,    960, 40}, { 65700, 66330,    720, 59},
 { 77280, 78000,    660, 62}, {     0,  7200,    480, 66},
 { 66900, 67920,    480, 77}, { 73080, 75000,    480, 32},
 { 75000, 76800,    480, 35}, { 76800, 77280,    480, 62},
 { 79200, 81900,    480, 62}, { 78000, 79200,    420, 62},
 { 25260, 25920,    360, 15}, { 12900, 16620,      0, 79},
 { 16620, 16800,      0, 60}, { 25920, 29100,      0, 15},
 { 52800, 54300,      0, 87}, { 64200, 65700,      0, 59},
 {  9540, 11820,   -420, 21}, { 11820, 12900,   -420, 79},
 { 54300, 58560,   -780, 76}, { 16800, 18300,   -960, 60},
 { 21000, 22470,   -960, 60}, { 64200, 64680,   -960, 77},
 { 65700, 66900,   -960, 77}, { 66900, 67920,   -960,  5},
 { 81900, 85800,   -960, 66}, { 38700, 41460,  -1440, 46},
 { 41460, 42600,  -1440, 87}, {     0,  1200,  -1680, 66},
 { 85800, 86400,  -1680, 66}, { 51300, 52800,  -1920, 87},
 { 57300, 58560,  -1920, 59}, { 72000, 73920,  -2160,  5},
 { 76800, 78720,  -2160,  4}, { 61800, 64680,  -2400, 59},
 { 21000, 29100,  -2640, 55}, { 17700, 18300,  -2640, 36},
 { 18300, 21000,  -2640, 60}, { 29100, 30120,  -2640, 42},
 { 34500, 38700,  -2640, 78}, { 42600, 46200,  -2640, 87},
 { 63300, 63600,  -2800, 59}, { 67920, 72000,  -2888,  5},
 { 17400, 17700,  -3480, 36}, { 73920, 76800,  -3600,  4},
 { 61800, 65700,  -3840, 77}, { 65700, 67920,  -3840, 75},
 { 30120, 30900,  -4080, 42}, { 58560, 58950,  -4380, 59},
 { 30900, 32700,  -4560, 42}, { 38700, 39000,  -4560, 29},
 { 58560, 58950,  -4620, 73}, { 56400, 57300,  -4800, 49},
 { 45300, 46200,  -5280, 28}, { 46200, 51300,  -5280, 87},
 { 32700, 35100,  -5760, 42}, {  6000,  9540,  -5852, 21},
 {  9540, 13500,  -5852, 36}, { 39000, 42600,  -5880, 29},
 { 42600, 45300,  -5880, 28}, { 51300, 53700,  -5880, 49},
 { 58560, 60300,  -5900, 59}, {     0,  6000,  -6120, 21},
 { 76800, 78720,  -6120, 16}, { 78720, 85800,  -6120,  4},
 { 85800, 86400,  -6120, 21}, { 35100, 36900,  -6360, 42},
 { 16920, 17400,  -6540, 36}, { 17400, 22020,  -6540, 48},
 { 72000, 76800,  -6720, 16}, { 36900, 38100,  -7000, 42},
 { 45300, 53700,  -7080, 42}, { 53700, 56400,  -7080, 49},
 { 56400, 57600,  -7080, 73}, { 16500, 16920,  -7200, 36},
 { 60300, 63360,  -7200, 59}, { 63360, 64200,  -7200, 72},
 { 38100, 39000,  -7480, 42}, { 22020, 26520,  -7920, 14},
 { 44100, 45300,  -7920, 42}, { 39000, 44100,  -8400, 42},
 { 12600, 13500,  -8640, 37}, { 30120, 33720,  -8820, 69},
 { 15360, 16500,  -8880, 36}, { 64200, 69000,  -8880, 72},
 { 76800, 82800,  -8880, 67}, { 82800, 84000,  -8880, 74},
 { 10800, 12600,  -9500, 37}, { 33720, 39600,  -9540,  2},
 {     0,  6000,  -9600, 74}, {  6000, 10800,  -9600, 37},
 { 13920, 15360,  -9600, 36}, { 84000, 86400,  -9600, 74},
 { 51000, 53700, -10080, 19}, { 56400, 57600, -10080, 50},
 { 57600, 59115, -10080, 73}, { 17400, 18000, -10320, 10},
 { 18000, 23700, -10320, 24}, { 28800, 30120, -10320, 68},
 { 12300, 13920, -10560, 36}, { 59115, 64200, -10920, 73},
 { 64200, 69000, -10920, 26}, { 69000, 73200, -10920, 72},
 { 73200, 76800, -10920, 54}, { 10800, 12300, -11040, 36},
 { 16200, 17400, -11160, 10}, { 55200, 56400, -11520, 50},
 {     0,  8400, -11560, 64}, {  9600, 10800, -11760, 36},
 { 14700, 15360, -11760, 41}, { 15360, 16200, -11760, 10},
 { 76800, 79200, -12000, 39}, { 21600, 28800, -12180, 68},
 { 28800, 29400, -12180, 86}, {  8700,  9600, -12240, 36},
 { 13800, 14700, -12240, 41}, {     0,  6600, -12360, 64},
 { 21600, 22200, -12600, 17}, { 29400, 30420, -12720, 86},
 { 12600, 13800, -12760, 41}, { 13800, 14400, -12760, 33},
 {     0,  5700, -12840, 64}, {  7800,  8700, -12960, 36},
 { 16200, 18000, -12960, 65}, { 54180, 55200, -12960, 50},
 { 30420, 31800, -13080, 86}, { 22200, 23400, -13200, 17},
 { 42600, 46200, -13200, 19}, { 51000, 54180, -13200, 50},
 { 54180, 55200, -13200, 57}, { 14400, 15600, -13560, 33},
 { 31800, 39600, -13560, 86}, { 39600, 40500, -13560, 19},
 { 63000, 64800, -13680,  6}, { 64800, 73200, -13680, 80},
 { 79200, 84000, -13680, 39}, { 11520, 12600, -13800, 41},
 { 18000, 19800, -13800, 65}, { 23400, 24600, -13920, 17},
 {     0,  4800, -14040, 64}, {  4800,  7800, -14040, 36},
 { 84000, 86400, -14040, 64}, { 15600, 16500, -14160, 33},
 { 55200, 59115, -14400, 57}, { 73200, 76800, -14400, 44},
 { 19800, 21600, -14640, 65}, { 54600, 55200, -14640, 23},
 { 59115, 59700, -14640,  6}, { 53700, 54600, -15260, 23},
 { 59700, 60300, -15260,  6}, { 21600, 24600, -15360, 65},
 { 24600, 32520, -15360, 17}, { 40500, 42600, -15360, 19},
 { 42600, 46200, -15360, 30}, { 46200, 52320, -15360, 19},
 { 48600, 49200, -15600, 23}, { 60300, 60600, -15600,  6},
 {  7800, 11520, -16200, 41}, { 11520, 16500, -16200, 70},
 { 53100, 53700, -16200, 23}, { 60600, 63000, -16200,  6},
 { 63000, 64800, -16200, 61}, { 79200, 84000, -16200, 83},
 { 16500, 23700, -16800, 33}, { 49200, 53100, -16800, 23},
 { 53100, 61200, -16800, 82}, {     0,  4800, -18000, 83},
 { 12600, 16500, -18000, 43}, { 23700, 32520, -18000, 88},
 { 32520, 40500, -18000, 17}, { 40500, 49200, -18000, 56},
 { 64800, 76800, -18000, 61}, { 76800, 84000, -18000, 44},
 { 84000, 86400, -18000, 83}, {  2700,  4800, -18240, 83},
 {     0, 12600, -19800, 43}, { 27600, 49200, -19800, 22},
 { 49200, 64800, -19800,  3}, { 12600, 27600, -20400, 53},
 {     0, 86400, -21601, 58} };


/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */

/* Compatability code */

#ifdef   NEED_SINCOS
void	sincos (double angle, double* s, double* c) {
    *s = sin (angle);
    *c = cos (angle);
}
#endif

/* ---------------------------------------------------------------------- */

/*  Complex Arithmetic as vect2 implementation */

vect2	csum (vect2 a, vect2 b) {
vect2	result;
    result.x = a.x + b.x;
    result.y = a.y + b.y;
    return result;
}

vect2	cdif (vect2 a, vect2 b) {
vect2	result;
    result.x = a.x - b.x;
    result.y = a.y - b.y;
    return result;
}

vect2	cmul (vect2 a, vect2 b) {
vect2	result;
    result.x = a.x * b.x - a.y * b.y;
    result.y = a.x * b.y + a.y * b.x;
    return result;
}

vect2	cdiv (vect2 a, vect2 b) {
/*  Division is multiply both by complex conjugate of b... */
vect2	result;
double	bs;
    bs = b.x * b.x + b.y * b.y;  // divisor
    result.x = (a.x * b.x + a.y * b.y) / bs;
    result.y = (a.y * b.x - a.x * b.y) / bs;
    return result;
}


/* ---------------------------------------------------------------------- */

/* Some basic vector things... */

vect3	make3vect (double x, double y, double z) {
vect3	result;
    result.x = x;
    result.y = y;
    result.z = z;
    return result;
}

vect2	make2vect (double x, double y) {
vect2	result;
    result.x = x;
    result.y = y;
    return result;
}

vect3	get3de2v (vect2 v, double z) {
vect3	result;
    result.x = v.x;
    result.y = v.y;
    result.z = z;
    return result;
}

vect2	get2de3v (vect3 v) {
vect2	result;
    result.x = v.x;
    result.y = v.y;
    return  result;
}

vect3	sum3vect (vect3 a, vect3 b) {
vect3	result;
    result.x = a.x + b.x;
    result.y = a.y + b.y;
    result.z = a.z + b.z;
    return result;
}

vect2	sum2vect (vect2 a, vect2 b) {
vect2	result;
    result.x = a.x + b.x;
    result.y = a.y + b.y;
    return result;
}

vect3	sub3vect (vect3 a, vect3 b) {
vect3	result;
    result.x = a.x - b.x;
    result.y = a.y - b.y;
    result.z = a.z - b.z;
    return result;
}

vect2	sub2vect (vect2 a, vect2 b) {
vect2	result;
    result.x = a.x - b.x;
    result.y = a.y - b.y;
    return result;
}

vect3	mul3vect (vect3 a, double b) {
vect3	result;
    result.x = a.x * b;
    result.y = a.y * b;
    result.z = a.z * b;
    return result;
}

vect2	mul2vect (vect2 a, double b) {
vect2	result;
    result.x = a.x * b;
    result.y = a.y * b;
    return result;
}

vect3	neg3vect (vect3 a) {
vect3	result;
    result.x = -a.x;
    result.y = -a.y;
    result.z = -a.z;
    return result;
}

vect2	neg2vect (vect2 a) {
vect2	result;
    result.x = -a.x;
    result.y = -a.y;
    return result;
}

double	dot3vect (vect3 a, vect3 b) {
    return (a.x*b.x + a.y*b.y + a.z*b.z);
}

double	dot2vect (vect2 a, vect2 b) {
    return (a.x*b.x + a.y*b.y);
}

double	vect3norm (vect3 a) {
    return (sqrt(a.x*a.x + a.y*a.y + a.z*a.z));
}

double	vect2norm (vect2 a) {
/*     return (sqrt(a.x*a.x + a.y*a.y)); */
    return  (hypot (a.x, a.y));
}

vect3	norm3vect (vect3 a) {
vect3	result;
double	r;
    r = vect3norm (a);
    if (r == 0.0) return a;
    result.x = a.x/r;
    result.y = a.y/r;
    result.z = a.z/r;
    return  result;
}

vect2	norm2vect (vect2 a) {
vect2	result;
double	r;
/*  r = vect2norm (a);  */
    r = hypot (a.x, a.y);
    if (r == 0.0) return a;
    result.x = a.x/r;
    result.y = a.y/r;
    return  result;
}

vect3	cross3vect (vect3 a, vect3 b) {
vect3	result;
    result.x = a.y*b.z - b.y*a.z;
    result.y = a.z*b.x - b.z*a.x;
    result.z = a.x*b.y - b.x*a.y;
    return result;
}

double	cross2vect (vect2 a, vect2 b) {
    return (a.x*b.y - a.y*b.x);
}

vect2	flip2vect (vect2 a) {
/* Useful for supplemental angle in vec2pol calls, for example */
vect2	result;
    result.x = -a.x;
    result.y =  a.y;
    return  result;
}

pvec2	polar2 (vect2 a) {	/* Convert to polar */
pvec2	result;
/*  retult.r = vect2norm (a);  */
    result.r = hypot (a.y, a.x);
    result.t = atan2 (a.y, a.x);
    return  result;
}

vect2	rect2 (pvec2 a) {	/* Convert to rectangular */
vect2	result;
#ifdef  HAVE_SC
double	s, c;
    sincos (a.t, &s, &c);
    result.x = a.r * c;
    result.y = a.r * s;
#else
    result.x = a.r * cos(a.t);
    result.y = a.r * sin(a.t);
#endif
    return  result;
}

pvec3	polar3 (vect3 a)  {	/* Convert to polar3 */
pvec3	result;
    result.r = hypot (a.y, a.x);
    result.t = atan2 (a.y, a.x);
    result.z = a.z;
    return  result;
}

vect3	rect3 (pvec3 a)  {	/* Convert to rectangular */
vect3	result;
#ifdef  HAVE_SC
double	s, c;
    sincos (a.t, &s, &c);
    result.x = a.r * c;
    result.y = a.r * s;
#else
    result.x = a.r * cos(a.t);
    result.y = a.r * sin(a.t);
#endif
    result.z = a.z;
    return  result;
}

double	znorm (vect3 a)  {	/* Ray tracing special norm */
double	N;

    N = 1.0 - a.x*a.x - a.y*a.y;
    return ( N < 0.0 ? -a.z : sqrt(N) );
}
/* znorm is used in ray tracing; the "cannonical momentum" of a ray
  is a unit vector multiplied by the index of refraction.  When crossing
  a boundry, (z is normal to the boundry) the x and y components remain
  the same but z is changed to normalize to the new index.  Going from
  index i to index r, multiply the normalized unit vector by i/r and
  then apply znorm to find z; then re-normalize the unit vector.  If 
  the resulting z would be immaginary, that is total internal reflection,
  and we return -z instead of the refracted normal z.  */



/* ---------------------------------------------------------------------- */

/*  ---  Geometry Utilities  ---  */

double	dist3 (vect3 a, vect3 b) {
    return  vect3norm (sub3vect (a, b) );
}

double	dist2 (vect2 a, vect2 b) {
    return  vect2norm (sub2vect (a, b) );
}

/* Distance along vector a to nearest point to p */
double	nearp3 (vect3 a, vect3 p) {
double	result;
    result = vect3norm (a);
    if (result > 0.0) result = dot3vect (a, p) / (result*result);
    return  result;
}

double	nearp2 (vect2 a, vect2 p) {
double	result;
    result = vect2norm (a);
    if (result > 0.0) result = dot2vect (a, p) / (result*result);
    return  result;
}

/* Distance from p to segment 0,a; point on line limited by ends */
double	dseg3 (vect3 a, vect3 p) {
double	f;
    f = nearp3 (a, p);
    if (f < 0.0) f = 0.0;
    if (f > 1.0) f = 1.0;
    return  dist3 (mul3vect(a,f), p);
}

double	dseg2 (vect2 a, vect2 p) {
double	f;
    f = nearp2 (a, p);
    if (f < 0.0) f = 0.0;
    if (f > 1.0) f = 1.0;
    return  dist2 (mul2vect(a,f), p);
}


/* Signed perpendicular distance from point p to vector a */
double	pdis2 (vect2 a, vect2 p) {
double	A;
    A = vect2norm (a);
    if (A == 0.0) return  vect2norm (p);
    return  cross2vect (a, p) / A;
/* or.. return  cross2vect (norm2vect(a), p);  ?? */
}

/*  Interval comparison -- integer decision function;
  Return -1, 0 or 1 for x relative to directed interval a,b:
  Result is -1 (a), 0 (inside), +1 (b) side of interval... */
/* Compare is inclusive of the boundry */
int	inside (double a, double x, double b) {
    if (a < b) {
	if (x < a) return -1;
	return x > b ? 1 : 0;
    } else {
	if (x > a) return -1;
	return x < b ? 1 : 0;
    }
}

/*  Minimum distance to parallelogram:  
  A parallelogram is defined by side vectors a and b, from a corner
  at origin (0,0).  Find least distance from point p to this parallelogram,
  with the distance being negative inside the figure.  */
double	dparel (vect2 a, vect2 b,  vect2 p)  {
double	result;
double	d;
int	in;

/* If either side is zero, use distance to segment */
    if (vect2norm (a) == 0.0) return dseg2 (b, p);
    if (vect2norm (b) == 0.0) return dseg2 (a, p);

/* Find if point e is interior to parallelogram - in = 0 if so */
    in = inside (0.0, cross2vect (a, p), cross2vect (a, b) ) |
	 inside (0.0, cross2vect (b, p), cross2vect (b, a) );

/* Find least distance to any of 4 segments bounding parallelogram */
    result = dseg2 (a, p);
    d = dseg2 (b, p);
    if (d < result) result = d;
    d = dseg2 (a, sub2vect (p, b));
    if (d < result) result = d;
    d = dseg2 (b, sub2vect (p, a));
    if (d < result) result = d;
    if (!in) result = -result;
    return  result;
}




/* ---------------------------------------------------------------------- */

/* Spherical geometry vector things */

vect3	rotate (vect3 v, double angle, int axis) {
vect3	result;
double	s, c;
    if (angle == 0.0) return  v;  /* null case */
#ifdef  HAVE_SC
    sincos (angle, &s, &c);
#else
    s = sin (angle);
    c = cos (angle);
#endif
    switch (axis) {
	case 1:
	    result.x = v.x;
	    result.y = v.y*c - v.z*s;
	    result.z = v.z*c + v.y*s;
	    break;
	case 2:
	    result.x = v.x*c + v.z*s;
	    result.y = v.y;
	    result.z = v.z*c - v.x*s;
	    break;
	case 3:
	    result.x = v.x*c - v.y*s;
	    result.y = v.y*c + v.x*s;
	    result.z = v.z;
	    break;
	default:
	    result = v;
    }
    return result;
}

vect2	rot2 (vect2 v, double angle) {
vect2	result;
double	s, c;
#ifdef  HAVE_SC
    sincos (angle, &s, &c);
#else
    s = sin (angle);
    c = cos (angle);
#endif
    result.x = v.x*c - v.y*s;
    result.y = v.y*c + v.x*s;
    return result;
}

vect3	sph2vec (double lng, double lat) {
vect3	result;
#ifdef  HAVE_SC
double	sa, ca;
double	sd, cd;
/*    result = make3vect (cos(lat), 0.0, sin(lat)); */
/*    result = rotate (result, lng, 3); */
    sincos (lng, &sa, &ca);
    sincos (lat, &sd, &cd);
    result.x = cd * ca;
    result.y = cd * sa;
    result.z = sd;
#else
double	cd;
    cd = cos (lat);
    result.x = cd * cos (lng);
    result.y = cd * sin (lng);
    result.z = sin (lat);
#endif
    return result;
}

vect3	had2vec (double lng, double lat) {
/* Similar to sph2vec, using hours and degrees arguments */
vect3	result;
#ifdef  HAVE_SC
double	sa, ca;
double	sd, cd;
/*    result = make3vect (cos(lat), 0.0, sin(lat)); */
/*    result = rotate (result, lng, 3); */
    sincos (lng*Hour,   &sa, &ca);
    sincos (lat*Degree, &sd, &cd);
    result.x = cd * ca;
    result.y = cd * sa;
    result.z = sd;
#else
double	cd;
double	h, d;
    h = lng * Hour;
    d = lat * Degree;
    cd = cos (d);
    result.x = cd * cos (h);
    result.y = cd * sin (h);
    result.z = sin (d);
#endif
    return result;
}


void	vec2sph (double *lng, double *lat, vect3 v) {
    *lng = atan2 (v.y, v.x);
    if (*lng < 0.0) *lng += TwoPi;
    *lat = atan2 (v.z, hypot (v.y, v.x) );
}

void	vec2had (double *lng, double *lat, vect3 v) {
/* Similar to vec2sph, returning hours and degrees arguments. */
    *lng = atan2 (v.y, v.x) / Hour;
    if (*lng < 0.0) *lng += 24.0;
    *lat = atan2 (v.z, hypot (v.y, v.x) ) / Degree;
}

vect2	pol2vec (double r, double a) {
vect2	result;
#ifdef  HAVE_SC
double	s, c;
    sincos (a, &s, &c);
    result.x = r * c;
    result.y = r * s;
#else
    result.x = r * cos (a);
    result.y = r * sin (a);
#endif
    return  result;
}

void	vec2pol (double *r, double *a, vect2 v) {
/*  *r = vect2norm (v);  */
    *r = hypot (v.y, v.x);
    *a = atan2 (v.y, v.x);
    if (*a < 0.0) *a += TwoPi;
}

void	disp3vect (char *left, vect3 v, char *right) {
    printf ("%s%12.9f %12.9f %12.9f%s",
	left, v.x, v.y, v.z, right );
}

void	disp2vect (char *left, vect2 v, char *right) {
    printf ("%s%12.9f %12.9f%s",
	left, v.x, v.y, right );
}

/* 
  The tangent plane coordinate system here has X along the reference
  vector direction, Y a tangent pointing east, and Z a tangent pointing
  north.  We can transform both to the tangent (tang3vec) and from it
  (vec3tang).  This is used to project points to a tangent plane with
  a center in the reference direction.  For multiple points, a projection
  matrix may be defined, in the same way as a precession matrix below.

  For a projected vector in tangent plane coordinates, one may plot
  the x,y (right, up) pair as (-v.y/v.x, v.z/v.x) in radian units;
  scale appropriately to effective radius.
*/

vect3	tang3vec (vect3 v, double lng, double lat) {
/* From vector to tangent plane */
vect3	result;
    result = rotate (v, -lng, 3);
    result = rotate (result, lat, 2);
    return result;
}

vect2	tang2vec (vect3 v) {
/* From tangent plane vector, find 2-vector in plane */
vect2	result;
    result.x = -v.y/v.x;
    result.y =  v.z/v.x;
    return  result;
}

vect3	vec2tang (vect2 v) {
/* Inverse of above (normalization is tricky; the 2-vector
  MUST be in radian units here) */
vect3	result;
    result.y = -v.x;
    result.z =  v.y;
    result.x = 1.0;
    return (norm3vect(result));
}

vect3	vec3tang (vect3 v, double lng, double lat) {
/* From tangent plane to vector; transform X to ref. vector */
vect3	result;
    result = rotate (v, -lat, 2);
    result = rotate (result, lng, 3);
    return result;
}

/* NEED -- DERIVE TANGENT PROJECTION MATRIX FROM EITHER A REFERENCE
 DIRECTION (A,D) OR A REFERENCE VECTOR. */

void	smatrix (vect3 *a, vect3 v) {
/* Obtain projection matrix a from reference vector v */
    a[0] = norm3vect(v);	/* Along reference vector */
    a[1] = norm3vect (make3vect(-v.y, v.x, 0.0));	/* East */
/*  if (vect3norm(a[1]) == 0.0) a[1] = make3vect (0.0, 1.0, 0.0); */
    if (vect3norm(a[1]) < 0.001) a[1] = make3vect (0.0, 1.0, 0.0);
    a[2] = norm3vect (cross3vect(a[0], a[1]));	/* North */
}

void	mrotate (vect3 *a, double angle, int axis) {
/* Rotate the transform in matrix a by the angle about axis... */
vect3	v, w;
double	s, c;
    if (angle == 0.0) return;  /* null case */
#ifdef  HAVE_SC
    sincos (angle, &s, &c);
#else
    s = sin (angle);
    c = cos (angle);
#endif
    switch (axis) {
	case 1:
	    v = a[1];
	    w = a[2];
	    a[1] = sub3vect ( mul3vect(v,c), mul3vect(w,s) );
	    a[2] = sum3vect ( mul3vect(w,c), mul3vect(v,s) );
/* note -- we may need to store each mul3vect result independently
since they are made up in local storage in that program, and could
conflict during execution... */
	    break;
	case 2:
	    v = a[0];
	    w = a[2];
	    a[0] = sum3vect ( mul3vect(v,c), mul3vect(w,s) );
	    a[2] = sub3vect ( mul3vect(w,c), mul3vect(v,s) );
	    break;
	case 3:
	    v = a[0];
	    w = a[1];
	    a[0] = sub3vect ( mul3vect(v,c), mul3vect(w,s) );
	    a[1] = sum3vect ( mul3vect(w,c), mul3vect(v,s) );
	    break;
	default:
	    ;
    }
}

/* ---------------------------------------------------------------------- */

/*  Geometry of Alt-Az Telescope  */
/* For the supprot of alt-az telescopes, and instruments mounted
  at the Naysmith focus, we add some basic telescope geometry.
  The Parallactic Angle is defined as the angle at an observed
  object from the direction to the North Celestial Pole to the
  current Zenith, measured counterclockwize from North through East.
  The Zenith Distance is the angle between the current Zenith and
  the object being observed.
  The Azimuth is the angle measured from North on the horizon,
  through East to the object being observed.

  When the object's position vector in celestial coordinates is rotated
  about the polar axis by siderial time, we get a position vector
  based on current hour angle.  This rotated by latitude gives a
  position vector in altitude and azimuth.

  The current zenith can be rendered as a celestial position, its
  right ascension is the sideral time, and its declination is the
  latitude.

  To find both zeinth distance and parallactic angle, find the current
  zenith vector in a coordinate system containing unit vectors along
  the object direction, the tangent vector pointing north at the object,
  and the tangent vector pointing east.  The zenith vector's position
  in a coordinate system centered on the object may be found by starting
  with the negative hour angle and latitude location and rotating by
  the object's declination.
.. */

/* Solve entire p-z-s navigation triangle at one step */
void	dntri ( double* az, double* zd, double* pa, 
		double* da, double* dz, double* dp,
		double  h,  double  d,  double  g)
/* Finds az=azimuth, zd=zeinth_dist, pa=parallactic and
   the time (hour angle) derivatives da, dz, dp from
   h=hour_angle, d=declination, g=latitude.
   All angles are in radians!                  */
{
double  st, ct;
double  sp, cp;
double  sd, cd;
double  x, y, z;
double	e, n;
//double	dx, dy;
//double	de, dn;
double	s;
double	r, rr;

/* Look up all trug functions first. */
#ifdef  HAVE_SC
    sincos (h, &st, &ct);	/* Hour Angle  */
    sincos (d, &sd, &cd);	/* Declination */
    sincos (g, &sp, &cp);	/* Latitude    */
#else
    st = sin (h);
    ct = cos (h);
    sd = sin (d);
    cd = cos (d);
    sp = sin (g);
    cp = cos (g);
#endif

/* Find vector at star position pointing to zenith */
    x = cp*st;			/* East tangent  */
    y = sp*cd - cp*sd*ct;	/* North tangent */
    z = sp*sd + cp*cd*ct;	/* Toward star   */

/* Find star vector in horizon system */
    e = -cd*st;			/* east  */
    n = cp*sd - sp*cd*ct;	/* north */
/* z is the same in both systems, star <dot> vertical */

/* Find component perpendicular to zenith, star vector */
    s = hypot (x, y);	/* Or, (e,n) as well */

/* Compute angles from vector components */
    *az = atan2 (e, n);
    *zd = atan2 (s, z);
    *pa = atan2 (x, y);

/* Now, compute the derivatives */
/* Limit size of derivatives by limiting size of 1/s.  At a
  distance of 20.6265 arc seconds, s is 1.0e-4 */
#define  MINIMX 1.0e-4
    r = 1.0 / ( (s < MINIMX) ? MINIMX : s );
    rr = r * r;

    *da =  cd * y * rr;
    *dz =  cd * x * r;  /* or: -cp * e * r */
    *dp = -cp * n * rr;

/* The older method was:
    de = -cd*ct;
    dn = sp*cd*st;
    dx = cp*ct;
    dy = cp*sd*st;

    *da = (n*de - e*dn) * rr;
    *dz = cp*cd*st * r;
    *dp = (y*dx - x*dy) * rr;
..*/
}


/* ---------------------------------------------------------------------- */

/* Refraction things */

vect2	difref (vect2 fp, vect2 zp)
{
/* Given zenith point, having magnitude of zenith distance in radians;
  and a field point, having magnitude of tangent plane coordinate in
  radians; will return the differential refraction displacement of the
  field point in tangent plane coordinates.  This result is to be
  multiplied by the actual refractive index.
  NOTE that this result is independent of tangent plane coordinate
  system, whether it is right handed or left handed, and of the 
  actual refraction.  The parallactic angle is in the direction of
  the zenith point here.
  This subroutine is intended to be a general solution to the problem
  of differential refraction in many applications.               ... */

vect2	dr;		/* The result  */
double	z1, z2;		/* zenith distances in radians  */
double	p1, p2;		/* paralactic angles at center, fp  */
double	r, g;		/* polar coordinates of fp  */
double	B, cz;		/* auxiliary angles in triangle  */
double	sr, cr;		/* sin,cos of r  */
double	sz1, cz1;	/* sin,cos of z1  */
double	sA, cA;		/* sin,cos of auxiliary angle A  */
vect2	r1, r2;		/* refraction displacement at center, fp  */

    vec2pol (&z1, &p1, zp);	/* z and parallactic at center  */
    r1 = pol2vec (tan(z1), p1);	/* refraction of center  */
/* We could substitute, both here and in the computation of r2 below,
  a better approximation than tan(z1) should it be desired... */
    vec2pol (&r, &g, fp);	/* Plane polar coords of fp  */

/* We have a spherical triangle; its corners are:
	A : the center of the field
	B : the field point
	C : the zenith
and we have the equations, in which the lower case letters are the
oposite sides of the triangle:
	y = sin a sin B = sin b sin A
	x = sin a cos B = cos b sin c - sin b cos c cos A
	z = cos a       = cos b cos c + sin b sin c cos A
and we substitute the needed values as:
	a : z2,  b : z1,  c : r.
	A : angle at center = p1 - g;
	B : angle at field point: p2 + B = g + Pi;
	thus p2 = g + Pi - B.
Using vec2sph we derive B and cz from the x,y,z values above,
and note that cz is the complement of z2.		... */

#ifdef  HAVE_SC
    sincos (z1, &sz1, &cz1);	/* side b  */
    sincos ( r,  &sr,  &cr);	/* side c  */
    sincos (p1-g, &sA, &cA);	/* Angle A  */
#else
    sz1 = sin (z1);
    cz1 = cos (z1);
    sr  = sin (r);
    cr  = cos (r);
    sA = sin (p1-g);
    cA = cos (p1-g);
#endif
    vec2sph ( &B, &cz,  make3vect (
	cz1 * sr - sz1 * cr * cA,
	sz1 * sA,
	cz1 * cr + sz1 * sr * cA)  );
    p2 = g + M_PI - B;		/* Parallactic of fp  */
    z2 = M_PI/2.0 - cz;		/* Actual z2  */
    r2 = pol2vec (tan(z2), p2);	/* Refraction at fp  */
    dr = sub2vect (r2, r1);	/* Differential Refraction  */
/* The correction for general refractive scale should be done outside
  this program as and if needed.  If the geometric position is input
  the result here will properly include the refractive scale effect.
  If this effect is not desired, or has already been included in the
  telescope focal length by calibrating the telescope inside the
  atmosphere by observing stars and using their geometrical positions,
  one should correct the result by adding fp, or fp multiplied by the
  air mass at which the telescope focal length was calibrated, to the
  result returned here.  This result is intended to be multiplied by
  the refraction constant to give the angular differential refraction.
  The "ndxref" program (below) may be used to obtain this constant.    */
//    dr = sum2vect (dr, fp);	/* Correct for refractive scale  */
    return  dr;
}

double	ndxref (double wvlen, double p, double t)
/* Approximate index of refraction of atmosphere */
/* Wavelength in angstroms, pressure in mm-hg, temp. in Celsius */
/* A good visual value of wvlen is 5500.0, and a good approximation
  for p at Las Campanas is 575.0  */
/* Returns index less 1.0 */
{
double	rwsq, q, c;

/* Refraction from Allen, Astrophysical Quantities, 3rd. edition */
    rwsq = (1.0e+8) / (wvlen*wvlen);
    q = 64.328 + 29498.1/(146.0-rwsq) + 255.4/(41.0-rwsq);
    c = p*(1.0 + (1.049-0.0157*t)*p*1.0e-6) / (720.88261*(1.0+0.003661*t));
    return (q*c*1.0e-6);
}
/* Using wvlen 5500, rwsq = 3.3, q = 277.8, and t = 0, p = 575.0,
  c = .798, refn = 221.737e-6;  .01/refn = .022 radian. */


vect3	refracted (vect3 v, double cr, int z)  {
/* Find refracted (horizon) vector fron non-refracted */
/* v is a 3 vector in horizon coordinates */
/* cr is coefecient of refraction, probably determined by ndxref above */
/* z is subscript of vertical component of v; 0=x, 1=y, 2=z */
#define  MINZ  0.01
#define  MAXP  (cr/MINZ)
vect3	r;
    r = norm3vect (v);
/*    if (z < 0 || z > 2) return r;  */
/*    if (r[z] < 0.01) r[z] += 0.1;  */
/*    else  r[z] += (cr / r[z]);  */
    switch  (z) {
	case 0:
	    if (r.x < MINZ)  r.x += MAXP;
	    else  r.x += (cr / r.x);
	    break;
	case 1:
	    if (r.y < MINZ)  r.y += MAXP;
	    else  r.y += (cr / r.y);
	    break;
	case 2:
	    if (r.z < MINZ)  r.z += MAXP;
	    else  r.z += (cr / r.z);
	    break;
	default:
	    return  r;
    }
    return  norm3vect (r);
}
/* Note refracted is not same refraction approximation used in
  the difref program above; yet. */


/* ---------------------------------------------------------------------- */

/* Precession things */

vect3	prec (double e, vect3 pf, double ef)
/* Result is pf at epoch ef precessed to epoch e. */
{
/* Compute vector position <p> at epoch <e> from position
   <pf> at epoch <ef>.
   Rigorous method -- precessional motion only.
   Uses IAU (1976) Precession constants
        (Astronomy and Astrophysics vol. 73, 282-284 (1979) )
   Epochs are in Julian years.  JE = 2000.0 + (JED - 2451545.0)/365.25
   Method is E.S.A.E. pages 30-31.   */
vect3	p;
double	t, tn;
static	double	zeta = 0.0;
static	double	z    = 0.0;
static	double	thta = 0.0;
static	double	es   = 0.0;
static	double	efs  = 0.0;

/* If same epochs as last time, skip the angle finding... */
    if ( (e != es) || (ef != efs) ) {
	t = (e - ef) * 1.0e-2;
	tn = (ef - 2000.0) * 1.0e-2;
     
	zeta = t * (11.1808609e-3 + tn*(6.77071e-6 - tn*674.e-12)
		    + t * (1.463556e-6 - tn*1.668e-9 + t*87.257e-9));
	z = zeta + t*t*(3.843603e-6 + tn*1.988e-9 + t*994.e-12);
	thta = t * (9.7171735e-3 - tn*(4.13692e-6 + tn*1.052e-9)
		    - t * (2.06846e-6 + tn*1.052e-9 + t*202.812e-9));
	es  = e;
	efs = ef;
    }

    p = rotate (pf, zeta, 3);
    p = rotate (p, -thta, 2);
    p = rotate (p, z, 3);
    return  p;
}

void	pprec (double *a, double *d, double e,
	double a1, double d1, double e1)  {
/* Precess from position a1, d1 at e1 to *a, *d at e.  Positions are
  in radians, and epoch in Julian years */
vect3	p;
    p = prec (e, sph2vec (a1, d1), e1);
    vec2sph (a, d, p);
}

void	hdprec (double *a, double *d, double e,
	double a1, double d1, double e1)  {
/* Precess from position a1, d1 at e1 to *a, *d at e.  Positions are
  in hours and degrees, and epoch in Julian years */
vect3	p;
    p = prec (e, had2vec (a1, d1), e1);
    vec2had (a, d, p);
}


/* ---------------------------------------------------------------------- */

/* (Simulated) Matrix things */

/* A matrix is an array of 3 3-vectors, or 2 2-vectors.  It is passed  by
  a pointer reference, and can be transposed, multiplied, copied, etc.  */

vect3	mmult3 (vect3 *a, vect3 v) {
vect3	result;
    result.x = dot3vect (a[0], v);
    result.y = dot3vect (a[1], v);
    result.z = dot3vect (a[2], v);
    return  result;
}

vect3	tmult3 (vect3* a, vect3 v) {
/* Multiply transpose of a by vector v */
vect3	result;
    result.x = a[0].x * v.x + a[1].x * v.y + a[2].x * v.z;
    result.y = a[0].y * v.x + a[1].y * v.y + a[2].y * v.z;
    result.z = a[0].z * v.x + a[1].z * v.y + a[2].z * v.z;
    return  result;
}


vect2	mmult2 (vect2 *a, vect2 v) {
vect2	result;
    result.x = dot2vect (a[0], v);
    result.y = dot2vect (a[1], v);
    return  result;
}

void	trpose3 (vect3 *a) {
/* Transpose a 3-matrix in place; this will invert a 3-d
  rotation matrix. */
double	t;
    t = a[0].y;   a[0].y = a[1].x;   a[1].x = t;
    t = a[0].z;   a[0].z = a[2].x;   a[2].x = t;
    t = a[1].z;   a[1].z = a[2].y;   a[2].y = t;
}

void	trpose2 (vect2 *a) {
/* Transpose a 2-matrix in place; this will invert a 2-d
  rotation matrix. */
double	t;
    t = a[0].y;   a[0].y = a[1].x;   a[1].x = t;
}

void	mcpy3 (vect3 *a, vect3 *b) {
/* Copy matrix b to location a */
int	i;
    for (i=0; i<3; i++) a[i] = b[i];
}

void	mcpy2 (vect2 *a, vect2 *b) {
/* Copy matrix b to location a */
int	i;
    for (i=0; i<2; i++) a[i] = b[i];
}

void	pmatrix (vect3 *p, double e, double ef) {
/* Make a 3-matrix p which will cause precession transformation to
  equinox e from equinox ef when multiplied by an equinox ef position. */
    p[0] = prec (e, make3vect (1.0, 0.0, 0.0), ef);
    p[1] = prec (e, make3vect (0.0, 1.0, 0.0), ef);
    p[2] = prec (e, make3vect (0.0, 0.0, 1.0), ef);
/* Transform the matrix to obtain correct multiplication.  */
    trpose3 (p);
/* To transform vector v1 at epoch e1 to vector v2 at epoch e2, do:
    vect3  pm[3];
    pmatrix (pm, e2, e1);
    v2 = mmult3 (pm, v1);       */
}

void	r2matrix (vect2 *a, double angle) {
/* Form a 2-d rotation matrix */
#ifdef  HAVE_SC
    sincos (angle, &(a[1].x), &(a[0].x));
#else
    a[0].x = cos (angle);
    a[1].x = sin (angle);
#endif
    a[0].y = -a[1].x;
    a[1].y =  a[0].x;
}

void	mxm3a (vect3* a, vect3* b, vect3* c) {
/* a = b x c, matrix wise... */
    a[0].x = b[0].x * c[0].x + b[0].y * c[1].x + b[0].z * c[2].x;
    a[0].y = b[0].x * c[0].y + b[0].y * c[1].y + b[0].z * c[2].y;
    a[0].z = b[0].x * c[0].z + b[0].y * c[1].z + b[0].z * c[2].z;
    a[1].x = b[1].x * c[0].x + b[1].y * c[1].x + b[1].z * c[2].x;
    a[1].y = b[1].x * c[0].y + b[1].y * c[1].y + b[1].z * c[2].y;
    a[1].z = b[1].x * c[0].z + b[1].y * c[1].z + b[1].z * c[2].z;
    a[2].x = b[2].x * c[0].x + b[2].y * c[1].x + b[2].z * c[2].x;
    a[2].y = b[2].x * c[0].y + b[2].y * c[1].y + b[2].z * c[2].y;
    a[2].z = b[2].x * c[0].z + b[2].y * c[1].z + b[2].z * c[2].z;
}

void	mxm3 (vect3* a, vect3* b, vect3* c) {
/* a = b x c, matrix wise... */
int	i;
    for (i=0; i<3; i++) {
	a[i].x = b[i].x * c[0].x + b[i].y * c[1].x + b[i].z * c[2].x;
	a[i].y = b[i].x * c[0].y + b[i].y * c[1].y + b[i].z * c[2].y;
	a[i].z = b[i].x * c[0].z + b[i].y * c[1].z + b[i].z * c[2].z;
    }
}

/* To both precess and project, we make a precession matrix b and
  a projection matrix a, and then the combined matrix c:
    smatrix (a, v)	-- projection, done second
    pmatrix (b, e, ecat)  -- precession, done first
    mxm3 (c, a, b)	-- combined precession and projection
.. */

/* ---------------------------------------------------------------------- */

/* Real Matrix things */

/* A (real) matrix is a structure, of 2 vect2 or 3 vect3 elements,
  labeled appropriately.  Functions will return a matrix value,
  with all the usual (needed) operations.  */


/* -- 2-dimensional Matrix things -- */

vect2	mmul2v (matrix2  mx,  vect2  v) {
/* Return the vector product of a 2-matrix and a 2-vector  */
vect2	result;
    result.x = dot2vect (mx.x, v);
    result.y = dot2vect (mx.y, v);
    return  result;
}

matrix2	mtr2 (matrix2 a) {
/* Return the transpose of a 2-matrix */
matrix2	result;
    result.x = make2vect (a.x.x, a.y.x);
    result.y = make2vect (a.x.y, a.y.y);
    return  result;
}

matrix2	minv2 (matrix2 a) {
/* Return the inverse of a 2-matrix */
matrix2	result;
double	d;
    d = a.x.x * a.y.y - a.x.y * a.y.x;
/* or d = cross2vect (a.x, a.y) */
    if (d == 0.0) {	/* Singular, return identity */
	result.x.x = result.y.y = 1.0;
	result.x.y = result.y.x = 0.0;
    } else {
	result.x.x =  a.y.y / d;
	result.y.y =  a.x.x / d;
	result.x.y = -a.x.y / d;
	result.y.x = -a.y.x / d;
/*	result.x.y = -a.y.x / d;  */
/*	result.y.x = -a.x.y / d;  */
    }
    return  result;
}

matrix2	mul2m (matrix2 a, matrix2 b) {
/* Standard matrix multiplication; direct. */
matrix2	result;
    result.x.x = a.x.x * b.x.x + a.x.y * b.y.x;
    result.x.y = a.x.x * b.x.y + a.x.y * b.y.y;
    result.y.x = a.y.x * b.x.x + a.y.y * b.y.x;
    result.y.y = a.y.x * b.x.y + a.y.y * b.y.y;
    return  result;
}

matrix2	rot2m (double a) {
/* Create a 2-d rotation matrix for angle a */
matrix2	result;
#ifdef  HAVE_SC
    sincos (a, &(result.y.x), &(result.x.x));
#else
    result.x.x = cos (a);
    result.y.x = sin (a);
#endif
    result.x.y = -result.y.x;
    result.y.y =  result.x.x;
    return  result;
}


/* -- 3-dimensional Matrix things -- */

vect3	mmul3v (matrix3  mx,  vect3  v) {
/* Return the vector product of a 3-matrix and a 3-vector  */
vect3	result;
    result.x = dot3vect (mx.x, v);
    result.y = dot3vect (mx.y, v);
    result.z = dot3vect (mx.z, v);
    return  result;
}

matrix3	mtr3 (matrix3 a) {
/* Return the transpose of a 3-matrix */
matrix3	result;
    result.x = make3vect (a.x.x, a.y.x, a.z.x);
    result.y = make3vect (a.x.y, a.y.y, a.z.y);
    result.z = make3vect (a.x.z, a.y.z, a.z.z);
    return  result;
}

/* PLEASE NOTE that for an orthonormal (rotation) matrix, the inverse
  and the transform are theoretically identical.  */

matrix3	minv3 (matrix3 a) {
/* Return the inverse of a 3-matrix, direct computation. */
matrix3	result;
matrix3 c;	/* The conjugate matrix */
double	d;	/* Determinant */

/* Find the conjugate matrix using cross products */
    c.x = cross3vect (a.y, a.z);
    c.y = cross3vect (a.z, a.x);
    c.z = cross3vect (a.x, a.y);
    d = dot3vect (a.x, c.x);	/* Determinant; y and z rows work too */
    if (d == 0.0) return rot3m (0.0, 0);	/* Return identity */
    d = (d == 0.0) ? 1.0 : 1.0 / d;	/* Scale by 1/d */
    result.x = mul3vect (c.x, d);
    result.y = mul3vect (c.y, d);
    result.z = mul3vect (c.z, d);
    return  mtr3 (result);	/* Actual result is transpose here */
}

matrix3	mul3m (matrix3 a, matrix3 b) {
/* Standard matrix multiplication, implemented by vectors */
matrix3	result;
matrix3	t;
    t = mtr3(b);
    result.x = mmul3v (a, t.x);
    result.y = mmul3v (a, t.y);
    result.z = mmul3v (a, t.z);
    return  mtr3(result);
}

matrix3	rot3m (double angle, int axis) {
/* Form rotation matrix which when multiplied by a vector will have
  the identical effect as the rotate subroutine */
/* If axis is not 1,2,3 the identity matrix is returned. */
matrix3	result;
double	s, c;
#ifdef  HAVE_SC
    sincos (angle, &s, &c);
#else
    s = sin (angle);
    c = cos (angle);
#endif
    switch (axis) {
	case 1:
	    result.x = make3vect (1.0, 0.0, 0.0);
	    result.y = make3vect (0.0,   c,  -s);
	    result.z = make3vect (0.0,   s,   c);
	    break;
	case 2:
	    result.x = make3vect (  c, 0.0,   s);
	    result.y = make3vect (0.0, 1.0, 0.0);
	    result.z = make3vect ( -s, 0.0,   c);
	    break;
	case 3:
	    result.x = make3vect (  c,  -s, 0.0);
	    result.y = make3vect (  s,   c, 0.0);
	    result.z = make3vect (0.0, 0.0, 1.0);
	    break;
	default:
	    result.x = make3vect (1.0, 0.0, 0.0);
	    result.y = make3vect (0.0, 1.0, 0.0);
	    result.z = make3vect (0.0, 0.0, 1.0);
    }
    return  result;
}

matrix3	precmat (double e, double ef) {
/* Return the 3-matrix p which will cause precession transformation to
  equinox e from equinox ef when multiplied by an equinox ef position. */
/* Update of pmatrix routine above. */
matrix3	result;
    result.x = prec (e, make3vect (1.0, 0.0, 0.0), ef);
    result.y = prec (e, make3vect (0.0, 1.0, 0.0), ef);
    result.z = prec (e, make3vect (0.0, 0.0, 1.0), ef);
/* This actually gives the transform of what is desired, so... */
    return  mtr3(result);
/* To transform vector v1 at epoch e1 to vector v2 at epoch e2, do:
    v2 = mmul3v (precmat(e2, e1), v1);       */
}

matrix3	precmat3 (double e, double ef) {
/* Return a precession matrix directly formed from precession constants. */
matrix3	result;
double	t, tn;
static	double	zeta = 0.0;
static	double	z    = 0.0;
static	double	thta = 0.0;
static	double	es   = 0.0;
static	double	efs  = 0.0;

/* Angles found exactly as in the prec subroutine */
/* If same epochs as last time, skip the angle finding... */
    if ( (e != es) || (ef != efs) ) {
	t = (e - ef) * 1.0e-2;
	tn = (ef - 2000.0) * 1.0e-2;
     
	zeta = t * (11.1808609e-3 + tn*(6.77071e-6 - tn*674.e-12)
		    + t * (1.463556e-6 - tn*1.668e-9 + t*87.257e-9));
	z = zeta + t*t*(3.843603e-6 + tn*1.988e-9 + t*994.e-12);
	thta = t * (9.7171735e-3 - tn*(4.13692e-6 + tn*1.052e-9)
		    - t * (2.06846e-6 + tn*1.052e-9 + t*202.812e-9));
	es  = e;
	efs = ef;
    }

/* Method 1...
    result = mul3m (rot3m (z, 3), rot3m (-thta, 2));
    result = mul3m (result, rot3m (zeta, 3));
.. */

/* Method 2 */
    result = mul3m (rot3m (-thta, 2), rot3m (zeta, 3));
    result = mul3m (rot3m (z, 3), result);

/* Method 3 ...
    result = rot3m (zeta, 3);
    result = mul3m (rot3m (-thta, 2), result);
    result = mul3m (rot3m (z, 3), result);
..*/

/* All 3 methods give the same result. */

    return  result;
}


/* ---------------------------------------------------------------------- */

/* R.A. overlap utilities */

double	fdifmod (double a, double b, double c) {
/* Difference limited to +/- c/2 */
double	result;
double	h;
    result = a - b;
    h = fabs (c / 2.0);
    while (result >  h) result -= c;
    while (result < -h) result += c;
    return  result;
}

double  fmodsum (double a, double b, double c) {
/* Sum, put into range 0<= x <c */
double	result;
double	h;
    result = a + b;
    h = fabs (c);
    while (result >=  h) result -= c;
    while (result < 0.0) result += c;
    return  result;
}

double	fabsdif (double a, double b, double c) {
/* Absolute value of difference in closest direction */
    return  fabs ( fdifmod(a, b, c));
}


/* ---------------------------------------------------------------------- */

/*  --  Arithmetic  --  */

int	Nint (double a)
/* Return nearest integer */
{
#ifdef  __sun
    return  nint(a);
#else
//    return  (int)round(a);
//    return  (int)nearbyint(a);
    return  (int)rint(a);
/* Could also use (int)rint(a), but nearbyint will not raise
  the inexact exception.  Uses current rounding direction.
  This is a c99 standard function.  */
#endif
}

void	qrem (long int* q, long int* r,  long int n, long int d)
/* General quotent+remainder code.  Division of numerator n by
  divisor d results in quotent q and remainder r; such that
  n = q*d + r, r is the same sign as d, and r is the minimum
  such remainder.    */
{
//ldiv_t	A;
int	s;
    if (d == 0) {	// correct for any divide by zero
	*q = n;   *r = 0;
	return;
    }
//    A = ldiv (n, d);
//    *q = A.quot;
//    *r = A.rem;
/* The ldiv function appears not to work.  Do it ourselves. */
    s = (d > 0) ? 1 : -1;
    *q = n / d;
    *r = n - *q * d;
    if ( *r * s < 0 ) { *r += d;  (*q)--; }
}

long int  mod (long int n, long int d)
/* Positive modulo function.  Divide n by d, and return the
  remainder; returned remainder will be of same sign as d.  
  This is a simplified sub-set of qrem above.    */
{
long int  r=0;
int	s;
    if (d == 0)  return r;  // for divide-by-zero error
    s = (d > 0) ? 1 : -1;
    r = n - ( (n/d) * d );  // do not save the quotent
    if ( r * s < 0 ) r += d;  // assure sign of remainder
    return  r;
}


/* ---------------------------------------------------------------------- */

/*  ==  Constellation Names  ==  */

int	ConIndex (double ra, double dec, double eq)
/* Return subscript for constellation in which the given ra, dec position
  (in hours and degrees and equinox eq years) is.  0 if not found.
  Uses Tbound array defined above.     */
{
double	a, d;
int	i, j, k;

// Precess the given position to 1875.0, convert to Qmin units for table
    hdprec (&a, &d, 1875.0,  ra, dec, eq);
    a *= 3600.0;
    d *= 240.0;

// Do binary search to find starting value for i
    i = Fbound;  k = 0;
    while ( (j = (i+k)/2) > k ) {
	if (d < (double)Tbound[j].dc) k = j;  // same as first continue below
	else    i = j;
    }
/* Above binary search algorithm assumes that test dec < list[0].dc will
  be true (as list[0] starts above +90.0) and that dec < list[n].dc will
  be false as list[n] is -90.0 or less.  If so, i will be set to the
  lowest value for which dec < list[i].dc is false.  It is also assumed
  that the list is sorted in decreasing declination.    */

// Do the search of bound list
    for (; i<Fbound; i++) {
//	if (d <  (double)Tbound[i].dc) continue;
	if (a <  (double)Tbound[i].lo) continue;
	if (a >= (double)Tbound[i].hi) continue;
// The southern boundry of required constellation has been found
	return  Tbound[i].cs;
    }
    return  0;
}

char*	English (int x)
/* Return a buffer containing the translated name for printing,
  given the constellation index as found above.  The buffer returned
  is in dynamic storage and should be freed after use.  */
{
int	j=0;
char	*c;
char	buf[32];	// minimum 29 required
char	*r;

    if (x < 0 || x > Ncon) {
	*buf = '\0';
    } else if (strlen(Cnam[x].e) == 0) {
	*buf = '\0';
    } else {
	if (Cnam[x].e[0] == '*') j = 1;
	c = Cnam[x].e + j;
	if (*c == ' ' || *c == '\0') c = Cnam[x].n;	// Substitute constellation
	sprintf (buf, "[%s%s]", j ? "" : "The ", c);
    }
    r = strdup (buf);
    return  r;
}

void	ConstellNames (char* buf, double ra, double dec, double eq)
/*  Print names into the given buffer.  The buffer should be at
  least 51 characters long.  */
{
int	ic;
char	*e;

    ic = ConIndex (ra, dec, eq);
    e = English (ic);

//    sprintf (buf, "%s : %s %s\n", Cnam[ic].a, Cnam[ic].n, e );
// If using the above format, maximum 58 characters needed.
    sprintf (buf, "%s %s", Cnam[ic].n, e );
/* Remember to free the dynamic buffer returned from English. */
    free (e);
}


/* ---------------------------------------------------------------------- */


/* -- Simple Time Programs -- */

double  jdnow (void)
/* Return current time as J.D. */
{
static long int        time_value;
/* static double	sysepoch = 2440587.5; */
/* 2440587.5 is J.D. at epoch of system time, 0h GMT on 1 Jan., 1970 */
/* SysEpoch is now defined in the kdcutil.h file. */
double          d;

    time (&time_value);		/* Seconds since the SysEpoch */
    d = (double)time_value / 86400.0;	/* Convert time to days */
    return (d + SysEpoch);	/* Convert clock time directly to J.D. */
}

/* Note that the returned JD and its associated GMT will be correct;
  however, local time would need the current time zone. */


void	jd2date (double jd, int* year, int* month, int* day)
/* Convert jd to Gregorian year, month and day (GMT date)
  Handles years less than 1 correctly.    */
{
long int        j, y, d, m;
long int	p, q;

    j = (long)(jd-0.5) - 1721118;	/* De-bias JD */
// Gj2date == Gregorian calendar, jd to y-m-d code
    qrem (&p, &q, j, 146097);	/* Gregorian period and 4-centuries */
    qrem (&y, &j, 4*q-1, 146097);	/* Century and quarter-days */
    y += 4*p;			/* Add Gregorian centuries of p */
    d = j/4;			/* Days in century */
    qrem (&j, &d, 4*d+3, 1461);   /* Year and quarter days */
    d = (d+4)/4;		/* Days */
    qrem (&m, &d, 5*d-3, 153);   /* Month and 1/5 days in month */
    d = (d+5)/5;		/* Day in month */
    y = 100*y + j;		/* Combine year, century */

    qrem (&p, &q, m+2, 12);
    *year = y + p;		/* Correct year */
    *month = 1 + q;		/* Correct month */
    *day = d;			/* Day of month */
    return;
}

double	jdate ( int year, int month, int day)
/* Convert Gregorian y-m-d to an actual J.D. at 0h GMT
  Correctly handles years less than 1.    */
{

long int  j, c;
long int  jy, jm;
long int  p, q, ky;

// Gdate2jd -- Gregorian calendar date to JD code
    qrem (&ky, &jm, month+9, 12);
    ky += year - 1;

    qrem (&c, &jy, ky, 100);	/* Century, year */
    qrem (&p, &q, c, 4);	/* Gregorian period, sub-century */
    j = 146097*p + 36524*q + (1461*jy)/4 + (153*jm+2)/5 + day + 1721118;

    return (double)j + 0.5;
}

/*  ===  Definitions for Julian Day conversions  ===  */
/* #define  JULIAN_YEAR  365.25  */
/* #define  JULIAN_EPOCH 1721045.0  */
/* #define  JYEAR(jd)  ((jd - JULIAN_EPOCH)/JULIAN_YEAR)  */
	/* JYEAR converts a julian date to standard epoch, years */
/* #define  JDEPOCH(epoch) (JULIAN_EPOCH + epoch*JULIAN_YEAR)  */
	/* JDEPOCH converts a standard epoch to julian date */
/* These are converted to subroutines below... */

double	Jyear (double  jd)
/* Convert a J.D. to a Julian Year epoch */
{
double	y;
    y = (jd - 1721045.0) / 365.25;
    return  y;
}

double	jdaye (double  year)
/* Convert a Julian Year epoch to a J.D. */
{
double	y;
    y = 1721045.0 + year * 365.25;
    return  y;
}


double	gst0 (double  jd)
/* Return the GST at 0 hours for the current J.D.  Just add the
  local solar time to get sidereal time.  Result in days. */
/* Also, lst = gst + east longitude */
{
double  tc;
double  gst;

    tc = (jd - 2451545.0) / 36525.0;
/* tc is time in Julian Centuries from the standard epoch 2000.0 */
    gst = 24110.54841
	+ tc*(8640184.812866
	+ tc*(0.093104
	- tc*6.2e-6));	/* GMST - UT1, seconds */
    gst /= 86400.0;	/* ST dif, days */
    gst -= (int)gst;
    if (gst < 0.0) gst += 1.0;	/* gst in days now */
    return  gst;
}


/* ---------------------------------------------------------------------- */


/*  ----  Support for Date structure  ----  */


static long int	jdG (int year, int month, int day)
/* Convert year, month and day of Gregorian calendar to
  the integral part of Julian Date for 0h GMT.  Add
  0.5 days to obtain true Julian Date.    */
{
long int  j, c;
long int  ky, jy, jm;

long int  p, q;
//long int  x, y;

// Gdate2jd -- Gregorian calendar date to JD code
    qrem (&ky, &jm, month+9, 12);
    ky += year - 1;

    qrem (&c, &jy, ky, 100);	/* Century, year */
    qrem (&p, &q, c, 4);	/* Gregorian period, sub-century */
//    qrem (&x, &y, jy, 4);	/* 4-years, years */
//    j = 146097*p + 36524*q + 1461*x + 365*y + (153*jm+2)/5 + day + 1721118;
// Since jy is non-negative, it can be shortened here...
    j = 146097*p + 36524*q + (1461*jy)/4 + (153*jm+2)/5 + day + 1721118;

    return  j;
}

static long int	jdJ (int year, int month, int day)
/* Convert year, month and day of Julian calendar to
  the integral part of Julian Date for 0h GMT.  Add
  0.5 days to obtain true Julian Date.    */
{
long int  j, c;
long int  jy, jm;
long int  k;

// Jdate2jd -- Julian calendar date to JD code
    qrem (&jy, &jm, month+9, 12);
    jy += year - 1;

    qrem (&c, &k, jy, 4);   /* 4-year periods, extra years */
    j = 1461*c + 365*k + (153*jm+2)/5 + day + 1721116;

    return  j;
}

long int	jDate (Date D)
/* Return the integer part of the Julian Date for the
  given calendar date.  Add 0.5 days to obtain true JD
  at 0h UTC on that date.  */
{
long int	r;
    if (D.Cal == 0) r = jdG (D.year, D.month, D.day);
    else	    r = jdJ (D.year, D.month, D.day);
    return  r;
}

int	dow (long int jd)
/* Return Day of Week for integer Julian Date.
  This is 0 for Sunday, 6 for Saturday.  
  The standard JD is UTC + 12h, so the integer jd
  for utc is (int)(JD - 0.5).  This jd is 0 mod 7
  on Tuesday, so we add 2 days before taking the
  modulo to return a Sunday based day.    */
{
    return  (int) mod (jd+2, 7);
}

int	dwD (Date D)
/* Return Day of Week for given Date.  Use the
  previously defined Day of Week program, and
  the Julian Day program.    */
{
    return  dow ( jDate (D) );
}

Date	MkDate (int year, int month, int day, int c)
/*  Utility to fill a Date structure  */
{
Date	D;

    D.year = year;
    D.month = month;
    D.day = day;
    D.Cal = c;

    return  D;
}

Date	Djd (long int jd, int c)
/* Convert the given integer part of JD to a Date in the
  given calender system (c).  Inverse of jdG or jdJ
  depending on calendar system.  The input integer jd is
  such that (jd + 0.5) is the Julian Date at 0h UTC on
  the resulting UT date.    */
{
Date	D;
long int        j, y, d, m;  //  , q, x;
long int	p, q;

  if (c == 0) {		// Gregorian
    j = jd - 1721118;		/* De-bias JD */
// Gj2date == Gregorian calendar, jd to y-m-d code
    qrem (&p, &q, j, 146097);	/* Gregorian period and 4-centuries */
    qrem (&y, &j, 4*q-1, 146097);	/* Century and quarter-days */
/* The y, j should work here, since q must be non-negative */
    y += 4*p;		/* Add Gregorian centuries of p */
    d = j/4;			/* Days in century */
    qrem (&j, &d, 4*d+3, 1461);   /* Year and quarter days */
    y = 100*y + j;		/* Combine year, century */
  } else {		// Julian or other
    j = jd - 1721116;		/* De-bias JD */
// Jd2date == Julian calendar, jd to y-m-d code
    qrem (&y, &d, 4*j-1, 1461);	/* Year and Quarter days in year */
  }
    d = (d+4)/4;		/* Days */
    qrem (&m, &d, 5*d-3, 153);	/* Month and 1/5 days in month */
    d = (d+5)/5;		/* Day in month */
    qrem (&p, &q, m+2, 12);
    D.year = y + p;		/* Correct year */
    D.month = 1 + q;		/* Correct month */
    D.day = d;			/* Day of month */
    D.Cal = c;

    return  D;
}



/* ---------------------------------------------------------------------- */

/* -- Time Stamp support -- */

// static	char	exchar (int c)
char	exchar (int c)
{
/* -- Returns the extended character coresponding to integer c;
  Should be in the ASCII sort compare sequence.  Normal return is
  0-9, A-Z, a-z.  A negative integer returns "."; there are
  62 values numbered 0-61; integers over 61 return "~" which is
  larger than any of the standard characters.  */

    if (c <  0) return '.';
    if (c < 10) return '0'+c;
    c -= 10;
    if (c < 26) return 'A'+c;
    c -= 26;
    if (c < 26) return 'a'+c;
    return '~';
}

/* --- Inverse extended character function... */
static	int	charex (char c)
{
/* -- Return an int which would result in c from exchar;
  The '.' returns -2, and '~' returns 63;
  Character values less than '.' give -3, between '.' and '0' give -1.
  Other unknown characters return 62.  '0'-'9', 'A'-'Z', 'a'-'z' give
  their proper extended values.  */
    if      (c <  '.') return -3;
    else if (c == '.') return -2;
    else if (c <  '0') return -1;
    else if (c <= '9') return c - '0';
    else if (c <  'A') return 62;
    else if (c <= 'Z') return 10 + c - 'A';
    else if (c <  'a') return 62;
    else if (c <= 'z') return 36 + c - 'a';
    else if (c == '~') return 63;
    else  return 62;
}

/* -- Time stamp formatting, from a unix system time... */
static	char*	tmstamp (time_t tx)
{
static	char	buffer[12];	/* Returns the value  */
struct	tm	*ut;

/* Find the ut time in a structure  */
    ut = gmtime (&tx);

/* Format the time into buffer  */
    sprintf (buffer, "%02d%02d%02d.%c%c%c",
	ut->tm_year % 100,
	ut->tm_mon + 1,
	ut->tm_mday,
	exchar(ut->tm_hour),
	exchar(ut->tm_min),
	exchar(ut->tm_sec)  );

    return  buffer;
}

/* -- A timestamp function, returning buffer with current time... */
char*	timestamp (time_t* tt)
{
time_t	now;

/* Find the current time, and return a time string for it, using
  the more general tmstamp routine above.  */
    time (&now);
    if (tt != NULL) *tt = now;
    return  tmstamp (now);
}

time_t	unstamp (char* ts)
{
/* Translate a time stamp, and provide the unix time; inverse of tmstamp */
int	y, m, d;
int	k;
char	ch, cm, cs;
int	days;
time_t	rt;		/* Result  */

    k = sscanf (ts, "%02d%02d%02d.%c%c%c",
	&y, &m, &d, &ch, &cm, &cs);
    if (k != 6) return -1;

/* Since there is no direct unix inverse to gmtime (mktime has the local
  time zone folded in, and is dodgy around daylight time changes) we find
  the unix time from our existing JD subroutines... */
    if (y < 70) y += 100;
    y += 1900;
    days = jdate (y, m, d) - SysEpoch;	/* Since System Epoch  */
    rt = days * 24 + charex (ch);	/* hours  */
    rt =  rt  * 60 + charex (cm);	/* minutes  */
    rt =  rt  * 60 + charex (cs);	/* seconds  */
    return  rt;
}


/* ---------------------------------------------------------------------- */

/* -- Linear Transforms -- */
/* Forward:  v = x*s + z;
   Inverse:  x = (v-z)/s;  */

double	tranfw (double x, double zp, double sf)
/* Forward transform */
{
    return (zp + x * sf);
}

double	tranrv (double x, double zp, double sf)
/* Inverse transform */
{
    if (sf == 0.0) return 0.0;
    return ( (x-zp) / sf);
}



/* ---------------------------------------------------------------------- */


		/* - - - - - - - - - - - - - - *\
		|				|
		|     Least Squares System	|
		|				|
		\* - - - - - - - - - - - - - - */

#undef  DBGLsq
//#define  DBGLsq
/* The DBGLsq symbol is defined for debug; it will cause reports of
  subscript usage and abort if any subscript falls out of range.  This
  will prove that no out of range subscripts occur in the use of the
  Least Squares solution.  It should be left undef in normal use.  */

/*  Instructions:

  1) Initialize the system
	For n coeficients, a double array dimensioned at least
	(n+2)^2 is required.  For convenience, it is dynamicaly
	allocated when the initialization call is made, and is
	released after the solution is obtained.  The only parameter
	to the initialization call is the number of coeficients.
	The value returned is an error indicator, 0 if all correct.

	err = kdclssq (n);

  2) Fill the matrix with each independent observation
	An equation of condition is
	  Sum(0<k<n)[a[k] * x[k]] = b
	where a is the array of coeficients to be determined,
	x is the array of independent variables, and b is the
	observed quantity.  For example, to fit a polynomial, we
	might have x[k] be simply x**k for k=0..n-1
	The fill entry is called with the x array pointer and b
	value for each observation; the user program fills the
	independent variable functions for each call.
	The value returned is an error indicator, 0 if correct.

	err = kdcfill (a, b);

  3) Solve the system to obtain fitted coeficients.
	The initialized and filled system is solved, the results
	are placed in supplied arrays and the storage is freed.
	The value returned is an error indicator, n is returned
	if correct, and a negative number if in error.

	error = kdcsolv (a, e);
	a contains the coeficients, e the errors; both are
	double arrays dimensioned at least n.

.. */



/* SubScript function mimics fortran subscripting */
#ifdef  DBGLsq
// Some debug globals
int	isl = 0;	// Internal subscript limit
int	ikp = 0;	// Internal kp value transfered
int	ikl = 0;	// Internal full subscript limit
int	dbn = 0;	// Debug initialized order

int	ilo, ihi;	// Subscript records
int	jlo, jhi;
int	klo, khi;

// Debug subroutines
static	int  SS (int i, int j)  // internal solver subscripting
// Debug version to check subscripts
{
int  k;

// Check kp versus initialized value
    if (ikp != (dbn+2)) {
	printf (" -- SS got kp = %d, should be %d.\n", ikp, dbn+2);
	fflush (stdout);
	abort ();
    }

// Check for limits on i and j
    if (i < 0) {
	printf (" --SS i (%d) less than 0.\n", i);
	fflush (stdout);
	abort ();
    }
    if (i > isl) {
	printf (" --SS i (%d) more than %d.\n", i, isl);
	fflush (stdout);
	abort ();
    }
    if (i < ilo) ilo = i;
    if (i > ihi) ihi = i;

    if (j < 0) {
	printf (" --SS j (%d) less than 0.\n", j);
	fflush (stdout);
	abort ();
     }
    if (j > isl) {
	printf (" --SS j (%d) more than %d.\n", j, isl);
	fflush (stdout);
	abort ();
    }
    if (j < jlo) jlo = j;
    if (j > jhi) jhi = j;

// Compute and check limits on k
    k = ikp * j + i;
    if (k < 0) {
	printf (" --SS k (%d) less than 0.\n", i);
	fflush (stdout);
	abort ();
    }
    if (k > ikl) {
	printf (" --SS k (%d) more than %d.\n", k, ikl);
	fflush (stdout);
	abort ();
    }
    if (k < klo) klo = k;
    if (k > khi) khi = k;

    return  k;
}

static	void	slreport (char* title, int low, int hiv, int blim, int tlim)
// Report in standard form
{
    printf (" %s] %d<%d:%d<%d", title, blim, low, hiv, tlim);
}

static	void	ssreport (char* title)
// Report SS subscript things
{
    printf ("I%s", title);
    slreport ("i", ilo, ihi, 0, isl);
    slreport ("j", jlo, jhi, 0, isl);
    slreport ("k", klo, khi, 0, ikl);
    printf ("\n");
    fflush (stdout);
}

#else  // on DBGLsq
#define  SS(i,j)  (i)+*kp*(j)
#endif


/* Locally defined solution program as copied from fortran... */

	/* - - - - - - - - - - *\
	|			|
	|      Local Solver	|
	|			|
	\* - - - - - - - - - - */

static	void  lssq (double *z, int *kp, int *kn, double* r, double* e)
/* add double *eps, int *ier */
{
/*
  PURPOSE:
	To solve a symmetric system of linear equations (such as the 
equations of condition in the problem of least squares) having one 
right-hand-side, and provide probable errors for least squares.

  PARAMETERS:
  Z  Double precision KP by KP matrix to solve AX = B where A is some
	M by N matrix with M not less than N, in a least squares sense
	(minimize norm of B-AX):
	1) Define A(M,N+1) = B(M) and KP = N + 2, and KN = M
	2) Do next step for (I=1, N+1) and (J=I, N+1)
	3) Set Z(I,J+1) = sum (K=1,M) of A(K,I) * A(K,J)
	The solution is returned in Z(KP-1,I) = X(I)
	The probable error is Z(KP,I) = PE(I)
 (The above refers to fortran subscripting)
  KP  - Workspace; for order N set KP = N+2
  KN  - If probable errors are desired, KN is the number of
	observations.  If not greater than KP-2, no errors
	are computed and zeros are returned

  REMARKS
	For best results the elements of the Z matrix should not
	differ from unity by more than a factor of about 2**32.

	Only the upper triangle of Z needs to be loaded.

  METHOD
	Choleski decomposition of a symmetric matrix, followed
	by a forward-ordered back-substitution.
*/



/* Former location of define SS here */

int	n, m;
int	i, j, k, l;
double	fno, sum;
double	p;
double	pef;

	m = *kp - 1;
	n = m - 1;
#ifdef  DBGLsq
	ikp = *kp;	// Pass it on...
#endif
/* -- Loop by rows over the matrix for decomposition -- */
	for (i=0; i<m; i++)  {
	  l = i + 1;

/*  Sum squares above diagonal  */
	  for (k=0; k<i; k++)
	    z[SS(i,l)] -= z[SS(k,l)] * z[SS(k,l)];

	  if (i >= n) continue;

/*  Find the diagonal element  */
	  z[SS(i,l)] = sqrt(z[SS(i,l)]);

/*  Compute forward decomposition on row  */
	  for (j=l+1; j<*kp; j++)  {
	    for (k=0; k<i; k++)
	      z[SS(i,j)] -= z[SS(k,l)] * z[SS(k,j)];
	    z[SS(i,j)] /= z[SS(i,l)];
	  }
	  z[SS(i,i)] = 1.0 / z[SS(i,l)];

/*  Compute forward ordered back substitution */
	  for (j=0; j<i; j++) {
	    sum = 0.0;
	    for (k=j; k<i; k++)
	      sum -= z[SS(k,l)] * z[SS(k,j)];
	    z[SS(i,j)] = z[SS(i,i)] * sum;
	  }
	}	/* end of loop by rows */

/*  Compute probable error factor */
	fno = *kn - n;
	pef = (fno < 1.0 || z[SS(n,m)] < 0.0) ? 0.0 :
		0.6744897502 * sqrt (z[SS(n,m)]/fno);
// Change the .67448 to 1.0 for standard error rather than probable error

#ifdef  DBGLsq
	ssreport ("Slvd subs:");
#endif

/*  Fill in the solution rows  */
	for (i=0; i<n; i++)  {
	  r[i] = sum = 0.0;
	  for (j=i; j<n; j++)  {
	    p = z[SS(j,i)];
	    r[i] += p * z[SS(j,m)];
	    sum += p * p;
	  }
	  if (e != NULL) e[i] = pef * sqrt(sum);
	}

#ifdef  DBGLsq
	ssreport ("fild subs:");
#endif
}

/*  NOTE:  Above is the matrix solution program, and below is the
code to fill the matrix, call the solution, and provide the solution
vectors.  These were previously two separate Fortran programs  */


	/* - - - - - - - - - - *\
	|			|
	|	 Globals	|
	|			|
	\* - - - - - - - - - - */

/* The least squares routines communicate by means of a static
  global storage, defined here. */

static int	kob;
static int	kn, kp;
static int	flag = 0;
static double	*work = NULL;
// #define  jsub(i,j)  (i)+(kp+1)*(j)
/* in fortran was  JSUB (I,J) = I + KP*(J-1)  */

/* SubScript function mimics fortran subscripting */
#ifdef  DBGLsq
// Some debug globals
int	fsl = 0;	// Internal subscript limit
int	fkp = 0;	// Internal kp value transfered
int	fkl = 0;	// Internal full subscript limit
int	fdbn = 0;	// Debug initialized order

int	ilof, ihif;	// Subscript records
int	jlof, jhif;
int	klof, khif;

// Debug subroutines
static	int  jsub (int i, int j)  //  fill subscripting
// Debug version to check subscripts
{
int  k;

// Check kp versus initialized value
    if (fkp != (fdbn+2)) {
	printf (" -- SS got kp = %d, should be %d.\n", fkp, fdbn+2);
	fflush (stdout);
	abort ();
    }

// Check for limits on i and j
    if (i < 0) {
	printf (" --SS i (%d) less than 0.\n", i);
	fflush (stdout);
	abort ();
    }
    if (i > fsl) {
	printf (" --SS i (%d) more than %d.\n", i, fsl);
	fflush (stdout);
	abort ();
    }
    if (i < ilof) ilof = i;
    if (i > ihif) ihif = i;

    if (j < 0) {
	printf (" --SS j (%d) less than 0.\n", j);
	fflush (stdout);
	abort ();
     }
    if (j > fsl) {
	printf (" --SS j (%d) more than %d.\n", j, fsl);
	fflush (stdout);
	abort ();
    }
    if (j < jlof) jlof = j;
    if (j > jhif) jhif = j;

// Compute and check limits on k
    k = fkp * j + i;
    if (k < 0) {
	printf (" --SS k (%d) less than 0.\n", i);
	fflush (stdout);
	abort ();
    }
    if (k > fkl) {
	printf (" --SS k (%d) more than %d.\n", k, fkl);
	fflush (stdout);
	abort ();
    }
    if (k < klof) klof = k;
    if (k > khif) khif = k;

    return  k;
}

// This is really the same as slreport above!
static	void	flreport (char* title, int low, int hiv, int blim, int tlim)
// Report in standard form
{
    printf (" %s] %d<%d:%d<%d", title, blim, low, hiv, tlim);
}

static	void	fsreport (char* title)
// Report SS subscript things
{
    printf ("J%s", title);
    flreport ("i", ilof, ihif, 0, fsl);
    flreport ("j", jlof, jhif, 0, fsl);
    flreport ("k", klof, khif, 0, fkl);
    printf ("\n");
    fflush (stdout);
}

#else  // on DBGLsq
#define  jsub(i,j)  (i)+(kp+1)*(j)
#endif


	/* - - - - - - - - - - *\
	|			|
	|      Initialize	|
	|			|
	\* - - - - - - - - - - */

int	kdclssq (int n)
/* Initialization; called once at start */
/* Returns error values:
	0 = Correctly initialized
	-1 = Storage not obtained
	-2 = Invalid parameter n (n < 1)
	-3 = Already initialized
.. */
{
int	k;

    if (n < 1) return  -2;  // exit(817);
    if (flag) return -3;	// Irregular initialize
    flag = 1;
    kn = n;
    kp = kn+1;
    kob = 0;
    k = kn+2;
    k *= k;		// For work array
/* Actually, fill and solve use subscripts 8:62 only, while the
  internal solution uses 0:63.  (for n=6)  */
    work = (double*) calloc ( (size_t) k, (size_t) (sizeof(double)) );
    if (work == NULL) {
	flag = 0;
	return  -1;
    }

#ifdef  DBGLsq
// Initialize global subscript tracking
    isl = n + 1;
    ikl = k - 1;
    dbn = n;
    ilo = jlo = n;
    klo = k+2;
    ihi = jhi = khi = -1;
// For fill subscripts
    fsl = n + 1;
    fkl = k - 1;
    fkp = kp + 1;
    fdbn = n;
    ilof = jlof = n;
    klof = k+2;
    ihif = jhif = khif = -1;
#endif  // on DBGLsq

    return  0;
}


	/* - - - - - - - - - - *\
	|			|
	|      Fill Matrix	|
	|			|
	\* - - - - - - - - - - */

/* #define  EPS  1.0e-15 */

int	kdcfill (double *a, double b)
/* Fill in values in array a (dimensioned n) and r.h.s. b; 
  returns error indicator:
	0 = all correct
	-1 = not initialized
.. */
{
int	i, j;

/* TEST FOR PROPER INITIALIZATION */
    if (!flag)  return -1;   // exit (818);
    kob++;

/* FORM SUMS FROM A */
    for (i=0; i<kn; i++)  {
	for (j=i; j<kn; j++)
	    work[jsub(i,j+1)] += a[i] * a[j];
    }

/* - - - -FORM SUMS FROM B */
    for (i=0; i<kn; i++)
	work[jsub(i,kp)] += b * a[i];
    work[jsub(kn,kp)] += b * b;

/* Return indicator */
    return  0;
}


	/* - - - - - - - - - - *\
	|			|
	|      Solve System	|
	|			|
	\* - - - - - - - - - - */

int	kdcsolv (double *v, double *p)
/* Solve the least squares system, and provide the coeficients
  and errors in the provided arrays, each of dimension n;
  returns an error indication:
	n = all correct
	-1 = not initialized
	-2 = insufficient observations
  v is the coeficients and p is the errors;
  After solving, the work storage is freed and cannot be used again
  for fill or solve without initializing it again.
.. */
{
int	j;

	if (!flag)  return -1;   // exit (818);
	if (kob < kn)  return -2;   // exit (819);
	
#ifdef  DBGLsq
	fsreport ("Fill subs:");  // change to report on fill subscripts
#endif

	j = kp+1;
	lssq (&work[0], &j, &kob, v, p);	// with direct fill

#ifdef  DBGLsq
	fsreport ("Solv subs:");
#endif

/* C- - - -SET FLAG TO INDICATE END */
/* Clear storage and indicator flag */
	if (work != NULL) { free (work); work = NULL; }
	flag = 0;
/* Return status */
	return  kn;
}


/* ---------------------------------------------------------------------- */


/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */

